#include <stdlib.h>
#include <stdio.h>

#include "bitoperations.h"

static const int testsize = 77;

int bit_is_set(int pos) {
  return (pos%2);
}

static void ltlassert(const char *msg, int expr) {
  if (!expr) {
    printf("assertion failed: %s\n.", msg);
    exit(1);
  }
}

static void ltlassertsame(const char *msg, int a, int b) {
  return ltlassert(msg, (a&&b) || (!a && !b));
}

void testcopy(int start, int size, int value) {
  bitfield *src,*dst;
  int i;
 
  src = (bitfield *) malloc(get_bitfield_memsize(testsize));
  dst = (bitfield *) malloc(get_bitfield_memsize(testsize));
  ltlassert("src allocated", src!=0);
  ltlassert("dst allocated", dst!=0);

  for (i=0; i<testsize; i++) {
    setbit(src, i, bit_is_set(i));
    setbit(dst, i, value);
  }
  copy_bitfield(src, dst, start, size);
  for (i=0; i<testsize; i++) {
    ltlassertsame("src unchanged", getbit(src, i), bit_is_set(i));
  }
  for (i=0; i<start; i++) {
    ltlassertsame("dst before start unchanged", getbit(dst, i), value);
  }
  for (i=start; i<start+size; i++) {
    ltlassertsame("dst in region copied", getbit(dst, i), bit_is_set(i));
  }
  for (i=start+size; i<testsize; i++) {
    ltlassertsame("dst behind region unchanged", getbit(dst, i), value);
  }

  free(src);
  free(dst);
}

void setlongbit(unsigned long *l, int pos, int value) {
  unsigned long mask = 1 << pos;
  if (value) {
    *l = *l | mask;
  } else {
    *l = *l & ~mask;
  }
}

void testlongcopy(int start, int size, int value) {
  unsigned long src;
  bitfield *dst;
  int i;

  dst = (bitfield *) malloc(get_bitfield_memsize(testsize));
  ltlassert("dst allocated", dst!=0);

  for (i=0; i<size; i++) {
    setlongbit(&src, i, bit_is_set(i));
  }
  for (i=0; i<testsize; i++) {
    setbit(dst, i, value);
  }
  setbit_by_long(dst, start, src, size);

  for (i=0; i<start; i++) {
    ltlassertsame("long: dst before start unchanged", getbit(dst, i), value);
  }
  for (i=start; i<start+size; i++) {
    ltlassertsame("long: dst in region copied", getbit(dst, i), bit_is_set(i-start));
  }
  for (i=start+size; i<testsize; i++) {
    ltlassertsame("long: dst behind region unchanged", getbit(dst, i), value);
  }

  free(dst);
}

void testsizes() {
  int size,i;
  for (size=0; size<1000; size++) {
    bitfield* field = (bitfield*) malloc(get_bitfield_memsize(size));
    for (i=0;i<size;i++) {
      setbit(field, i, 1);
    }
    for (i=0;i<size;i++) {
      setbit(field, i, 0);
    }    
    free(field);
  }
}

void testclone(int size) {
  bitfield *src = (bitfield*) malloc(get_bitfield_memsize(size));
  bitfield *dst = (bitfield*) malloc(get_bitfield_memsize(size));
  int i;
  for (i=0; i<size; i++) {
    setbit(src, i, bit_is_set(i));
  }
  clear_bitfield(dst, size);
  clone_bitfield(src, dst, size);
  for (i=0; i<size; i++) {
    ltlassertsame("field not equal after clone",
		  getbit(dst, i), bit_is_set(i));
  }
  free(src);
  free(dst);
}

void testinsert(int size, int offset, int range) {
  bitfield *src = (bitfield*) malloc(get_bitfield_memsize(range));
  bitfield *dst0 = (bitfield*) malloc(get_bitfield_memsize(size));
  bitfield *dst1 = (bitfield*) malloc(get_bitfield_memsize(size));
  int i;
  for (i=0; i<range; i++) {
    setbit(src, i, bit_is_set(i));
  }
  for (i=0; i<size; i++) {
    setbit(dst0, i, 0);
    setbit(dst1, i, 1);
  }
  insert_bitfield(src, dst0, offset, range);
  insert_bitfield(src, dst1, offset, range);
  for (i=0; i<range; i++) {
    ltlassertsame("insert: src modified", getbit(src,i), bit_is_set(i));
  }
  for (i=0; i<offset; i++) {
    ltlassertsame("insert: dst0 modified before offset", getbit(dst0,i), 0);
    ltlassertsame("insert: dst1 modified before offset", getbit(dst1,i), 1);
  }
  for (i=offset; i<offset+range; i++) {
    ltlassertsame("insert: dst0 not correctly inserted",
		  getbit(dst0,i), bit_is_set(i-offset));
    ltlassertsame("insert: dst1 not correctly inserted",
		  getbit(dst1,i), bit_is_set(i-offset));
  }
  for (i=offset+range; i<size; i++) {
    ltlassertsame("insert: dst0 modified after range", getbit(dst0,i), 0);
    ltlassertsame("insert: dst1 modified after range", getbit(dst1,i), 1);
  }
  free(src);
  free(dst0);
  free(dst1);
}

void testequalbitfields(int size) {
  bitfield *ref = (bitfield*) malloc(get_bitfield_memsize(size));
  bitfield *cpy = (bitfield*) malloc(get_bitfield_memsize(size));
  int i;
  for (i=0; i<size; i++) {
    setbit(ref, i, bit_is_set(i));
  }
  clone_bitfield(ref, cpy, size);
  ltlassert("equal_bitfields returned FALSE for equal bitfields",
	    equal_bitfields(ref, cpy, size));
  for (i=0; i<size; i++) {
    clone_bitfield(ref, cpy, size);
    setbit(cpy, i, !bit_is_set(i));
    ltlassert("equal_bitfields returned TRUE for different bitfields",
	      !equal_bitfields(ref, cpy, size));
  }
  free(ref);
  free(cpy);
}

void testnullbitfield(int size) {
  bitfield *f = (bitfield*) malloc(get_bitfield_memsize(size+1));
  int i;
  for (i=0; i<size; i++) {
    setbit(f, i, 0);
  }
  /* set the bit _behind_ the official bitfield, this should leave
     the bitfield be null */
  setbit(f, size, 1);
  ltlassertsame("is_null_bitfield() failed", 1, is_null_bitfield(f, size));
  free(f);
}

void testsubbitfield(int size) {
  bitfield *a = (bitfield*) malloc(get_bitfield_memsize(size));
  bitfield *b = (bitfield*) malloc(get_bitfield_memsize(size));
  int i;
  for (i=0; i<size; i++) {
    switch (i%3) {
    case 0: 
      setbit(a, i, 0);
      setbit(b, i, 0);
      break;
    case 1:
      setbit(a, i, 0);
      setbit(b, i, 1);
      break;
    case 2:
      setbit(a, i, 1);
      setbit(b, i, 1);
      break;
    default:
      ltlassert("uncovered case",0);
    }
  }
  ltlassertsame("is_sub_bitfield() failed (1)", 1, is_sub_bitfield(a, b, size));
  setbit(a, size-1, 1);
  setbit(b, size-1, 0);
  ltlassertsame("is_sub_bitfield() failed (2)", 0, is_sub_bitfield(a, b, size));
  free(a);
  free(b);
}

int main(int argc, char *argv[]) {
  bitfield *field;
  int i;

  bitoperations_init();

  field = (bitfield *) malloc(get_bitfield_memsize(testsize));
  ltlassert("field allocated", field!=0);

  for (i=0; i<testsize; i++) {
    setbit(field,i,bit_is_set(i));
  }
  for (i=0; i<testsize; i++) {
    ltlassertsame("bit is correctly set", getbit(field,i), bit_is_set(i));
  }

  clear_bitfield(field, testsize);
  for (i=0; i<testsize; i++) {
    ltlassert("after clear_bitfield: field is 0", getbit(field, i)==0);
  }

  free(field);

  testcopy(0,50,1);
  testcopy(0,50,0);
  testcopy(31,20,1);
  testcopy(31,20,0);
  testcopy(6,5,1);
  testcopy(6,5,0);
  testcopy(31,3,0);
  testcopy(20,52,1);
  testcopy(20,52,0);

  testlongcopy(0,30,1);
  testlongcopy(0,30,0);
  testlongcopy(0,15,1);
  testlongcopy(0,15,0);
  testlongcopy(29,30,1);
  testlongcopy(29,30,0);
  testlongcopy(29,15,1);
  testlongcopy(29,15,0);

  testclone(8);
  testclone(32);
  testclone(33);
  testclone(145);
  
  testinsert(8, 4, 2);
  testinsert(8, 6, 2);
  testinsert(70, 28, 16);
  testinsert(280, 10, 210);

  testequalbitfields(8);
  testequalbitfields(32);
  testequalbitfields(33);
  testequalbitfields(145);

  testnullbitfield(8);
  testnullbitfield(250);

  testsubbitfield(8);
  testsubbitfield(32);
  testsubbitfield(33);
  testsubbitfield(64);
  testsubbitfield(65);

  testsizes();

  printf("bitoperations: test passed.\n");

  return 0;
}

