#include <assert.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "hashtrie.h"
#include "shm_common.h"

const index_t root = (index_t) {0};

#ifndef UPDATE_INFOS
static void update_infos(void *infos, void *new_infos) {
    char *old = (char*) infos;
    char *new = (char*) new_infos;
    int i;
    for (i = 0; i < TRIE_ADDITIONAL_DATA; i++) {
        old[i] |= new[i];
    }
}
#else
extern void update_infos(void *infos, void *new_infos);
#endif




int calculate_pos(char *key, int depth) {
    // pattern repeats after 8 times
    int offset = depth / 8;
    int moddepth = depth % 8;

    char *offi = key + 5 * offset;

    uint64_t a = 0;
    int i;
    for (i = 0; i < 5; i++) {
       a += (uint64_t) (offi[i] * pow(2, 8*i));
    }
    uint64_t c = offi[0] + offi[1] * pow(2, 8) + offi[2] * pow(2, 16)
               + offi[3] * pow(2, 24) + offi[4] * pow(2, 32);
    assert(c == a);

    uint64_t b = a >> (moddepth * 5);
    int x = b & 31;
    return x;
}


int is_digest(tNode *node, int pos) {
    index_t index = node->indices[pos];
    int bit_at = (index.a >> 15) & 1;
    return bit_at;
}



index_t skip_first_layer(char key[TRIE_HASHLEN], int *branch) {
    *branch = calculate_pos(key, 0);
    index_t index = (lookup_node(root))->indices[*branch];
    return index;
}


size_t trie_get_checked_states_count(void) {
    struct hash_meta_infos x;
    get_meta_infos(&x);
    return x.checked_states;
}

size_t trie_count_elements(void) {
    struct hash_meta_infos x;
    get_meta_infos(&x);
    return x.pos_hashes;
}

char *trie_iterate(size_t *blocklen) {
    struct hash_meta_infos x;
    get_meta_infos(&x);
    lock_digests_read();
    char *res = get_shm_digests(x.pos_hashes);
    release_digests_read();
    *blocklen = x.pos_hashes * TRIE_DATALEN;
    return res;
}

int get_max_depth(void) {
    struct hash_meta_infos x;
    get_meta_infos(&x);
    return x.max_depth;
}

static inline int is_empty_pos(tNode *node, int pos) {
    // if it's empty then everything is at max value
    index_t index = node->indices[pos];

    return index.a == (uint16_t) -1 && index.b == (uint32_t) -1;
}

static size_t h_teardown(tNode *node) {
    size_t i, sum = 0;
    for (i = 0; i < 32; i++) {
        if (is_empty_pos(node, i))
            continue;
        if (is_digest(node, i)) {
            sum++;
        } else {
            sum += h_teardown(lookup_node(node->indices[i]));
        }
    }
    return sum;
}

size_t teardown_hashtrie(void) {
    size_t res = h_teardown(lookup_node(root));
    return res;
}

inline int is_checked(void *infos) {
    unsigned char *ptr = (unsigned char*) infos;
    // TODO: implementation
    // could be: first bit indicates checked/unchecked status
    return (ptr[0] & 128) == 128;
}

static void *h_get(index_t id, char *key, int depth, const int branchpos) {
    tNode *node = lookup_node(id);
    int pos = calculate_pos(key, depth);
    if (!is_digest(node, pos)) {
        return h_get(node->indices[pos], key, depth+1, branchpos);
    } else {
        lock_nodes_read(branchpos);
        if (is_digest(node, pos)) {
            if (is_empty_pos(node, pos)) {
                release_nodes_read(branchpos);
                return NULL;
            } else {
                char *ref = lookup_digest(node->indices[pos]);
                release_nodes_read(branchpos);
                if (memcmp(key, ref, TRIE_HASHLEN) != 0) {
                    return NULL;
                } else {
                    return ref + TRIE_HASHLEN;
                }
            }
        } else {
            release_nodes_read(branchpos);
            return h_get(node->indices[pos], key, depth+1, branchpos);
        }
    }
}

void *get(char key[TRIE_HASHLEN]) {
    int branch;
    index_t node = skip_first_layer(key, &branch);
    return h_get(node, key, 1, branch);
}

int trie_contains(char key[TRIE_HASHLEN]) {
    void *resptr = (void*) get(key);
    if (resptr == NULL) {
        return 0;
    } else if (is_checked(resptr)) {
        return 1;
    } else {
        return -1;
    }
}

int trie_contains_processed(char key[TRIE_HASHLEN]) {
    return trie_contains(key) == 1;
}


static index_t transform_with(index_t digest_id, int newdepth);

static void init_node(struct shm_node *res) {
    alloc_node(res);
    memset(res->nodeptr->indices, 0xFF, 32 * sizeof(index_t));
}

static void set_digest(tNode *node, int pos) {
    uint16_t bits = node->indices[pos].a;
    uint16_t newbit = 1 << 15;
    node->indices[pos].a = bits | newbit;
}

static void set_node(tNode *node, int pos) {
    uint16_t bits = node->indices[pos].a;
    uint16_t newbit = 1 << 15;
    node->indices[pos].a = bits & ~newbit;
}


static index_t insert(char *digest, void *infos) {
    struct shm_digest x;
    lock_digests_write();
    alloc_digest(&x);
    char *target = x.ptr;
    memcpy(target, digest, TRIE_HASHLEN);
    memcpy(target + TRIE_HASHLEN, infos, TRIE_ADDITIONAL_DATA);

    release_digests_write();

    int checked = is_checked(infos);
    if (checked) {
        inc_checked_states();
    }
    return x.index;
}


static void add(index_t index, char key[TRIE_HASHLEN], void *infos, int depth, const int branchpos) {
    tNode *node = lookup_node(index);
    assert(depth < 31);
    int pos = calculate_pos(key, depth);

    if (!is_digest(node, pos)) {
        add(node->indices[pos], key, infos, depth+1, branchpos);
    } else {
        lock_nodes_write(branchpos);
        if (is_digest(node, pos)) {
            if (is_empty_pos(node, pos)) {
                node->indices[pos] = insert(key, infos);
                set_digest(node, pos);
                release_nodes_write(branchpos);
                update_max_depth(depth);
            } else {
                char *digest = lookup_digest(node->indices[pos]);
                if (memcmp(digest, key, TRIE_HASHLEN) == 0) {
                    release_nodes_write(branchpos);
                    // TODO: this line seems incorect, might throw off the counting
                    if (is_checked(infos) && !is_checked(digest + TRIE_HASHLEN))
                        inc_checked_states();
                    update_infos(digest + TRIE_HASHLEN, infos);
                    /* TODO: move this code
                    int i;
                    for (i = 0; i < TRIE_ADDITIONAL_DATA; i++) {
                        digest[TRIE_HASHLEN + i] |= infos[i];
                    }
                    */
                } else {
                    index_t succ = transform_with(node->indices[pos], depth+1);
                    node = lookup_node(index); // memory block may be moved; recalculate the pointer is necessary
                    node->indices[pos] = succ;
                    set_node(node, pos);
                    release_nodes_write(branchpos);
                    add(node->indices[pos], key, infos, depth+1, branchpos);
                }
            }
        } else {
            release_nodes_write(branchpos);
            add(node->indices[pos], key, infos, depth+1, branchpos);
        }
    }
}

// MAYMOVE shm node structure
static index_t transform_with(index_t digest_index, int newdepth) {
    struct shm_node r;
    init_node(&r);
    tNode *node = r.nodeptr;
    char *digest = lookup_digest(digest_index);

    int pos = calculate_pos(digest, newdepth);
    assert(is_empty_pos(node, pos));
    node->indices[pos] = digest_index;
    set_digest(node, pos);

    return r.index;
}

void trie_put(char digest[TRIE_HASHLEN], void *infos) {
    //assert(digest[0] == 0 || digest[0] == 1);
    int branch;
    index_t node = skip_first_layer(digest, &branch);
    add(node, digest, infos, 1, branch);
}

