% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(b_interpreter_components,[
  reset_component_info/1,
  b_trace_test_components/3, b_trace_test_components/4,
  unsat_component/1, unsat_component_enumeration_warning/1,
  unsat_component_abort_error/1,
  unsat_conjunct_inside_component/3,
  det_solution_for_constant/2,
  b_tracetest_boolean_expression/5,
  observe_state/3,

  construct_optimized_exists/3, construct_optimized_exists/4
  ]).

:- use_module(b_interpreter,[b_test_boolean_expression/6]).
:- use_module(preferences).
:- use_module(kernel_waitflags).
:- use_module(tools).
:- use_module(translate).
:- use_module(error_manager).
:- use_module(library(lists)).
:- use_module(b_enumerate, [b_trace_tighter_enumerate_values/2]).
:- use_module(chrsrc(chr_integer_inequality),[attach_chr_integer_inequality_to_wf_store/1]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,interpreter).
:- module_info(description,'This module can be used to split a predicate into components and solving each componente individually (with tracing feedback).').

/* -------------- */
/*  dynamic bits  */
/* -------------- */

:- dynamic skipped_unsat_conjunct/0, unsat_conjunct_inside_component/3.
reset_unsat_conjunct_inside_component_info :-
       retractall(skipped_unsat_conjunct),
       retractall(unsat_conjunct_inside_component(_,_,_)).
reset_tracetest_for_component :- retractall(skipped_unsat_conjunct). % to be called before trace checking a component

% individual conjuncts which are directly unsatisfiable
assert_skipped_failed_conjunct(ComponentNr,BE) :- assert_skipped_unsat_conjunct,
        assert(unsat_conjunct_inside_component(ComponentNr,BE,false)),
        (preferences:get_preference(provide_trace_information,false) -> true
          ; print(' *** INCONSISTENCY DETECTED, SKIPPING: '), print_bexpr_with_limit(BE,150),nl,
            print(' *** COMPONENT NUMBER: '), print(ComponentNr),nl).
assert_skipped_unknown_conjunct(ComponentNr,BE) :- assert_skipped_unsat_conjunct,
        assert(unsat_conjunct_inside_component(ComponentNr,BE,unknown)),
        (preferences:get_preference(provide_trace_information,false) -> true
          ; print(' *** POTENTIAL INCONSISTENCY DETECTED (** ENUMERATION WARNING **), SKIPPING: '),
            print_bexpr_with_limit(BE,150),nl,
            print(' *** COMPONENT NUMBER: '), print(ComponentNr),nl).
assert_skipped_unsat_conjunct :-
  (skipped_unsat_conjunct -> true ; assert(skipped_unsat_conjunct)).


:- dynamic unsat_component/1, unsat_component_enumeration_warning/1, unsat_component_abort_error/1.
reset_unsat_component_info :-
     retractall(unsat_component(_)),
     retractall(unsat_component_enumeration_warning(_)),
     retractall(unsat_component_abort_error(_)).

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(clear_specification,reset_unsat_component_info,
                           'Reset b_interpreter_components unsat component info.').

store_unsat_component_nr(X) :-
  (unsat_component(X) -> true ; assert(unsat_component(X))),
  ((enumeration_warning_occured_in_error_scope,
   \+ unsat_component_enumeration_warning(X))
    -> assert(unsat_component_enumeration_warning(X)),
       debug_println(9,enumeration_warning_in_component(X))
    ; true),
  ((abort_error_occured_in_error_scope,
   \+ unsat_component_abort_error(X))
    -> assert(unsat_component_abort_error(X)), print(abort(X)),nl,
       debug_println(9,unsat_component_abort_error(X))
    ; true).
l_ground_wait_flags(WF) :- l_ground_wait_flags(WF,Result),
   (Result=false -> (!,fail) ; true).
unsat_component_exists :- unsat_component(_),!.

:- public nr_of_components/1.
:- dynamic nr_of_components/1.
:- dynamic  allow_skipping_over_components/0.
reset_component_info(AllowSkip) :- retractall(allow_skipping_over_components),
   (AllowSkip==true -> assert(allow_skipping_over_components) ; true),
   reset_unsat_component_info,
   reset_unsat_conjunct_inside_component_info.


/* -------------- */


:- use_module(library(avl)).
% You can use the term one as ComponentNr if you don't know the number or it is not relevant
b_tracetest_boolean_expression(BE,LS,S,WF,ComponentNr) :-
   empty_avl(Ai),
   b_tracetest_boolean_expression1(BE,LS,S,WF,ComponentNr,Ai,_Ao).

b_tracetest_boolean_expression1(b(Expr,_,Infos),LS,S,WF,ComponentNr,Ai,Ao) :- !,
	b_tracetest_boolean_expression2(Expr,Infos,LS,S,WF,ComponentNr,Ai,Ao).
b_tracetest_boolean_expression2(conjunct(LHS,RHS),_Info,LocalState,State,WF,ComponentNr,Ai,Ao) :-
     % \+(Info = [try_smt|_]),
     !,
%%   print('check lhs of conjunct: '), translate:print_bexpr(LHS),nl,
        b_tracetest_boolean_expression1(LHS,LocalState,State,WF,ComponentNr,Ai,Aii),
%%   print('check rhs of conjunct: '), translate:print_bexpr(RHS),nl,
        b_tracetest_boolean_expression1(RHS,LocalState,State,WF,ComponentNr,Aii,Ao).
b_tracetest_boolean_expression2(BE,Infos,LS,S,WF,ComponentNr,Ai,Ao) :-
    (preferences:get_preference(provide_trace_information,true)
      -> nl,print(' ====> '),print_bexpr_with_limit(BE,250),nl ; true),
    if(safe_b_test_boolean_expression(ComponentNr,b(BE,pred,Infos),LS,S,WF,Ai,Ao),true,   %% b_det_test_boolean_expression2
       (assert_skipped_failed_conjunct(ComponentNr,b(BE,pred,Infos)),
        Ai=Ao)).

safe_b_test_boolean_expression(ComponentNr,BE,LS,S,WF,Ai,Ao) :-
    catch_enumeration_warning_exceptions(
       b_test_boolean_expression(BE,LS,S,WF,Ai,Ao), % this should not backtrack (without WF0 set); otherwise we may get problems with error blocks nesting
        (assert_skipped_unknown_conjunct(ComponentNr,BE),Ai=Ao), true). % true: add exceptions as event errors

:- use_module(bsyntaxtree,[predicate_components_in_scope/3,
           predicate_identifiers/2,predicate_identifiers_in_scope/3]).

get_typedval_id(typedval(_Val,_Type,ID,_),ID).
b_trace_test_components(Properties,ConstantsState,TypedVals) :-
   %format('b_trace_test_components(~w,~w,~w)',[Properties,ConstantsState,TypedVals]),
   b_trace_test_components(Properties,ConstantsState,TypedVals,all).
b_trace_test_components(Properties,ConstantsState,TypedVals,RelevantVariables) :-
    b_trace_test_components(Properties,ConstantsState,TypedVals,RelevantVariables,[]).
% in constract to /3 and /4 b_trace_test_components/5 does not fail if the predicate is unsatisfiable.
% instead it returns the unsatisfiable component as the last argument
% the other predicates still fail because they expect to find an empty list there
b_trace_test_components(Properties,ConstantsState,TypedVals,RelevantVariables,UnsatConjuncts) :-
    % we now treat all TypedVals as local variables: ensure that we count them in predicate_components/predicate_identifiers (in case they clash with global identifiers)!
    maplist(get_typedval_id,TypedVals,LocalVars),
    (preferences:get_preference(partition_predicates,true)
      -> predicate_components_in_scope(Properties,LocalVars,Components),
         length(Components,Len), print_trace_message(nr_of_components(Len))
      ;  predicate_identifiers_in_scope(Properties,LocalVars,PropVars),
         Components = [component(Properties,PropVars)], Len=1
     ),
    retractall(nr_of_components(_)),assert(nr_of_components(Len)),
    reset_observe_state,
    % print(components(Components)),nl,
    b_trace_test_components2(Components,ConstantsState,TypedVals,RelevantVariables,1,WFList),
    get_unsat_component(WFList,Components,UnsatConjuncts).


get_unsat_component(_,Components,UnsatConjuncts) :-
    unsat_component_exists, !,
    debug_println(9,unsatisfiable_predicate),
    unsat_component(UnsatComponentNr),
    nth0(UnsatComponentNr,Components,component(UnsatConjuncts,_)).
get_unsat_component(WFList,_Components,[]) :-
    print_trace_message(grounding_wait_flags),
    l_ground_wait_flags(WFList).
get_unsat_component(_,Components,UnsatConjuncts) :-
    unsat_component(UnsatComponentNr),
    nth0(UnsatComponentNr,Components,component(UnsatConjuncts,_)).


:- use_module(b_ast_cleanup,[predicate_level_optimizations/2]).
print_trace_message(Msg) :-
   (preference(provide_trace_information,false) -> true
     ; print_message(Msg)).
b_trace_test_components2([],_ConstantsState,TypedVals,_,Nr,ResWF) :-
  (TypedVals=[] -> print_trace_message(finished_processing_component_predicates),
                   ResWF=[]
     ; print_trace_message(enumerating_constants_without_constraints(TypedVals)),
       init_wait_flags(WF),
       b_trace_tighter_enumerate_values(TypedVals,WF), ResWF=[wf_comp(WF,Nr)]).
b_trace_test_components2([component(Pred,ComponentVars)|T],ConstantsState,TypedVals,RelevantVariables,
                         Nr,ResWF) :-
		N1 is Nr+1,
    ((RelevantVariables\=all, ComponentVars\= all,
      lists_are_disjoint(RelevantVariables,ComponentVars))
     -> print_trace_message(skipping_irrelevant_component(Nr,ComponentVars)),
		b_trace_test_components2(T,ConstantsState,TypedVals,RelevantVariables,N1,ResWF)
     ;  debug_println(9,checking_component_properties(Nr,ComponentVars)), %%
        %% print(component(Nr)),nl, translate:nested_print_bexpr_as_classicalb(Pred),debug:nl_time,
        ResWF = [wf_comp(WF,Nr)|TWF],
		init_wait_flags(WF),
        attach_chr_integer_inequality_to_wf_store(WF),
		(ComponentVars=all -> InTypedVals=TypedVals, OutTypedVals=[], ProjectedState=ConstantsState
		  ; project_typed_vals(TypedVals,ComponentVars,InTypedVals,OutTypedVals),
			project_state(ConstantsState,ComponentVars,ProjectedState) % relevant, e.g., for exists which may wait on state becoming ground
		 ),
		observe_state(InTypedVals,WF,Nr), %%
		reset_tracetest_for_component,
		predicate_level_optimizations(Pred,OPred),
		%%nl,nl,print(component(Nr,ComponentVars,ProjectedState)),nl,nl,translate:print_bexpr(Pred),nl,nl,
		b_tracetest_boolean_expression(OPred,[],ProjectedState,WF,Nr),
		%print(done_comp(Nr,WF)),nl,
		(skipped_unsat_conjunct
		 -> store_unsat_component_nr(Nr),
		    allow_skipping_over_components % check if we allow skipping over unsat components to maybe find as many values for constants as possible ,...
		  ; %print_bt_message(start_enumerating_constants(Nr)), %%(TypedVals)), %%
			b_trace_tighter_enumerate_values(InTypedVals,WF),
			(preferences:get_preference(disprover_mode,false) -> true
			  ; ground_wait_flag0(WF)) % ensure that we give each component the chance to run WF0 inconsistency checks
		),
		% we could ground waitflag 1 (deterministic computations) ?
		b_trace_test_components2(T,ConstantsState,OutTypedVals,RelevantVariables,N1,TWF)
	).

:- use_module(debug,[debug_println/2, debug_mode/1]).
:- use_module(tools,[start_ms_timer/1, stop_ms_timer/2]).
l_ground_wait_flags([],true).
l_ground_wait_flags([wf_comp(WF,ComponentNr)|TWF],SatResult) :- %trace,
   (debug_mode(off) ->  true ; print(grounding_component(ComponentNr)),nl, start_ms_timer(Timer)),
   % print_bt_message(grounding_component(ComponentNr)), portray_waitflags(WF),nl, %%
   % maybe we should set up new error scope ? <--- TO DO
   if(ground_wait_flags(WF),
       ((debug_mode(on) -> stop_ms_timer_with_msg(Timer,component(ComponentNr)) ; true),
         l_ground_wait_flags(TWF,SatResult) % continue with other components
        ),
       ((debug_mode(on) -> print(unsatisfiable_component(ComponentNr)),nl,
                                 stop_ms_timer_with_msg(Timer,component(ComponentNr)) ; true),
        store_unsat_component_nr(ComponentNr),
        preferences:get_preference(disprover_mode,false), % in Disprover mode we can stop straightaway; we are not interested in finding values for identifiers in other components; unless we have an enumeration warning !!
        % TO DO: keep track of enumeration warnings per component
        SatResult=false)).

project_typed_vals([],_,[],[]).
project_typed_vals([typedval(Val,Type,VarID,EnumWarning)|Rest],ComponentVars,In,Out) :-
   (member(VarID,ComponentVars)
     -> In=[typedval(Val,Type,VarID,EnumWarning)|In2], Out2=Out
     ;  Out=[typedval(Val,Type,VarID,EnumWarning)|Out2], In2=In
   ), project_typed_vals(Rest,ComponentVars,In2,Out2).
project_state([],_,[]).
project_state([bind(VarID,Value)|Rest],ComponentVars,Out) :-
   (member(VarID,ComponentVars)
     -> Out=[bind(VarID,Value)|Out2]
     ;  Out=Out2
   ), project_state(Rest,ComponentVars,Out2).


:- use_module(state_packing,[pack_value/2,unpack_value/2]).
 :- volatile det_solution_for_constant_packed/2.
 :- dynamic det_solution_for_constant_packed/2.
assert_det_solution_for_constant(Var,Val) :- %print(sol(Var)),tools:print_memory_used_wo_gc,nl,nl,
    (value_too_large(Val) -> PVal = '$TOO_LARGE' ,debug_println(4,not_storing(Var))
     ; pack_value(Val,PVal)),
    assert(det_solution_for_constant_packed(Var,PVal)).
    %,tools:print_memory_used_wo_gc,nl.
det_solution_for_constant(Var,Val) :- det_solution_for_constant_packed(Var,PVal),
    (PVal = '$TOO_LARGE'
      -> Val=[], add_warning(no_det_value_stored_for_constant,'No deterministic value stored for constant (set ALLOW_INCOMPLETE_SETUP_CONSTANTS to TRUE to force storing all constants): ',Var)
      ;  unpack_value(PVal,Val)
    ).

:- use_module(avl_tools,[avl_height_less_than/2]).
value_too_large(avl_set(A)) :-
    preference(allow_incomplete_partial_setup_constants,false), %if true we always store the value
    \+ avl_height_less_than(A,7).
    % TO DO: it could be a small set but include large subsets !

:- use_module(tools_printing,[print_term_summary/1]).
reset_observe_state :- retractall(det_solution_for_constant_packed(_,_)).
observe_state(Env,WF,ComponentNr) :-
  get_wait_flag0(WF,WF0), get_wait_flag1(observe_state,WF,WF1),
  (preferences:get_preference(provide_trace_information,true)
    -> when((ground(Env);nonvar(WF0)),print_success(WF0,ComponentNr,Env,WF))
      %% ,when(nonvar(WF1),(nl, print('Starting Phase 1'),nl, portray_waitflags(WF)))
    ; true),
  observe_state2(Env,WF0,WF1).
  %,when(nonvar(WF0),(print(wf0),nl,portray_waitflags(WF))),  when(nonvar(WF1),(print(wf1),nl)).

print_success(WF0,_ComponentNr,_,_WF) :- nonvar(WF0),!.
print_success(_,ComponentNr,Env,_WF) :-
    format('~n-->> SUCCESS; all constants valued in component ~w~n',[ComponentNr]),
    (debug_mode(on)
     -> findall(V,member(typedval(_,_,V,_),Env),Vars),
        statistics(walltime,[W,_]), format('-->> (walltime: ~w ms) : variables = ~w~n',[W,Vars])
      ; true),
                     %%,check_for_multiple_solutions(WF,ComponentNr) %% comment in to check for multiple solutions
    nl.

observe_state2([],_,_).
observe_state2([typedval(Val,_Type,VarID,_EnumWarning)|T],WF0,WF1) :-
    observe_variable(VarID,Val,WF0,WF1),observe_state2(T,WF0,WF1).
%observe_state2([bind(Var,Val)|T],WF0) :- observe_variable(Var,Val,WF0), observe_state2(T,WF0).

% print a warning message if we find multiple solutions *after* all constants have been valued
%:- public check_for_multiple_solutions/2.
%check_for_multiple_solutions(WF,ComponentNr) :- get_enumeration_finished_wait_flag(WF,EWF),
%    retractall(solution_found(ComponentNr,_)),
%    add_solution(EWF,ComponentNr).
%:- volatile solution_found/2.
%:- dynamic solution_found/2.
%:- block add_solution(-,?).
%add_solution(_,ComponentNr) :- retract(solution_found(ComponentNr,Nr)),!,
%    print('*** Multiple solutions found *after valuation* for component '), print(ComponentNr),nl,
%    N1 is Nr+1,
%    print('*** Solution: '),print(N1),nl,
%    assert(solution_found(ComponentNr,N1)).
%add_solution(_,ComponentNr) :- % print(add_sol(ComponentNr)),nl,
%    assert(solution_found(ComponentNr,1)).

observe_variable(Var,Val,WF0,_WF1) :- preferences:get_preference(provide_trace_information,false),!,
    observe_variable_no_trace(Var,Val,WF0).
observe_variable(Var,Val,WF0,WF1) :- ground_value_check(Val,GrVal),
   when((nonvar(GrVal);nonvar(WF0)),
        (var(WF0) ->  print('-->> '),print(Var), nl,print('     '),print_term_summary(value(Val)),print_walltime,nl,
                      assert_det_solution_for_constant(Var,Val)
                   ;  print(' :?: '),print(Var),print(' '), print_term_summary(value(Val)),
                      observe_variable1(Var,Val,WF1)
                   )
                   ).
:- use_module(kernel_tools,[ground_value/1,ground_value_check/2]).
:- block observe_variable_no_trace(?,-,-).
observe_variable_no_trace(Var,Val,WF0) :- nonvar(WF0),!, % important e.g. for test 1368
   (ground_value(Val) -> assert_det_solution_for_constant(Var,Val) ; true).
observe_variable_no_trace(Var,Val,WF0) :- ground_value_check(Val,GV),
   %print_term_summary(gr_val_check(Var,Val,GV)),nl,
   observe_variable_no_trace4(Var,GV,Val,WF0).
:- block observe_variable_no_trace4(?,-,?,-).
observe_variable_no_trace4(_,GV,_,_) :- var(GV),!.
observe_variable_no_trace4(Var,_,Val,_) :- assert_det_solution_for_constant(Var,Val). % this could be expensive !?

print_walltime :- statistics(walltime,[Tot,Delta]),
   (debug_mode(on) -> print('     '),tools:print_memory_used_wo_gc ; true),
   format('     ~w ms (Delta ~w ms) ',[Tot,Delta]).

observe_variable1(Var,Val,WF1) :- ground_value_check(Val,GV),
   when((nonvar(GV);nonvar(WF1)),
      (nonvar(GV) -> print('--1-->> '),print(Var), nl,print('     '),print_term_summary(value(Val)),
          assert_det_solution_for_constant(Var,Val)
        ; print('  :?1?: '),print(Var),print(' '), print_term_summary(value(Val))
         %% , when(ground(Val), (print('--ground-->> '),print(Var),print(' '), print_term_summary(value(Val)))) % comment in to see grounding
         %% , external_functions:observe_state([bind(Var,Val)]) % comment in to observe values
         %% , translate:print_bstate([bind(Var,Val)]),nl % comment in to print full value with CLP(FD) info
      ) ).

/*
comment back in for finer-grained observation and uncomment call above
:- block observe_instantiation(-,?,?).
observe_instantiation(X,Var,Val) :-
  print(bind(X)),nl,translate:print_bstate([bind(Var,Val)]),nl, (debug_mode(on) -> trace ; true),
   X =.. [_|Args],
   observe2(Args,Var,Val).

observe2([],_,_).
observe2([H|T],Var,Val) :- (var(H) -> observe_instantiation(H,Var,Val)
  ; H=.. [_|Args], observe2(Args,Var,Val)),
  observe2(T,Var,Val).
*/

% --------------------


:- use_module(bsyntaxtree,[definitely_not_empty_set/1]).


construct_optimized_exists(Parameters,Body,Pred) :-
   construct_optimized_exists(Parameters,Body,Pred,true).

% we could also partition inside b_interpreter ?
construct_optimized_exists(Parameters,Body,Pred,Simplify) :-
% we can actually do more: first project out all conjuncts which do not have any variables in common with Parameters
%  #x.(P(x) & Q)  <=>  #x.(P(x)) & Q
% we can also use a stronger version of predicate components  #(x,y).( P(x,v) & Q(v,y)) <=>  #(x).P(x,v) & #(y).Q(v,y)
   get_texpr_ids(Parameters,PIDs),
   %print(constructing_optimized_exists(PIDs)),nl,
   predicate_components_with_restriction(Body,[],PIDs,Components), % should we use LocalVars instead of [] ?
   %length(Components,Len), print(nr_of_components_for_exists(Len,Components)),nl,
   conjunct_components(Components,Parameters,Pred,Simplify).
%   print('* EXISTS: '),translate:print_bexpr(Pred),nl.

:- use_module(bsyntaxtree).
conjunct_components(C,Parameters,Res,Simplify) :- conjunct_components2(C,Parameters,List,Simplify),
  conjunct_predicates(List,Res).

conjunct_components2([],_Parameters,[],_). % :- (Parameters=[] -> true ; print(unused_existential_var(Parameters)),nl).
conjunct_components2([component(Pred,ComponentVars)|T],Parameters,Res,Simplify) :-
   augment_ids(ComponentVars,CV), % translate atomic id names into b(identifier(...
   tools:list_intersection(CV,Parameters,ExistParas),
   %print(ord_inter(CV,Parameters,ExistParas)),nl,
   (Simplify=true
    -> create_and_simplify_exists(ExistParas,Pred,EP)
     ; create_unsimplified_exists(ExistParas,Pred,EP)),
   (EP=b(truth,pred,[]) -> Res=TRes
   ; Res=[EP|TRes]),
   conjunct_components2(T,Parameters,TRes,Simplify).

% translate a list of atomic ides into ids with the b/3 wrapper
augment_ids([],[]).
augment_ids([ID|T],[b(identifier(ID),_,_)|TT]) :- augment_ids(T,TT).

extract_equalities(Parameters,Pred, ReducedParas,OutPred) :-
   extract_equalities2(Pred,Parameters,ReducedParas,NewPred,[],Subst),
   %sort(Subst,SortedSubst),
   %print(extracted(Subst)),nl,
   apply_subst(Subst,NewPred,OutPred).
   % apply equality inside expressions: note: this may duplicate expressions !
   % However, it also works if the equalities contain parameters

extract_equalities2(b(conjunct(A,B),pred,Info),Ids,RemIds,Res) --> !,
   {Res=b(conjunct(PA,PB),pred,Info)},
   extract_equalities2(A,   Ids,RemIds1,PA),
   extract_equalities2(B,RemIds1,RemIds,PB).
extract_equalities2(b(equal(ID,E),pred,_Info),Ids,RemIds,NewPred,InSubst,OutSubst) :-
    same_id_member(ID,Ids,SID,RemIds),
    %\+ occurs_in_expr(SID,E), % not cyclic; we check it again after applyng InSubst to E below
    \+ in_subst_dom(SID,InSubst), % ensure we don't already have an equation for the same ID
    compose_with_subst(InSubst,SID,E,OutSubst), % we can compose with existing substitution without occurs-check problem
    !,
    (always_well_defined_or_disprover_mode(E)
      -> NewPred=b(truth,pred,[])
       ; NewPred=b(truth,pred,[]),
        % note: if we keep E=E for deadlock checking we would still not find more deadlock counter examples not(E=E) either fails or if E is not WD it also fails (sometimes generating error message though)
        (debug_mode(on)
           -> print('Ignoring potential WD problem: '), translate:print_bexpr(E),nl
            ; true)
    )
        % NewPred = b(equal(E,E),pred,_Info)),  % this ensures that we evaluate E to detect any WD problems
   .
extract_equalities2(X,Ids,Ids,X) --> [].

% compose a new binding Var/Expr with existing substitution Subst; where Subst was not yet applied to Expr
compose_with_subst(Subst,Var,Expr,[Var/NewExpr|ModSubst]) :-
   apply_subst(Subst,Expr,NewExpr),
   \+ occurs_in_expr(Var,NewExpr), % otherwise cyclic
   apply_binding_to_substitution(Subst,Var,NewExpr,ModSubst).
apply_binding_to_substitution([],_,_,[]).
apply_binding_to_substitution([TV/TE|T],Var,Expr,[TV/TE2|NT]) :-
   \+ occurs_in_expr(TV,Expr), % otherwise we get a cyclic substitution --> fail
   replace_id_by_expr(TE,Var,Expr,TE2),
   apply_binding_to_substitution(T,Var,Expr,NT).

apply_subst([],X,X).
apply_subst([Var/Expr|T],E,NewE) :-
  (replace_id_by_expr(E,Var,Expr,E2)  % replace identifier Var by expression Expr
    -> %apply_to_tail(T,Var,Expr,NT), % apply binding to tail of substitution
       apply_subst(T,E2,NewE)
    ;  add_internal_error('Call failed: ',apply_subst([Var/Expr|T],E,NewE)),NewE=E).


in_subst_dom(X,[I/_|T]) :- (X=I -> true ; in_subst_dom(X,T)).

same_id_member(X,[H|T],SID,Rest) :-
  (same_id(X,H,SID) -> Rest=T ; Rest=[H|TR],same_id_member(X,T,SID,TR)).

% TO DO: investigate what is the exact difference with create_exists_opt
create_and_simplify_exists(Paras,Pred,ResultingPred) :-
   % print(exists(Paras)),nl, translate:print_bexpr(Pred),nl,
   extract_equalities(Paras,Pred,NewParas,NewPred),
   % translate:print_bexpr(NewPred),nl,
   %print(create_and_simplify_exists2(NewParas,NewPred)),nl,
   create_and_simplify_exists2(NewParas,NewPred,ResultingPred).

%:- use_module(b_interpreter_check,[arithmetic_op/4]). % TODO REFACTOR and move this and code below to separate module
create_and_simplify_exists2([],Pred,R) :- !,R=Pred.
% simplify #ID.(ID:E1 & ID:E2)  <=> E1 /\ E2 /= {} , etc... now done in ast_cleanup
% simplify #ID.(ID/:E)  <=> E /= FullType also done in ast_cleanup
% simplify #SID.(SID > Expr) <=> truth now done in ast_cleanup
create_and_simplify_exists2(Parameters,Body,Res) :- create_unsimplified_exists(Parameters,Body,Res).
% TO DO: treat nested exists



:- use_module(b_ast_cleanup,[clean_up_pred/3]).
:- use_module(bsyntaxtree,[is_a_disjunct_or_implication/4,add_texpr_info_if_new/3]).
create_unsimplified_exists(Parameters,Body,Res) :-
     % print(create_unsimplified_exists(Parameters,Body)),nl,
     create_exists(Parameters,Body,Res1),
     add_texpr_info_if_new(Res1,removed_typing,Res2), % unconditionally add this info, avoid generating warnings in exists_body_warning, see test 625
     clean_up_pred(Res2,[],Res). % also computes used_ids,
                 % and performs some existential quantifier optimisations like replace_exists_by_not_empty
% previously removed_typing was only added for disjuncts or implications:
%     (is_a_disjunct_or_implication(Body,_,_,_)
%      -> add_texpr_infos(Res2,[removed_typing],Res3) % avoid generating warnings in exists_body_warning; see, e.g., test 625
%      ; Res3=Res2),
