% a small utility to check that the block declarations in a Prolog
% file are immediately followed by a predicate declaration for a predicate
% of same name and arity

check(N) --> ":-",!, check_directive(N).
check(N) --> ws,!, check(N).
check(N) --> [_X],!,check(N).
check(_N) --> [].

check_directive(N) --> ws,!, check_directive(N).
check_directive(N) --> "block", ws,
   (block_decl(P,Arity,BA)
     -> {name(Name,P) ,vprintln(detected_block(Name,Arity,BA,nr(N))) },!,
         block_decl_rest(P,Arity),
    (predicate_head(P,Arity2)
      -> ({Arity=Arity2} -> {true} ; {print_error(non_matching_block_arity(Name,Arity,BA,Arity2))})
       ; {print_error(non_matching_block_declaration(Name,Arity,BA))} ),
         {N1 is N+1},
         check(N1)
     ;  {print_error(illegal_block(nr(N))),fail}).
check_directive(N) --> check(N).

predicate_head(ID,Arity) --> ws, !, predicate_head(ID,Arity).
predicate_head(ID,Arity) --> cond_compilation_matter, !, predicate_head(ID,Arity).
predicate_head(ID,Arity) --> id(ID), anyws, "(", pred_args(Arity).
pred_args(Arity) --> ",",!, pred_args(A1), {Arity is A1+1}.
pred_args(1) --> ")",!.
pred_args(A) --> ws, !, pred_args(A).
pred_args(A) --> "(", !,term(0), ")", pred_args(A).
pred_args(A) --> "[",anyws,"]", !, pred_args(A).
pred_args(A) --> "[", !,term(0), anyws, "]", pred_args(A).
pred_args(A) --> [_], !, pred_args(A).

% skip over things like :- if(environ(prob_safe_mode,true)).
cond_compilation_matter --> ":-", anyws, "if", anyws, "(", term(0), anyws, ")", anyws, ".".
cond_compilation_matter --> ":-", anyws, "endif", anyws, ".".

term(N) --> ws,!, term(N).
term(NestingL) --> {NestingL>0}, ")",!,{N1 is NestingL-1},term(N1).
term(NestingL) --> {NestingL>0}, "]",!,{N1 is NestingL-1},term(N1).
term(N) --> "(", !, {N1 is N+1}, term(N1).
term(N) --> "[", !, {N1 is N+1}, term(N1).
term(N) --> [X], {nonmember(X,")]")},!,term(N).
term(0) --> "".

block_decl(ID,Arity,BA) --> anyws,id(ID),
                   anyws,"(",block_args(ID,Arity,BA).

block_args(_,1,[BA]) --> block_arg(BA),anyws,")",!.
block_args(ID,Arity,[BA|T]) --> block_arg(BA),anyws,",",!,block_args(ID,A1,T), {Arity is A1+1}.
block_args(ID,A,BA) --> ws, !, block_args(ID,A,BA).
block_args(ID,_,_) --> [X], {name(N,ID),name(NX,[X]),print_error(illegal_block_argument(N,NX)),fail}.

block_arg('-') --> "-".
block_arg('?') --> "?".

block_decl_rest(_P,_Arity) --> ".",!.
block_decl_rest(P,Arity) --> ",",!,
    block_decl(P,Arity,_),block_decl_rest(P,Arity).
block_decl_rest(P,Arity) --> ws,block_decl_rest(P,Arity).
block_decl_rest(P,Arity) --> {print_error(illegal_block_decl_rest(P,Arity)),fail}.

id(ID) --> "'",!,id2(ID),"'".
id([H|T]) --> alpha(H), id2(T).
id2([H|T]) --> alphadigit(H),!,id2(T).
id2([]) --> [].

alpha(X) --> [X],{X>=97, X=<122}.
alphadigit(X) --> [X], ({X>=97,X=<122} ; {X>=65, X=<90} ; {X=95} ; {X>=48, X=<57}).

anyws --> "%",!, go_to_line_end,anyws.
anyws --> "/*",!, go_to_end_comment,anyws.
anyws --> ws,!,anyws.
anyws --> [].

go_to_line_end --> new_line,!.
go_to_line_end --> [_], go_to_line_end.

go_to_end_comment --> new_line,!, go_to_end_comment.
go_to_end_comment --> "*/",!.
go_to_end_comment --> [_], go_to_end_comment.

ws --> "%",!, go_to_line_end.
ws --> "/*",!, go_to_end_comment.
ws --> " "; "\t" ; new_line.

new_line --> "\n",{inc_line_nr}.

:- dynamic cur_line/1.
cur_line(1).
inc_line_nr :- retract(cur_line(N)), N1 is N+1, assert(cur_line(N1)).
reset_line_nr :- retract(cur_line(_)), assert(cur_line(1)).

:- dynamic verbose/0.
vprintln(S) :- (verbose -> print(S),nl ; true).



print_error(Error) :-
    cur_line(LineNr),
    current_output(X),
    set_output(user_error),
    nl,
    write('! Line: '),write_term(LineNr,[]),nl,
    (var(Error)  -> print_message(error,'_')
    ;  write('! '),write_term(Error,[max_depth(20),numbervars(true)]),nl),
    %% flush_output(user_error), %%
    set_output(X).

% ---------------------------

read_string_from_file(Filename,String) :-
    open(Filename,read,S),
    read_string(S,String),
    close(S).

read_string(S,String) :-
    get_code(S,C),!,
    (C= -1
    -> String = []
    ;  String=[C|Rest],read_string(S,Rest)).

% ------------------------------

chk([]) :- !.
chk([H|T]) :- !,chk(H),chk(T).
chk('-v') :- !, assert(verbose).
chk(F) :- read_string_from_file(F,S),
  print(checking(F)),nl,
  reset_line_nr,
  check(1,S,[]).

test :- chk(['kernel_objects.pl','custom_explicit_sets.pl','bsets_clp.pl',
             'custom_explicit_sets.pl', 'bsets_clp.pl',
             'kernel_z.pl','kernel_records.pl','kernel_sym_break_order.pl',
             'b_interpreter.pl', 'kernel_mappings.pl', 'kernel_freetypes.pl',
             'kernel_equality.pl', 'fd_utils_clpfd.pl', 'bool_pred.pl',
             'b_enumerate.pl', 'b_ast_cleanup.pl',
             'b_interpreter_check.pl','store.pl', 'btypechecker.pl',
             'external_functions.pl','ltl.pl']).

go_cli :-
    prolog_flag(argv,ArgV),
    chk(ArgV).

runtime_entry(start) :- go_cli.
