% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(graph_iso_nauty,[print_nauty_graph/1,
                           nauty_init/0, nauty_clear/0,
                           add_nauty_graph/2]).


/* Translate a state graph, as produced by
   - into an format that can be input into dreadnaut
   - insert it directly into the C-Interface to nauty with its internal DB of canoncial forms
 */
 :- use_module(module_information).

:- module_info(group,symmetry).
:- module_info(description,'Translate a state graph into a form for nauty and insert into internal DB of nauty C-interface.').

:- use_module(library(lists)).
:- use_module(library(avl)).
:- use_module(self_check).
:- use_module(error_manager).
:- use_module(tools_meta,[reraise_important_exception/1]).

sorted_list_to_keylist([],_,[]).
sorted_list_to_keylist([H|T],N,[H-N|KT]) :- N1 is N+1, sorted_list_to_keylist(T,N1,KT).
slist_to_avl(List,AVL) :-
   sorted_list_to_keylist(List,1,KL), % assign each element a number, starting at 1
   list_to_avl(KL,AVL).


:- dynamic computed_label_colours/4.

get_label_colours(SG,Cols,AVLCols,AVLRange,TotalNrLabelCols) :-
      (computed_label_colours(Cols,AVLCols,AVLRange,TotalNrLabelCols) -> true
        ; add_error_fail(graph_iso_nauty,'Could not get label colours',SG)).

compute_label_colours :-
      retractall(computed_label_colours(_,_,_,_)),
      b_get_machine_variables(TypedVarList),
      b_get_machine_constants(TypedCstList),
      append(TypedCstList,TypedVarList,TL),
      ~~mnf(user:convert_typed_varlist(TL,Cols,_Types)),
      retractall(cst_var_colournr(_,_)), retractall(colornr_used(_)),
      generate_col_keylist(TL,3,KL,TotalNrLabelCols), % start at 3, because of $from and $to_el
      % TO DO: check if $from and $to_el are really required given types of variables,constants !
      print(label_colours(Cols,KL,TotalNrLabelCols)),nl,
      list_to_avl(['$from'-1,'$to_el'-2|KL],AVLCols), %%print(AVLCols),nl,
      avl_range(AVLCols,AVLRange),
      assert(computed_label_colours(Cols,AVLCols,AVLRange,TotalNrLabelCols)).

 :- dynamic cst_var_colournr/2, colornr_used/1.

generate_col_keylist([],N,[],TotalNrLabelCols) :- TotalNrLabelCols is N-1.
generate_col_keylist([b(identifier(ID),Type,_)|T],N,[ID-IDNr|KT],TotalNrLabelCols) :-
    (find_existing_colour(Type,IDNr) -> N1=N /* we can reuse an existing colour nr without conflict */
     ; IDNr=N,N1 is N+1,
       assert(colornr_used(IDNr)) /* we need to use a new colour number; entailing a new level */
    ), assert(cst_var_colournr(Type,IDNr)),
    generate_col_keylist(T,N1,KT,TotalNrLabelCols).

find_existing_colour(Type,Nr) :- /* check whether an existing colour number can be used for a variable/constant with type Type */
   colornr_used(Nr),
   \+ find_incompatible_type(Type,Nr).
find_incompatible_type(Type,Nr) :- cst_var_colournr(Type2,Nr),
    incompatible_type(Type,Type2).

% TO DO: Check and extend this !!!!!!!!!!
%  two different types may still be problematic if put onto the same level??
%    especially since there are currently only 4 colours for abstract nodes ?!
incompatible_type(T,T).
incompatible_type(T,set(T)).  %% membership arrow could conflict with = arrow; but only at top-level ??
incompatible_type(set(T),T).




get_nodes(SG,Nodes,AVLNodes) :-
   findall(N,(sg_arc(Orig,_C,Dest,SG),(N=Orig;N=Dest)),Ns),
   sort(Ns,Nodes),
   slist_to_avl(Nodes,AVLNodes).
%% sg_arc(Orig,arc,Dest, StateGraph) :-
sg_arc(Orig,ArcColour,Dest, StateGraph) :-
   member((Orig,ArcColour,Dest),StateGraph).


:- dynamic partition/3.
print_nauty_graph(SG) :- nauty_initialised,!,
  get_graph_info(SG,  Cols,AVLRange,NrCols,Nodes,AVLNodes,NrNodesPerLevel,TotalNodes,TSG),
  format('% LabelColors=~w=Levels, Nodes/Level=~w, Cols=~w\n',[NrCols,NrNodesPerLevel,Cols]),
  format('n=~w $=0 g\n',[TotalNodes]),
  print_nodes(AVLRange,Nodes,AVLNodes,NrNodesPerLevel,NrCols,TSG),
  print('$$'),nl,
  %get_node_colours(Nodes,NCols),
  % print(node_colors(NCols)),nl,
  print_partitions.
print_nauty_graph(_). % don't print it; we don't have the necessary info
%get_node_colours(Nodes,NCols) :-
%   findall(fd(NC), member(fd(_,NC),Nodes), NCs), sort(NCs, NCols1),
%   findall(NC2, (member(NC2,Nodes), NC2 \= fd(_,_)), NC2s),
%   sort(NC2s, NCols2),append(NCols1,NCols2,NCols).


print_partitions :-  /* print the starting partion: partition into every level plus in
  each level partition according to the node colours */
  findall(part(Col,Type),partition(_,Col,Type),Ps),
  sort(Ps,Partitions),
  print('% Starting Partitions: '),print(Partitions),nl,
  print('f ['),
  print_part(Partitions),
  nl.


print_part([]).
print_part([part(Col,Type)|T]) :-
   findall(Id,partition(Id,Col,Type),Ids),
   sort(Ids,SIds),
   print_individual_part(SIds),
   (T=[] -> print(']') ; print(' | ')),
   print_part(T).

print_individual_part([]).
print_individual_part([H|T]) :- print(H), print_individual_part2(T).
print_individual_part2([]).
print_individual_part2([H|T]) :- print(', '),print(H), print_individual_part2(T).


print_nodes(AVLRange,Nodes,AVLNodes,NrNodesPerLevel,NrCols,TSG) :-
   get_node(AVLRange,Nodes,AVLNodes,NrNodesPerLevel,NrCols,TSG,  NrC,Node,ID,Successors),
   print_node(Node, NrC, ID, Successors),
   fail.
print_nodes(_AVLRange,_Nodes,_AVLNodes,_NrNodes,_NrCols,_TSG).


translate_state_graph([],_AVLCols,_AVLNodes,[]).
translate_state_graph([(Node,Col,Dest)|T],AVLCols,AVLNodes,[(NrN,NrC,NrD)|TT]) :-
  avl_member(Node,AVLNodes,NrN),
  (avl_member(Col,AVLCols,NrC) -> true
    ; add_error(translate_state_graph,'Unknown Label: ',Col:AVLCols),NrC=1),
  avl_member(Dest,AVLNodes,NrD),
  translate_state_graph(T,AVLCols,AVLNodes,TT).

 /* get general info about a state graph : */
get_graph_info(StateGraph,Cols,AVLRange,TotalNrLabelCols,
               Nodes,AVLNodes,NrNodesPerLevel,TotalNodes,TranslatedStateGraph) :-
  get_label_colours(StateGraph,Cols,AVLCols,AVLRange,TotalNrLabelCols),
  get_nodes(StateGraph,Nodes,AVLNodes),
  length(Nodes,NrNodesPerLevel),
  TotalNodes is NrNodesPerLevel*TotalNrLabelCols,
  translate_state_graph(StateGraph,AVLCols,AVLNodes,TranslatedStateGraph),!.
get_graph_info(StateGraph,_,_,_,_,_,_,_,_) :-
               add_error(graph_iso_nauty,'get_graph_info failed: ',StateGraph),fail.

 /* get a node, its colour, its ID and its successors : */


get_node(AVLRange,_Nodes,AVLNodes,NrNodesPerLevel,TotalNrLabelCols,TSG,
         NrofLabelColour,Node,NodeID,Successors) :-
   avl_member(Node,AVLNodes,NrN),
   restrict_sg_for_node(TSG,NrN,RestrictedTSG),
   member(NrofLabelColour,AVLRange),
  % print(nr(NrN,NrofLabelColour)),nl,

   NodeID is (NrofLabelColour-1)*NrNodesPerLevel+NrN-1,
   (NrofLabelColour>1
    -> N1 is (NrofLabelColour-2)*NrNodesPerLevel+NrN-1, Successors = [N1|S2]
       /* add link to layer above */
    ; Successors=S2),

%    S2=S3,  /* add no link to layer above; seems to be much slower ! */
   (NrofLabelColour<TotalNrLabelCols
   -> N2 is (NrofLabelColour)*NrNodesPerLevel+NrN-1, S2 = [N2|S3]
       /* add link to layer above */
    ; S2=S3),

   findall( DestID,  (member((NrN,NrofLabelColour,DestNr),RestrictedTSG),
                      DestID is (NrofLabelColour-1)*NrNodesPerLevel+DestNr-1), S3).

% translate a node into the corresponding id for the label colour, i.e. level
% has been inlined above for performance reasons
%get_node_id(NrOfNode,NrOfLabelColour, NrNodesPerLevel,  NodeID) :-
%    NodeID is (NrOfLabelColour-1)*NrNodesPerLevel+NrOfNode-1.

% ----------------------------------------------------------

restrict_sg_for_node([],_,[]).
restrict_sg_for_node([(Nr,Col,Dest)|T],NodeNr,Res) :-
  (Nr = NodeNr -> Res = [(Nr,Col,Dest)|TR]  ; Res=TR),
  restrict_sg_for_node(T,NodeNr,TR).

:- assert_must_succeed(( graph_iso_nauty:get_subgraph([(1,2,3),(1,2,5),(1,4,1),(1,4,2),(2,1,1),(2,1,2)],Sub,N,C),  N==2, C==1, Sub == [(2,1,2),(2,1,1)] )).

get_subgraph(StateGraph,SubGraph,SubNode,SubCol) :-
   get_subgraph(StateGraph, [],_,_, SubGraph,SubNode,SubCol).

get_subgraph([],SubGraph,Node,Col,SubGraph,Node,Col).
get_subgraph([(Src,Col,Dest)|T], CurSubGraph,CurNode,CurCol,  SubGraph,SubNode,SubCol) :-
   ((CurNode=Src, CurCol = Col)
     -> get_subgraph(T,[(Src,Col,Dest)|CurSubGraph],CurNode,CurCol,SubGraph,SubNode,SubCol)
     ;  ( (SubGraph = CurSubGraph, SubNode=CurNode, SubCol=CurCol)
          ;
          get_subgraph(T, [(Src,Col,Dest)],Src,Col, SubGraph,SubNode,SubCol)
        )
   ).

% ----------------------------------------------------------

print_node(Node, NrC, ID, Successors) :-
   (Node = fd(_,Type) -> assert(partition(ID,NrC,fd(Type))) ; assert(partition(ID,NrC,Node))),
   print(' '), print(ID), print(' : '),
   member(DestID,Successors),
   print(DestID), print(' '),
   fail.
print_node(_Node, _NrC, _ID, _Successors) :- print(';'),nl.


/* ------------------------------------------------------- */


/* INTERFACE FOR ADDING GRAPHS TO OUR C-Library for NAUTY */
:- dynamic nauty_initialised/0.

%nauty_use_module
:- use_module(graphiso,'../extensions/nauty/graphiso.pl',[graphiso_init/0,
          prob_start_graph/1,
          prob_add_edge/2,
          prob_get_number_of_colours/1,
          prob_set_colour_of_node/2,
          prob_exists_graph/1,
          prob_free_storage/0]).

nauty_init :-  %%print(nauty_init),nl,
               initialise_node_colours,
               % nauty_use_module,
               (nauty_initialised -> print(prob_free_storage),nl, prob_free_storage
                    ; assert(nauty_initialised) ),
               %% print(graphiso_init),nl,
               on_exception(Exception, graphiso:graphiso_init,
                  (add_error(nauty_init,'Exception during graphiso_init: ',Exception),
                   reraise_important_exception(Exception),fail)),
               nr_of_node_colours_per_level(NrOfNodeColours),
               max_nr_of_label_colours(MaxNrOfLabelColours),
               MaxNrOfNodeColours is NrOfNodeColours * MaxNrOfLabelColours,
     %% print(prob_get_number_of_colours(MaxNrOfNodeColours,NrOfNodeColours)),nl, flush_output(user), %%
               nauty_clear,
               prob_get_number_of_colours(MaxNrOfNodeColours),
               assert(total_nr_of_node_colours(MaxNrOfNodeColours)),
               print(done),nl,flush_output(user),
               compute_label_colours.
nauty_clear :-
    retractall(colour_overflow),
    retractall(total_nr_of_node_colours(_)),
    retractall(computed_label_colours(_,_,_,_)).


:- use_module(bmachine,[b_get_machine_constants/1,b_get_machine_variables/1]).

max_nr_of_label_colours(Nr) :-
   b_get_machine_constants(C),
   length(C,NrOfConstants),
   b_get_machine_variables(V),
   length(V,NrOfVars),
   NrOfExtraArrows = 2, /* '$from', '$to_el'*/
   Nr is NrOfConstants+NrOfVars + NrOfExtraArrows.

%nauty_reset :- prob_free_storage.

:- dynamic nr_of_node_colours_per_level/1.
nr_of_node_colours_per_level(5).

:- dynamic total_nr_of_node_colours/1.
total_nr_of_node_colours(5).

:- dynamic node_fd_colour/3.
:- dynamic colour_for_int0/1.
:- volatile colour_for_int0/1.

:- use_module(b_global_sets,[b_global_deferred_set/1, is_used_b_global_constant/3,is_unused_b_global_constant/3,
  b_partially_deferred_set/1]).
:- use_module(preferences,[get_preference/2]).

initialise_node_colours :- retractall(nr_of_node_colours_per_level(_)), retractall(node_fd_colour(_,_,_)),
   assert(nr_of_node_colours_per_level(8)), % start off with number 8 ( last abs_cn colour is 7)
   is_used_b_global_constant(Type,Nr,_Cst),
   retract(nr_of_node_colours_per_level(NC)),
   NC1 is NC+1,
   assert(nr_of_node_colours_per_level(NC1)),
   assert(node_fd_colour(Type,Nr,NC)),
  %% print(enum_colour(Type,Nr,NC,_Cst)),nl, %%
   fail.
initialise_node_colours :-
   b_partially_deferred_set(Type),
   retract(nr_of_node_colours_per_level(NC)),
   NC1 is NC+1,
   assert(nr_of_node_colours_per_level(NC1)),
   is_unused_b_global_constant(Type,Nr,_),
   assert(node_fd_colour(Type,Nr,NC)),
  %% print(partial_enum_colour(Type,Nr,NC)),nl, %%
   fail.
initialise_node_colours :-
   b_global_deferred_set(Type),
   retract(nr_of_node_colours_per_level(NC)),
   NC1 is NC+1,
   assert(nr_of_node_colours_per_level(NC1)),
   assert(node_fd_colour(Type,_,NC)),
  %% print(node_fd_colour(Type,'ANY',NC)),nl, %%
   fail.
initialise_node_colours :- nr_of_node_colours_per_level(C),
    preferences:get_preference(minint,MININT),
    preferences:get_preference(maxint,MAXINT),
    CNew is C + MAXINT+1-MININT,
    CNewIncreased is CNew + 14,  /* to allow numbers which are greater than MAXINT or smaller than MININT */
    retract(nr_of_node_colours_per_level(_)),
    assert(nr_of_node_colours_per_level(CNewIncreased)),
    retractall(colour_for_int0(_)),
    assert(colour_for_int0(C)),
    print('Colours initialised: '), print(CNewIncreased),nl, flush_output(user).

get_colour_of_node(X,C) :- var(X), !,
   add_error(get_colour_of_node,'Argument is var: ',get_colour_of_node(X,C)),
   fail.
get_colour_of_node(sg_root,C) :- !, C=0.
%get_colour_of_node('$Empty_Set',C) :- !, C=1.
get_colour_of_node([],C) :- !, C=1.
get_colour_of_node(pred_true /* bool_true */,Colour) :- !, Colour = 3.
get_colour_of_node(pred_false /* bool_false */,Colour) :- !, Colour = 2.
get_colour_of_node(abs_cn(_,AbsCol),Colour) :- !, Colour is 4 + (AbsCol mod 4).  %% TO DO: make more flexible
get_colour_of_node(fd(Nr,Type),Colour) :- !, node_fd_colour(Type,Nr,Colour).
get_colour_of_node(int(Nr),Colour) :- !, colour_for_int0(Max),   /* + number of cn's !!! */
  (Nr>=0 -> Colour is Max+Nr*2
         ; Colour is Max-1-(Nr*2)
  ).
  /* get_colour_of_node(global_set(_)) :-  ADD ... */
get_colour_of_node(string(S),C) :- !, print(strings_not_supported_in_nauty(S)),nl, C=0.
get_colour_of_node((_A,_B),Col) :- !, Col = 4. /* internal node; just like abs_cn(_,_) */
get_colour_of_node(avl_set(A),Col) :- !, print(avl_set_as_node(A)),nl, Col = 1.  /* should be handled in state graph ! */
get_colour_of_node(Node,0) :- print(unknown_node(Node)),nl.


/* add SG as nauty graph to the nauty C library and check if it already exists */
add_nauty_graph([],Exists) :- !,
  prob_start_graph(1), prob_set_colour_of_node(0,0),
  prob_exists_graph(Exists).
add_nauty_graph(SG,Exists) :-
  get_graph_info(SG,  _Cols,AVLRange, NrLabelCols,
                      Nodes,AVLNodes,NrNodesPerLevel,TotalNodes,TSG),
  %% format('% LabelColors=~w=Levels, Nodes/Level=~w\n',[NrLabelCols,NrNodesPerLevel]), %%
              % nr_of_node_colours_per_level(NrOfNodeColours),
              % print(prob_get_number_of_colours(NrOfNodeColours)),nl,
              % prob_get_number_of_colours(NrOfNodeColours),
 %% nl,nl,print(prob_start_graph(TotalNodes)),nl, %%
  prob_start_graph(TotalNodes),
  add_nodes(AVLRange,Nodes,AVLNodes,NrNodesPerLevel,NrLabelCols,TSG),
  %print(prob_exists_graph(Exists)),nl,
  prob_exists_graph(Exists)
  .%,print(added_graph(SG,Exists)),nl,nl.



:- dynamic colour_overflow/0.

add_nodes(AVLRange,Nodes,AVLNodes,NrNodesPerLevel,TotalNrLabelCols,TSG) :-
   %% print(add_nodes(AVLRange,Nodes,AVLNodes,NrNodesPerLevel,TotalNrLabelCols,TSG)),nl,
   %% portray_avl(AVLNodes),nl,
   total_nr_of_node_colours(MaxNrOfNodeColours),
   get_node(AVLRange,Nodes,AVLNodes,NrNodesPerLevel,
            TotalNrLabelCols,TSG,  NrOfLabelColour,Node,ID,Successors),
   (get_colour_of_node(Node,Colour) -> true
      ; add_error(add_nodes,'Call failed: ',get_colour_of_node(Node,Colour)),Colour=0),
 %% print(node(Node,Colour)),nl, %%
   RealColour is (Colour*TotalNrLabelCols + NrOfLabelColour-1),
   (RealColour >= MaxNrOfNodeColours
     -> (RealColour2 is RealColour mod MaxNrOfNodeColours,
         (colour_overflow -> true
           ; (assert(colour_overflow),
              add_error(graph_iso_nauty,'Warning: Colour overflow in Nauty Symmetry Library. Some symmetry classes may not be explored!!! Try and increase MAXINT.',RealColour))
        ))
     ;  RealColour2 = RealColour
   ),  %print(prob_set_colour_of_node(ID,RealColour,nodec(Colour),level(NrOfLabelColour),nroflabelcolours(TotalNrLabelCols))),nl, flush_output(user),%%
   prob_set_colour_of_node(ID,RealColour2),
   member(DestID,Successors),
   %% print(prob_add_edge(ID,DestID)),nl, flush_output(user),%%
   prob_add_edge(ID,DestID),
   fail.
add_nodes(_AVLRange,_Nodes,_AVLNodes,_NrNodes,_TotalNrLabelCols,_TSG).


/* ------------------------------------------------------- */
/* TESTS */

/*
state_graph([
 (fd(2,'Code'),activec,sg_root),
 (fd(1,'Code'),activec,sg_root),
 (fd(2,'Name'),active,sg_root),
 (fd(1,'Name'),active,sg_root),
 (fd(2,'Name'),db,fd(1,'Code')),
 (fd(1,'Name'),db,fd(2,'Code'))
 ]).


tn :- state_graph(SG), print_nauty_graph(SG).
tn2 :- state_graph(SG), nauty_init, add_nauty_graph(SG,Ex), print(exists(Ex)),nl,
      add_nauty_graph(SG,Ex2), print(exists(Ex2)),nl.

% graph_iso_nauty:bench. For Phonebook7
 | ?- graph_iso_nauty:bench.
bench_state_graph
time_state_graph(0)
time_nauty_wo_exists(2390)
time_nauty(2670)
sg_overhead_ratio(0.0)


%:- assert_must_succeed( (graph_iso_nauty:state_graph(SG), graph_iso_nauty:nauty_init,
%                         graph_iso_nauty:add_nauty_graph(SG,Ex), Ex== -1,
%                         graph_iso_nauty:add_nauty_graph(SG,Ex2), Ex2==1, graph_iso_nauty:nauty_init)).



bench :- bench(10000).
bench(N) :- print(bench_state_graph),nl,
         statistics(runtime,[_,_]),
         sg(N,SG),
         statistics(runtime,[_,T1]),
         print(time_state_graph(T1)),nl,
         statistics(runtime,[_,_]),
         adw(N,SG),
         statistics(runtime,[_,T2]),
         print(time_nauty_wo_exists(T2)),nl,
         statistics(runtime,[_,_]),
         ad(N,SG),
         statistics(runtime,[_,T3]),
         print(time_nauty(T3)),nl,
         Ratio is T1/T3, print(sg_overhead_ratio(Ratio)),nl.

sg(1,SG) :- !,state_graph(SG).
sg(N,SG) :- state_graph(_), N1 is N-1, sg(N1,SG).

adw(1,SG) :- !,add_nauty_graph(SG,_).
adw(N,SG) :- add_nauty_graph_wo_check(SG), N1 is N-1, adw(N1,SG).
ad(1,SG) :- !,add_nauty_graph(SG,_).
ad(N,SG) :- add_nauty_graph(SG,_), N1 is N-1, ad(N1,SG).



 add_nauty_graph_wo_check(SG) :-
  get_graph_info(SG,  _Cols,AVLRange,
                       TotalNrLabelCols,Nodes,AVLNodes,NrNodesPerLevel,TotalNodes,TSG),!,
 % print(get_graph_info(SG,  Cols,AVLCols,TotalNrLabelCols,Nodes,NrNodesPerLevel,TotalNodes)),nl,
  prob_start_graph(TotalNodes),
  add_nodes(AVLRange,Nodes,AVLNodes,NrNodesPerLevel,TotalNrLabelCols,TSG).


bench3 :- prob_free_storage, graphiso_init, prob_get_number_of_colours(16),
         statistics(runtime,[_,_]),
         addg(X1),
         addg(X2),
         statistics(runtime,[_,T1]),
         print(time(T1,X1,X2)),nl.

addg(X) :-
   prob_start_graph(15),
prob_set_colour_of_node(8,6),
prob_add_edge(8,3),
prob_add_edge(8,13),
prob_add_edge(8,5),

prob_set_colour_of_node(6,6),
prob_add_edge(6,1),
prob_add_edge(6,11),
prob_add_edge(6,5),

prob_set_colour_of_node(4,5),
prob_add_edge(4,9),
prob_add_edge(4,0),

prob_set_colour_of_node(2,5),
prob_add_edge(2,7),
prob_add_edge(2,0),

prob_set_colour_of_node(14,5),
prob_add_edge(14,9),
prob_add_edge(14,13),

prob_set_colour_of_node(12,5),
prob_add_edge(12,7),
prob_add_edge(12,11),

prob_set_colour_of_node(0,0),
prob_set_colour_of_node(1,0),
prob_set_colour_of_node(3,0),
prob_set_colour_of_node(5,0),
prob_set_colour_of_node(7,0),
prob_set_colour_of_node(9,0),
prob_set_colour_of_node(10,0),
prob_set_colour_of_node(11,0),
prob_set_colour_of_node(13,0),

prob_exists_graph(X).

             */




