% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(kernel_equality,[equality_objects_wf/4, equality_objects_lwf/4, equality_objects/3,
           equality_objects_with_type/4,
           empty_set_test/2, empty_cartesian_product/3,
           or_eq_obj/3,
           get_cardinality_wait_flag/4, get_cardinality_wait_flag/5,
           get_cardinality_powset_wait_flag/5,
           get_cardinality_relation_over_wait_flag/6,
           %get_cardinality_partial_function_wait_flag/7,
           check_and_remove/4,
           membership_test/3, membership_test_wf/4,
           membership_test_wf_with_force/4,
           cartesian_pair_test_wf/5,
           in_nat_range_test/4,

           dif/3, integer_dif/2, bool_dif/2, eq_atomic/4,

           subset_test/4, subset_strict_test/4,
           conjoin_test/4]).

:- use_module(kernel_objects,[equal_object/2, equal_object/3,
                              not_equal_object/2, cardinality_as_int_for_wf/2,
                              safe_mul/3, safe_add/3, safe_pow2/2, safe_pown/3,
                              empty_set/1, not_empty_set/1,
                              exhaustive_kernel_check/2, exhaustive_kernel_check/1,
                              exhaustive_kernel_check_wf/2, exhaustive_kernel_fail_check_wf/2,
                              exhaustive_kernel_fail_check/1]).
:- use_module(b_interpreter_check,[conjoin/6]).
:- use_module(kernel_dif).

:- use_module(self_check).


:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module provides reified equality and membership predicates.').

:- use_module(error_manager).

% kernel_equality
% :- ensure_loaded(kernel_equality).

:- use_module(custom_explicit_sets).
:- use_module(kernel_waitflags).

:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([int(2),int(1)],[int(2)],pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([int(2),int(1)],[],pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([int(2),int(1)],[int(1),int(2)],pred_true))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([[int(2)],[int(1)]],[[int(1)],[int(2)]],pred_true))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([string('STRING1')],[string('STRING2')],pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([string('STRING2'),string('STRING1')],
                       [string('STRING1'),string('STRING2')],pred_true))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([(int(1),int(2)),(int(1),int(3))],[(int(1),int(3)),(int(1),int(2))],pred_true))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([(int(2),int(2)),(int(1),int(3))],[(int(2),int(3)),(int(1),int(2))],pred_false))).
:- assert_must_succeed((equality_objects(A,B,R), A=[int(2),int(3)],B=[int(X),int(Y)],
                     X=4,Y=2, R==pred_false)).
:- assert_must_succeed((equality_objects(A,B,R), A=[int(2),int(3)],B=[int(X),int(Y)],
                     X=2,Y=4, R==pred_false)).
:- assert_must_succeed((equality_objects(A,B,R), A=[int(2),int(3)],B=[int(X),int(Y)],
                     X=3,Y=2, R==pred_true)).
:- assert_must_succeed((equality_objects(A,B,R), A=[int(2),int(3)],B=[int(X),int(Y)],
                     X=2,Y=3, R==pred_true)).
:- assert_must_succeed((equality_objects(A,B,R), A=int(2),B=int(X),
                     R=pred_true, X==2)).


/* equality_objects/3 is a function which compares two objects
   and returns either pred_true or pred_false in the third argument,
   as soon as it can be decided whether the objects are equal */


equality_objects_wf(X,Y,Res,WF) :-
  equality_objects_lwf(X,Y,Res,LWF),
  ((var(Res),WF\==no_wf_available) -> get_last_wait_flag(equality_objects_wf,WF,LWF) ; true).

equality_objects_lwf(X,Y,Res,_LWF) :- X==Y,!, % NOTE: what if X,Y are not well-defined (e.g., division by 0); setting Res=pred_true could lead to failure and missing a WD-problem
   Res=pred_true.
equality_objects_lwf(X,Y,Res,LWF) :-
   equality_objects(X,Y,Res),
   force_equality_objects_lwf2(X,Y,Res,LWF).

:- block force_equality_objects_lwf2(?,?,-,-).
force_equality_objects_lwf2(_X,_Y,Res,_LWF) :-
   (  nonvar(Res) -> true
    ; (  %% print(forcing_equal(X,Y,_WF)),nl, %%
      % alternative would be to propagate LWF down to and_equality ?
         %Res \== pred_false, equal_object(X,Y,force_equality_objects_lwf2), % not required anymore as and_equality can propagate pred_true down  (commented out 27.2.2015)
         Res=pred_true
       ; %Res \== pred_true, not_equal_object(X,Y), % note required anymore (commented out 27.2.2015)
         Res=pred_false
      )).


:- block equality_objects_with_expansion(-,-,?), equality_objects_with_expansion(?,-,-), equality_objects_with_expansion(-,?,-).
% will expand second argument; use if first argument is a tail of a list (representing a set)
equality_objects_with_expansion(X,Y,R) :-
    nonvar(Y), is_custom_explicit_set_nonvar(Y),!,
    expand_custom_set(Y,ExpSet),
    equality_objects(X,ExpSet,R).
equality_objects_with_expansion(X,Y,R) :- equality_objects(X,Y,R).


:- use_module(bool_pred).
% a version of equality_objects with type information available: allows to dispatch to boolean version
equality_objects_with_type(boolean,X,Y,R) :- !, bool_pred:bool_equality(X,Y,R).
equality_objects_with_type(_,X,Y,R) :- (X==Y -> R=pred_true ; equality_objectsb(X,Y,R)).

:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([int(1)],[],pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([int(1)],[int(2)],pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects([int(1),int(2)],[int(2),int(1)],pred_true))).
:- assert_must_succeed((kernel_equality:equality_objects([int(1)],[],R),R==pred_false)).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects(freeval(id1,case1,int(1)),freeval(id1,case1,int(1)),pred_true))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects(freeval(id1,case1,int(1)),freeval(id1,case1,int(2)),pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:equality_objects(freeval(id1,case1,int(1)),freeval(id1,case2,int(1)),pred_false))).

:- use_module(b_global_sets,[is_global_set_of_cardinality_one/2]).
equality_objects(X,Y,R) :- (X==Y -> R=pred_true ; equality_objectsb(X,Y,R)).
% This is in principle more precise: but entails about a 10 % penalty for make cruise
%equality_objects(X,Y,R) :- when((?=(X,Y);nonvar(X);nonvar(Y)),  (X==Y -> R=pred_true  ; equality_objectsb(X,Y,R))).
:- block equality_objectsb(-,-,-).
equality_objectsb(X,Y,R) :-
   (nonvar(X) -> X=fd(XN,GS) ; nonvar(Y) -> Y=fd(XN,GS)),
   (var(GS) -> print(var_gs(fd(XN,GS))),nl,fail ; true),
   is_global_set_of_cardinality_one(GS,N),!, % there exist only one object of that type
   XN=N, Y=X, R=pred_true.
%equality_objectsb(X,Y,R) :- X==Y,!,R=pred_true. % only makes sense if all three are variables; cannot happen here; we could have int(X)==int(Y), but then code further below will trigger
equality_objectsb(X,Y,R) :- equality_objects0(X,Y,R).

%:- block %equality_objects0(-,-,?),
%         equality_objects0(?,-,-), equality_objects0(-,?,-).
equality_objects0(X,Y,R) :- nonvar(R),!, % we know the result of the comparison
  ( R=pred_true  -> equal_object(X,Y,equality_objects0)
  ; R=pred_false -> not_equal_object(X,Y)
  ; add_error_fail(equality_objects0,'Illegal result: ',equality_objects0(X,Y,R))
  ).
% now we know that either X or Y must be nonvar
equality_objects0(X,Y,R) :-
   (nonvar(Y)
     -> (X==[] -> eq_empty_set(Y,R)
          ;  equality_objects1(Y,X,R)
        )
     ;  equality_objects1(X,Y,R)
   ).

 :- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(\+ environ(disable_chr, true)).
:- use_module(chrsrc(chr_integer_inequality),[chr_eq/3]).
:- endif.
:- use_module(clpfd_interface,[post_constraint2/2]).
equality_int(X,Y,Res) :- nonvar(X),nonvar(Y),!,
  (X=Y -> Res=pred_true ; Res=pred_false).
equality_int(X,Y,Res) :-
   % check for attached dif co-routines; diabled if CHR is used (no dif coroutine needed)
   % !! there might still be dif coros set up by equality_epr and dis
   % preferences:preference(use_chr_solver,false),
   fd_frozen_dif(X,Y),!, Res=pred_false.
equality_int(X,Y,Res) :-
   (preferences:preference(use_chr_solver,true) -> chr_eq(X,Y,Res) ; true),
   % with CHR we detect e.g. r=TRUE in (X=Y <=> r=TRUE) & X:1..100 & Y > X
   % see code for equality_fd
   clpfd_interface:post_constraint2( ((X#=Y) #<=> R01), Posted),
   (Posted = true
      -> prop_eq_01(Res,R01),
         (var(R01) -> syntactic_equality_test(X,Y,Res) ; true)  % CLPFD does not detect variable equality
       ; eq_atomic_simple(X,Y,integer,Res)).


%:- block equality_objects1(?,-,-). % avoid instantiating arg2 if result not known, problem for test 1488 : R:INTEGER<->INTEGER & dom(R)=D & X /: D & dom({(X,Y),B}) = DS & D=DS & Y:INTEGER
% equality_objects1 assumes first argument is nonvar:
%%equality_objects1(X,Y,R) :- print(equality_objects1(X,Y,R)),nl,fail. %%
equality_objects1(pred_true /* bool_true */,X,Res) :- !, bool_equality(pred_true /* bool_true */,X,Res). %bool_eq_atomic(X,bool_true,pred_false /* bool_false */,Res).
equality_objects1(pred_false /* bool_false */,X,Res) :- !,bool_equality(pred_false /* bool_false */,X,Res). %bool_eq_atomic(X,bool_false,bool_true,Res).
equality_objects1(int(X),IY,Res) :- !, IY=int(Y), equality_int(X,Y,Res).
equality_objects1(term(X),TY,Res) :- !, TY=term(Y), eq_atomic(X,Y,term,Res).
equality_objects1(string(X),SY,Res) :- !, SY=string(Y), eq_atomic(X,Y,string,Res).
equality_objects1(fd(X,Type),FY,Res) :- !,  FY=fd(Y,Type), %print(equality_fd(X,Y,Type,Res)),nl,
   equality_fd_objects(X,Y,Type,Res).
equality_objects1(freeval(Id,Case1,Value1),FY,Res) :- !, FY=freeval(Id,Case2,Value2),
    and_equality(CaseRes,ValueRes,Res),
    eq_atomic(Case1,Case2,freeval_case,CaseRes),
    equality_freeval_content(CaseRes,Value1,Value2,ValueRes).
equality_objects1(rec(X),RY,Res) :- !, RY=rec(Y), equality_fields(X,Y,Res).
equality_objects1((X,Y),XY2,Res) :- !, XY2=(X2,Y2),
     and_equality(RY,RX,Res),
     equality_objects(X,X2,RX),
     (RX==pred_false -> true
       ; equality_objects(Y,Y2,RY)).
equality_objects1([],S,R) :- !,eq_empty_set(S,R).
%equality_objects1(X,[],R) :- !,eq_empty_set(X,R).
equality_objects1(X,Y,R) :- equality_objects2(X,Y,R).

:- block equality_freeval_content(-,?,?,?).
% if the case is different, we don't care anymore about the content; and the types could be different
equality_freeval_content(pred_false,_Value1,_Value2,_).
equality_freeval_content(pred_true,Value1,Value2,Res) :-
    % if we do not wait until we now that the case is the same,
    % an internal type error might be raised
    equality_objects(Value1,Value2,Res).


:- use_module(error_manager,[add_warning/2]).
:- use_module(tools_printing,[print_term_summary/1]).
:- block equality_objects2(?,-,-). % also wait on second or third argument
equality_objects2(X,Y,Res) :- nonvar(Res),!,% we know the result of the comparison
  ( Res=pred_true  -> %% print(forcing_equal(X,Y)),nl,%%
                 equal_object(X,Y,equality_objects2_1)
  ; Res=pred_false -> not_equal_object(X,Y)
  ; add_error_fail(equality_objects2,'Illegal result: ',equality_objects2(X,Y,Res))
  ).
equality_objects2([H|T],S,Res) :- !,eq_cons(S,H,T,Res). % move to equality_objects1 ?!
equality_objects2(avl_set(A),S,Res) :-
 (S=[H|T] -> !, eq_cons(avl_set(A),H,T,Res)
  ; S = [] -> !, (A=empty -> add_warning(equality_objects2,'Empty avl_set'), Res=pred_true
                           ; Res=pred_false)).
equality_objects2(ES1,ES2,Res) :- equality_explicit_sets(ES1,ES2,ERes), !, Res=ERes.
equality_objects2(ES1,S,Res) :- is_custom_explicit_set(ES1,eq_cons),!, % move to custom_explicit_sets?
    % S = [] or ES1 = [] is already covered above
    (S=[H|T] -> eq_cons(ES1,H,T,Res) % maybe we can avoid expanding ES1
     ; (is_infinite_explicit_set(ES1),is_custom_explicit_set(S,eq_cons))
        -> expand_custom_set(S,ExpSet), equality_objects0(ExpSet,ES1,Res)
     ; expand_custom_set(ES1,ExpSet), equality_objects0(ExpSet,S,Res)).
equality_objects2(X,Y,Res) :-
    nl,print_term_summary(unknown_equality_objects(X,Y,Res)),nl, % could happen with abort(.) terms
    trace,
    (equal_object(X,Y,equality_objects2_2),Res=pred_true
    ;
     not_equal_object(X,Y),Res=pred_false
    ).

:- use_module(b_global_sets,[b_fd_type/3]).
:- use_module(library(clpfd)).
equality_fd_objects(X,Y,_,Res) :- number(X),number(Y),!,
  (X=Y -> Res=pred_true ; Res= pred_false).
equality_fd_objects(X,Y,_,Res) :- % check for attached dif co-routines; later we should probably use CHR
   fd_frozen_dif(X,Y),!,
   % THIS DROPS THE PERFORMANCE OF probsli ../prob_examples/examples/B/ClearSy/alloc_large.mch -init -p TIME_OUT 5000
   Res=pred_false.
% Is it really necessary to keep code below: clp(FD) will do this for us ?!
equality_fd_objects(X,Y,Type,Res) :-
    nonvar(Type),b_global_sets:b_get_fd_type_bounds(Type,LowBnd,UpBnd),!,
    (LowBnd=UpBnd
      -> X=Y,Res=pred_true /* only one object of Type exists */
      ;  (LowBnd+1 =:= UpBnd
          -> bin_eq_atomic(X,Y,Res,LowBnd,UpBnd) /* two objects exist */
          ;  eq_atomic_simple(X,Y,global(Type),Res),
             ((var(Res),(var(X);var(Y)))
             ->  X in LowBnd..UpBnd, Y in LowBnd..UpBnd, % still required ! even with fd_utils_clpfd ! (for v8 of Bosch CrCtl_Comb2_Final_mch.eventb)
                 equality_fd(X,Y,Res)
             ;  true
             )
         )
    ).
equality_fd_objects(X,Y,Type,Res) :-
    add_internal_error('Unknown type: ',equality_fd_objects(X,Y,Type,Res)),
    eq_atomic_simple(X,Y,global(Type),Res).


% a reified version of equality
%equality_fd(X,Y,Res) :- X==Y,!, Res=pred_true. % not necessary; syntactic_equality_test below will trigger
equality_fd(X,Y,Res) :-
  ((X#=Y) #<=> Reification01),
  %  (preferences:preference(use_chr_solver,true) -> chr_eq(X,Y,Res) ; true),  Probably not useful as we do not keep track of info for FD variables; Sebastian: please investigate
  prop_eq_01(Res,Reification01),
  (var(Reification01)
   -> syntactic_equality_test(X,Y,Res)
    ; true).


% propagate pred_false/true <-> 0/1
:- block prop_eq_01(-,-).
% prop_eq_01(A,B) :- print(prop(A,B)),nl,fail.
prop_eq_01(P,B) :- var(P), !, prop_01_eq(B,P).
prop_eq_01(pred_true,1).
prop_eq_01(pred_false,0).
prop_01_eq(1,pred_true).
prop_01_eq(0,pred_false).


:- block eq_empty_set(-,-).
eq_empty_set(X,R) :- var(X),!,
  (R=pred_true -> empty_set(X) ; not_empty_set(X)).
eq_empty_set([],R) :- !,R=pred_true.
eq_empty_set([_|_],R) :- !,R=pred_false.
eq_empty_set(CS,R) :- is_custom_explicit_set(CS,eq_empty_set),!,
   test_empty_explicit_set(CS,R).
eq_empty_set(S,R) :- add_internal_error('Illegal Set: ',eq_empty_set(S,R)),fail.

%%eq_cons(S,H,T,Res) :- print(eq_cons(S,H,T,Res)),nl,fail.
eq_cons([],_,_,Res) :- !, Res=pred_false.
eq_cons([H2|T2],H1,T1,Res) :- !,
   check_and_remove([H2|T2],H1,NewSet2,RemoveSuccesful),
    %% print(removed(H1,H2,T2,NewSet2,RemoveSuccesful)), %%
   eq_cons2(RemoveSuccesful,H1,T1,H2,T2,NewSet2,Res).
eq_cons(avl_set(A),H1,T1,Res) :- !,
   (var(Res) -> kernel_tools:ground_value_check(H1,GrH1) ; true), % if Res already bound, no need to do ground check
   eq_cons_avl_set(A,H1,T1,Res,GrH1).
eq_cons(global_set(G),_H,T,Res) :-  %print(chk(G,H,T)),nl,
   is_infinite_global_set(G),!,
   expand_custom_set_to_list(T,_ET,_Done,eq_cons), % this will either succeed and we have a finite list ; or it will raise an enumeration warning exception
   Res = pred_false.
eq_cons(ES,H1,T1,Res) :- is_custom_explicit_set(ES,eq_cons),expand_custom_set(ES,ExpSet),
   eq_cons1(ExpSet,H1,T1,Res).

% this treatment is relevant for avoiding unnecessary choice points
% see e.g. codespeed test probsli ../prob_examples/examples/B/Alstom/vesg_Aug11/gradient_train_altitude.mch -animate 5 -p TIME_OUT 300000000 -p MAXINT 2147483647 -p MININT -2147483647 -p CLPFD TRUE
:- block eq_cons_avl_set(?,?,?,-,-).
eq_cons_avl_set(A,H1,T1,Res,_) :- Res==pred_true,!,
    kernel_objects:equal_custom_explicit_set_cons(avl_set(A),H1,T1).
eq_cons_avl_set(A,H1,T1,Res,_) :-
    element_can_be_added_or_removed_to_avl(H1),!, % should hopefully be true as H1 is now ground
    (remove_element_from_explicit_set(avl_set(A),H1,RA)
      ->  equality_objects_with_expansion(T1,RA,Res)  %%% PROBLEM: T1 has to be in list form !
       ;  Res = pred_false
    ).
eq_cons_avl_set(A,H1,T1,Res,_) :-
   % TO DO: we could call propagate_avl_element_information(X,AVL,ApproxSize,WF) if Res=pred_true and with WF or something similar higher up to detect when an element cannot be within the AVL set
   expand_custom_set(avl_set(A),ExpSet),
   eq_cons1(ExpSet,H1,T1,Res).

:- block eq_cons1(-,?,?,?).
eq_cons1(ExpSet,H1,T1,Res) :- eq_cons(ExpSet,H1,T1,Res).

:- block eq_cons2(-, ?,?,?,?, ?,-).
eq_cons2(RemoveSuccesful, H1,T1,H2,T2, _NewSet2,Res) :- var(RemoveSuccesful),!,
    % result has been instantiated before Remove finished; now force equal_object or not_equal_object
    % note: without this the remove could be pending (see test 1112, 1105)
    % TO DO: look whether we could force remove and continue from there ?
    equality_objects2([H1|T1],[H2|T2],Res).
eq_cons2(not_successful, _,_T1,_,_, _NewSet2,pred_false).
eq_cons2(successful,     _, T1,_,_,  NewSet2,Res) :-  equality_objects(T1,NewSet2,Res).

:- block check_and_remove(-,?,?,?).
check_and_remove([],_,ResultingSet,Res) :- !, ResultingSet=[],Res=not_successful.
check_and_remove([H|T],ElementToRemove,ResultingSet,Res) :- !,
   equality_objects(H,ElementToRemove,RH),
   %%print(equality_objects_result(H,ElementToRemove,RH)),nl,
   check_and_remove2(RH,H,T,ElementToRemove,ResultingSet,Res).
check_and_remove(CS,ElementToRemove,ResultingSet,Res) :-
   expand_custom_set_to_list(CS,ESet,_,check_and_remove),
   check_and_remove(ESet,ElementToRemove,ResultingSet,Res).

:- block check_and_remove2(-,?,?,?,?,?).
check_and_remove2(pred_true, _H,T,_ElementToRemove,T,successful).
check_and_remove2(pred_false,H,T,ElementToRemove,[H|RT],Res) :-
   check_and_remove(T,ElementToRemove,RT,Res).


:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:and_equality(pred_false,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:and_equality(pred_true,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:and_equality(pred_true,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:and_equality(pred_true,pred_true,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:and_equality(pred_true,pred_false,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:and_equality(pred_false,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:and_equality(pred_false,pred_false,pred_true))).
:- block and_equality(-,-,-).
and_equality(X,Y,Res) :-
   (   X==pred_true  -> Res=Y
     ; X==pred_false -> Res=pred_false
     ; Y==pred_true  -> Res=X
     ; Y==pred_false -> Res=pred_false
     ; Res==pred_true -> X=pred_true, Y=pred_true
     ; Res==pred_false -> and_equality2(X,Y,Res)  % we know that one of X,Y must be pred_false
     ; add_error_fail(and_equality,'Illegal call: ',and_equality(X,Y,Res))
   ).
:- block and_equality2(-,-,?).
and_equality2(X,Y,Res) :-
   (   X==pred_true  -> Res=Y
     ; X==pred_false -> Res=pred_false
     ; Y==pred_true  -> Res=X
     ; Y==pred_false -> Res=pred_false
     ; add_error_fail(and_equality,'Illegal call: ',and_equality2(X,Y,Res))
   ).

/*
:- block bool_eq_atomic(-,?,?,-).
bool_eq_atomic(X,Y,OtherY,Res) :- nonvar(Res),!,
   (Res=pred_true -> X=Y ; X=OtherY).
bool_eq_atomic(X,Y,_,Res) :- X==Y -> Res=pred_true ; Res=pred_false. */


:- block bin_eq_atomic(-,?,-,?,?), bin_eq_atomic(?,-,-,?,?).
% no need to check for frozen_dif: a dif would instantly trigger instantiation to other value
% However, we could use when(?=(X,Y) ; nonvar(Res) as trigger ! maybe this would slowdown SATLIB tests ??
bin_eq_atomic(X,Y,Res,V1,V2) :- nonvar(Res),!,
   %print(forcing_bin_eq_atomic(X,Y,Res,V1,V2)),nl,
   (Res=pred_true -> X=Y
    ; %print(bin_dif(X,Y)),nl,
      ((X=V1,Y=V2) ; (X=V2,Y=V1))
    ). % difference with eq_atomic: we now know the value !
bin_eq_atomic(X,Y,Res,_,_) :-
  %print(triggering_bin_eq(X,Y,Res)), translate:print_frozen_info(Res),nl,
  (X==Y -> Res=pred_true ; Res=pred_false).

eq_atomic(X,Y,_Type,Res) :- nonvar(X),nonvar(Y),!,
  (X=Y -> Res=pred_true ; Res=pred_false).
eq_atomic(X,Y,_,Res) :- % check for attached dif co-routines; later we should probably use CHR
   frozen_dif(X,Y),!, Res=pred_false.
eq_atomic(X,Y,Type,Res) :- when((?=(X,Y);nonvar(Res)),eq_atomic2(X,Y,Type,Res)).

% version of eq_atomic without frozen_dif check (if already performed)
% eq_atomic_simple(X,Y,_Type,Res) :- % in all calling contexts this test is already performed
% nonvar(X),nonvar(Y),!, (X=Y -> Res=pred_true ; Res=pred_false).
eq_atomic_simple(X,Y,Type,Res) :- when((?=(X,Y);nonvar(Res)),eq_atomic2(X,Y,Type,Res)).

/*
Comment: % for integers CLPFD reification will not ground Res if X==Y,   need to do when((?=(X,Y)...
| ?- X in 1..10, Y in 1..10, (X#=Y  #<=> Z), X=Y.
Y = X,
X in 1..10,
Z in 0..1 ?  */

eq_atomic2(X,Y,Type,Res) :- nonvar(Res),!,
   (Res=pred_true -> X=Y ; dif(Type,X,Y)).
eq_atomic2(X,Y,_Type,Res) :-
   (X=Y -> Res=pred_true
         ; Res=pred_false).

:- use_module(fd_utils_clpfd,[neq_fd/3]).
dif(boolean,A,B) :- !, bool_pred:negate(A,B). % was: bool_dif(A,B).
dif(integer,A,B) :- !, integer_dif(A,B).
dif(global(T),A,B) :- !, neq_fd(A,B,T).
dif(_,A,B) :- dif(A,B).

:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:integer_dif(0,1))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:integer_dif(1,1))).
:- use_module(clpfd_interface,[clpfd_neq/2]).
integer_dif(A,B) :- clpfd_neq(A,B). % clpfd_neq calls dif on overflow


:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:bool_dif(pred_false,pred_true))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:bool_dif(pred_true,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:bool_dif(pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:bool_dif(pred_false,pred_false))).
bool_dif(A,B) :- bool_pred:negate(A,B).

/*  old version:
:- block eq_atomic(-,?,-), eq_atomic(?,-,-).
eq_atomic(X,Y,Res) :- nonvar(Res),!,
   (Res=pred_true -> X=Y ; dif(X,Y)).
eq_atomic(X,Y,Res) :- X==Y -> Res=pred_true
                            ; Res=pred_false.
*/

:- use_module(kernel_records,[check_field_name_compatibility/2]).
:- block equality_fields(-,?,-), equality_fields(?,-,-).
% if last argument is known we should propagate the field names from one record to another; see test 1289 (Hansen12)
equality_fields([],[],pred_true).
equality_fields([field(Name1,V1)|T1],[field(Name2,V2)|T2],Res) :-
 % should we wait for Name1 or Name2 to become nonvar?
  check_field_name_compatibility(Name1,Name2),
  and_equality(RH,RT,Res),
  equality_objects(V1,V2,RH),
  (RH==pred_false -> true
                ; equality_fields(T1,T2,RT)).



empty_set_test(Set,EqRes) :- kernel_equality:eq_empty_set(Set,EqRes).
empty_cartesian_product(Set1,Set2,EqRes) :-
   eq_empty_set(Set1,EqRes1),
   (EqRes1==pred_true -> EqRes=pred_true /* avoid inspecting Set2 */
    ; or_eq_obj(EqRes1,EqRes2,EqRes),eq_empty_set(Set2,EqRes2)).


:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:or_eq_obj(pred_false,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_check([commutative],kernel_equality:or_eq_obj(pred_false,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:or_eq_obj(pred_true,pred_true,pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:or_eq_obj(pred_true,pred_true,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:or_eq_obj(pred_true,pred_false,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:or_eq_obj(pred_false,pred_true,pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:or_eq_obj(pred_false,pred_false,pred_true))).
:- block or_eq_obj(-,-,-).
% very similar to disjoin in b_interpreter_check; no waitflag to instantiate if result false
or_eq_obj(X,Y,XorY) :-
  ( (X==pred_true ; Y==pred_true) -> XorY=pred_true
   ; XorY==pred_false -> X=pred_false,Y=pred_false
   ; X==pred_false -> XorY = Y
   ; Y==pred_false -> XorY = X
   ; XorY==pred_true -> or_eq_obj2(X,Y)
   ; add_internal_error('Illegal call: ',or_eq_obj(X,Y,XorY)) ).
:- block or_eq_obj2(-,-).
or_eq_obj2(X,Y) :- ((X==pred_true; Y=pred_true) -> true
            ; X==pred_false -> Y=pred_true
            ; Y==pred_false -> X=pred_true
            ; add_internal_error('Illegal call: ',or_eq_obj2(X,Y))).

% ------------------------------------------------------------------


% not used at the moment:
%get_partial_subset_priority(Set,WF,LWF) :- get_partial_set_size(Set,Size),
%              safe_pow2(Size,Prio),
%              get_wait_flag(Prio,psubset(Set),WF,LWF).
%get_partial_set_priority(Set,WF,LWF) :- get_partial_set_size(Set,Size),
%              get_wait_flag(Size,pset(Set),WF,LWF).
%get_partial_set_size(X,R) :- var(X),!,R=3.
%get_partial_set_size([],R) :- !,R=0.
%get_partial_set_size([_|T],R) :- !,get_partial_set_size(T,RT), R is RT+1.
%get_partial_set_size(_,10).


get_cardinality_wait_flag(Set,Info,WF,LWF) :-
    get_cardinality_wait_flag(Set,Info,WF,_Card,LWF).
get_cardinality_wait_flag(Set,Info,WF,Card,LWF) :-
    cardinality_as_int_for_wf(Set,Card),
  % when(nonvar(X), (print(get_card_waitflag(X,Set,WF,LWF)),nl)),
   (var(Card)
    -> kernel_waitflags:get_last_wait_flag(get_cardinality_wait_flag(Info),WF,LastWF)
     ; true), % ensure that we trigger choice points waiting on LWF before infinite enumeration starts
    get_cardinality_wait_flag2(Card,Info,WF,LastWF,LWF).
:- block get_cardinality_wait_flag2(-,?,?,-,?).
get_cardinality_wait_flag2(X,_Info,_WF,LastWF,LWF) :- var(X),!,LWF=LastWF.
get_cardinality_wait_flag2(X,Info,WF,_,LWF) :- get_bounded_wait_flag(X,Info,WF,LWF).

get_cardinality_powset_wait_flag(Set,Info,WF,CardSet,LWF) :-
    cardinality_as_int_for_wf(Set,CardSet),
    get_cardinality_powset_wait_flag2(CardSet,Info,WF,LWF).
:- block get_cardinality_powset_wait_flag2(-,?,?,?).
get_cardinality_powset_wait_flag2(X,Info,WF,LWF) :-
     priority_pow2(X,X2),
     %safe_mul(X2,10000,X3), % as generating subsets this way will generate repetitions
     %print(powset_waitflag(X,X2)),nl,
     (X2=inf
      -> integer_pow2_priority(R), get_wait_flag(R,powset_card(Info),WF,LWF)
      ; get_bounded_wait_flag(X2,powset_card(Info),WF,LWF) % finite prio; ensure we do not exceed infinite prios if X2 very large
      ).

priority_pow2(X,R) :- X==inf,!, R=inf. %R=10000010.  % TO DO: provide a separate module for combining priorities in a systematic way !
priority_pow2(X,R) :- safe_pow2(X,R).

% CARDINALITY FOR NUMBER OF PARTIAL FUNCTIONS: (not used at the moment)
%get_cardinality_partial_function_wait_flag(DomSet,RanSet,Info,WF,CardDomSet,CardRanSet,LWF) :-
%    cardinality_as_int_for_wf(DomSet,CardDomSet),
%    cardinality_as_int_for_wf(RanSet,CardRanSet),
%    get_cardinality_partial_function_wait_flag(CardDomSet,CardRanSet,Info,WF,LWF).
%:- block get_cardinality_partial_function_wait_flag(-,?,?,?,?),
%         get_cardinality_partial_function_wait_flag(?,-,?,?,?).
%get_cardinality_partial_function_wait_flag(DC,RC,Info,WF,LWF) :-
%     priority_pow1n(RC,DC,PFCARD), print(prio(PFCARD)),nl,
%     get_bounded_wait_flag(PFCARD,powset_card(Info),WF,LWF).
% % Cardinality of the set of partial functions :  (card(Range)+1)^card(Domain)
%priority_pow1n(X,Y,R) :- (X==inf ; Y==inf),!, integer_pow2_priority(R). %R=10000010.  % TO DO: provide a separate module for combining priorities in a systematic way !
%priority_pow1n(X,Y,RR) :- X1 is X+1, safe_pown(X1,Y,R),
% (R==inf -> integer_pow2_priority(RR) %RR=10000010
%  ; RR=R).

% CARDINALITY FOR NUMBER OF RELATIONS:
get_cardinality_relation_over_wait_flag(Dom,Ran,WF,LWF,MaxCardOfRel,MaxNrOfRels) :-
    cardinality_as_int_for_wf(Dom,X),
    cardinality_as_int_for_wf(Ran,Y),
    get_cardinality_relation_over_wait_flag2(X,Y,WF,LWF,MaxCardOfRel,MaxNrOfRels).

:- block get_cardinality_relation_over_wait_flag2(-,?,?,?,?,?), get_cardinality_relation_over_wait_flag2(?,-,?,?,?,?).
get_cardinality_relation_over_wait_flag2(X,Y,WF,LWF,MaxCardOfRel,MaxNrOfRels1) :-
    safe_mul(X,Y,MaxCardOfRel),
    safe_pow2(MaxCardOfRel,MaxNrOfRels),
    safe_add(MaxNrOfRels,1,MaxNrOfRels1),  /* add 1 to give priority to tighter_enum */
   % print(get_rel_card(X,Y,XY2)),nl,
    get_bounded_wait_flag(MaxNrOfRels1,powset_rel_card,WF,LWF).

% ------------------------------------------------------------------

:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_equality:membership_test_wf([int(2),int(1)],int(2),pred_true,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_equality:membership_test_wf([int(2),int(1)],int(1),pred_true,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_equality:membership_test_wf([int(2),int(1),int(3),int(4),int(5)],int(4),pred_true,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_equality:membership_test_wf([int(2),int(1)],int(3),pred_false,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_check_wf(kernel_equality:membership_test_wf([],int(3),pred_false,WF),WF)).
:- assert_must_succeed(exhaustive_kernel_fail_check_wf(kernel_equality:membership_test_wf([int(2),int(1)],int(3),pred_true,WF),WF)).
:- assert_must_succeed((membership_test_wf([int(2),int(1)],int(2),R,_WF),R==pred_true)).
:- assert_must_succeed((membership_test_wf([int(2),int(1)],int(X),R,_WF),var(R),X=1,R==pred_true)).
:- assert_must_succeed((membership_test_wf([int(2),int(1)],int(3),R,_WF),R==pred_false)).
:- assert_must_succeed((membership_test_wf(global_set('Name'),fd(1,'Name'),R,_WF),R==pred_true)).

:- assert_must_succeed((kernel_equality:membership_test_wf(avl_set(node(rec([field(cardinality,int(2)),field(col,fd(2,setField1)),field(fill,fd(1,setField2)),field(shape,fd(1,setField1))]),true,0,empty,empty)),
 rec([field(cardinality,int(2)),field(col,fd(2,setField1)),field(fill,fd(1,setField2)),field(shape,fd(1,setField1))]),R,_WF), R==pred_true)).
:- assert_must_succeed((
kernel_equality:membership_test_wf([rec([field(cardinality,int(2)),field(col,fd(2,setField1)),field(fill,fd(1,setField2)),field(shape,fd(1,setField1))])],
rec([field(cardinality,int(2)),field(col,fd(2,setField1)),field(fill,fd(1,setField2)),field(shape,fd(1,setField1))]),R,_WF), R==pred_true
)).
/* A function that instantiates last argument when membership test can be decided */
membership_test(Set,X,Res) :- print(obsolete_membership_without_wf(Set,X,Res)),nl,
   membership_test_wf(Set,X,Res,_WF).

% a version which also creates a choice point in SMT mode
membership_test_wf_with_force(Set,X,Res,WF) :-
   membership_test_wf(Set,X,Res,outer,WF),
   ((var(Res),preferences:preference(use_smt_mode,true))
    -> get_last_wait_flag(membership_test_wf,WF,LWF),
       force_equality_result(Res,LWF)
    ;  true).
:- block force_equality_result(-,-).
force_equality_result(pred_true,_).
force_equality_result(pred_false,_).

membership_test_wf(Set,X,Res,WF) :- membership_test_wf(Set,X,Res,outer,WF).

:- load_files(library(system), [when(compile_time), imports([environ/2])]).
:- if(environ(prob_data_validation_mode,true)).
:- block membership_test_wf(-,?,?,?,?). % avoid instantiating Set
:- else.
:- block membership_test_wf(-,?,-,?,?).
:- endif.
membership_test_wf(Set,X,Res,_,WF) :- var(Set),!,
   (Res==pred_true
    -> kernel_objects:mark_as_non_free(X), % attach co-routine to indicate that we cannot freely chose X
       Set=[X|_] % TO DO: normalise X first ?
    ; membership_not_el(Set,X,WF)).
membership_test_wf(Set,X,Res,OuterInner,WF) :-
    membership_test_wf2(Set,X,Res,OuterInner,WF).

:- block membership_not_el(-,?,?).
membership_not_el([],_,_WF) :- !.
membership_not_el([H|T],X,WF) :- !,kernel_objects:not_equal_object_wf(H,X,WF), membership_not_el(T,X,WF).
membership_not_el(avl_set(A),X,WF) :- !,membership_custom_set_wf(avl_set(A),X,pred_false,WF).
membership_not_el(CS,X,WF) :- is_custom_explicit_set_nonvar(CS),!,
    membership_custom_set_wf(CS,X,pred_false,WF).

membership_test_wf2([],_X,Res,_,_WF) :- !, Res=pred_false.
membership_test_wf2([H|T],X,Res,OuterInner,WF) :- !,
  (Res==pred_false
    -> kernel_objects:not_equal_object_wf(H,X,WF), membership_test_wf(T,X,Res,WF)
    ; T==[] -> equality_objects(H,X,Res)
    % TO DO: at the outer level we could check whether the list has the same length as the type of X
    ; (OuterInner=outer, is_definitely_maximal_set([H|T])) -> Res=pred_true
    ; (OuterInner=outer, % avoid re-checking ground_list_with_fd over and over again
       preferences:preference(use_clpfd_solver,true),
       ground_list_with_fd([H|T]),
       construct_avl_from_lists([H|T],CS))
    % this slows down: probsli ../prob_examples/public_examples/EventBPrologPackages/SET_Game/SET_GAM_Sym_NoSet20_mch.eventb -mc 10000000 -expcterr goal_found -p SYMMETRY_MODE hash -p TIME_OUT 7000 -p CLPFD TRUE -df -goal "n=18" -p MAX_OPERATIONS 82 -cc 155 833
    % TO DO: call clpfd_reify_inlist directly ? avoid translating to AVL and back to list; detect maximal list
        -> membership_custom_set_wf(CS,X,Res,WF)  %, print(conv_to_avl(X,[H|T],CS,Res)),nl
    ;  %print(eq(Res,H,X,T,WF)),nl,equality_objects_wf_dbg(H,X,HXRes,WF), % makes test 1003 : plavis-MemoryLoad_SP_55 fail
       equality_objects(H,X,HXRes),cons_el_of(HXRes,T,X,Res,WF)
      % ,(var(HXRes) -> print(mem(X,[H|T],Res)), tools_printing:print_arg(X),nl ; true)
   ).
membership_test_wf2(avl_set(A),X,Res,_,WF) :- !, membership_avl_set_wf(A,X,Res,WF).
membership_test_wf2(CS,X,Res,_,WF) :- is_custom_explicit_set(CS,membership_test_wf),!,
    membership_custom_set_wf(CS,X,Res,WF).
membership_test_wf2(Z,X,Res,OuterInner,WF) :-
    add_error_fail(membership_test_wf,'Not Set as first argument: ', membership_test_wf(Z,X,Res,OuterInner,WF)).

% detect certain lists containing FD values and convert them to avl sets: they have improved reification and can detect when a list is complete
ground_list_with_fd(V) :- var(V),!,fail.
ground_list_with_fd([H|T]) :- nonvar(H),ground_val(H),ground_list_with_fd(T).
ground_list_with_fd([]).

:- use_module(kernel_tools,[ground_value/1]).
ground_val(int(X)) :- number(X).
ground_val(pred_false).
ground_val(pred_true).
ground_val(fd(X,T)) :- number(X),ground(T).
ground_val((A,B)) :- nonvar(A),ground_val(A),ground_value(B).
% TO DO: add more checks; do we need this check ?

:- block cons_el_of(-,?,?,-,?).
cons_el_of(EQRES,T,X,MEMRES,WF) :- var(EQRES),!, %print(cons_el_of(EQRES,T,X,MEMRES)),nl,
   (MEMRES=pred_false -> EQRES=pred_false, /* otherwise element of set */
                           membership_not_el(T,X,WF)
      ; (T==[] -> %print(forcing_cons_el_of(T,X,MEMRES)),nl,
                  EQRES=pred_true
       ; cons_el_of_mem_obj(EQRES,T,X,WF)) /* we have to wait for result of comparison with X */
    ).
cons_el_of(pred_true,_,_,pred_true,_WF).
cons_el_of(pred_false,T,X,Res,WF) :- (Res==pred_true -> T=[_|_] ; true),
     membership_test_wf(T,X,Res,inner,WF).

:- block cons_el_of_mem_obj(-,?,?,?).
cons_el_of_mem_obj(pred_true,_,_,_WF).
cons_el_of_mem_obj(pred_false,T,X,WF) :- (var(T) -> T=[_|_] ; true),
     membership_test_wf(T,X,pred_true,inner,WF).


:- assert_must_succeed((kernel_equality:cartesian_pair_test_wf((int(1),int(1)),
           [int(1),int(2)],[int(2),int(3)],R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:cartesian_pair_test_wf((int(3),int(1)),
           [int(1),int(2)],[int(2),int(3)],R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:cartesian_pair_test_wf((int(2),int(3)),
           [int(1),int(2)],[int(2),int(3)],R,_WF),R==pred_true)).
:- assert_must_fail((kernel_equality:cartesian_pair_test_wf((int(1),int(3)),
           [int(1),int(2)],[int(2),int(3)],pred_false,_WF))).

cartesian_pair_test_wf((X,Y),XType,YType,Res,WF) :- check_cartesian_pair1(X,XType,Y,YType,Res,WF).

:- block check_cartesian_pair1(-,?,-,?,-,?).
%check_cartesian_pair1(X,XType,Y,YType,Res) :- Res==pred_true,!,
%   is_cartesian_pair_wf((X,Y),XType,YType,WF).
%check_cartesian_pair1(X,XType,Y,YType,Res) :- Res==pred_false,!,
%   not_is_cartesian_pair((X,Y),XType,YType,WF).
check_cartesian_pair1(X,XType,Y,YType,Res,WF) :-
   ((var(X),nonvar(Y)) -> check_cartesian_pair2(Y,YType,X,XType,Res,WF)
     ; check_cartesian_pair2(X,XType,Y,YType,Res,WF)).

check_cartesian_pair2(X,XType,Y,YType,Res,WF) :- prop_pred_true(Res,MemResX),
   membership_test_wf(XType,X,MemResX,WF),
   ((var(MemResX), is_definitely_maximal_set(YType)) % TO DO: maybe actually also call membership_test_wf
     -> Res=MemResX
     ;  check_cartesian_pair3(MemResX,Y,YType,Res,WF)).

:- block check_cartesian_pair3(-,?,?,?,?).
check_cartesian_pair3(pred_true,Y,YType,Res,WF) :- membership_test_wf(YType,Y,Res,WF).
check_cartesian_pair3(pred_false,_Y,_YType,pred_false,_WF).

:- block prop_pred_true(-,?).
prop_pred_true(pred_true,pred_true).
prop_pred_true(pred_false,_).





:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:in_nat_range_test(int(9),int(10),int(12),pred_false))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:in_nat_range_test(int(11),int(10),int(12),pred_true))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:in_nat_range_test(int(22),int(22),int(22),pred_true))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:in_nat_range_test(int(22),int(22),int(21),pred_false))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:in_nat_range_test(int(9),int(10),int(12),pred_true))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:in_nat_range_test(int(22),int(22),int(22),pred_false))).
:- assert_must_succeed((in_nat_range_test(X,int(11),int(12),pred_false), X=int(10) )).
:- assert_must_succeed((kernel_equality:in_nat_range_test(X,int(11),int(12),pred_false), X=int(13) )).
:- assert_must_succeed((in_nat_range_test(X,int(11),int(12),pred_true), X=int(11) )).
:- assert_must_succeed((kernel_equality:in_nat_range_test(int(11),X,int(12),pred_true), X=int(11) )).
:- assert_must_succeed((in_nat_range_test(int(11),X,int(12),pred_false), X=int(12) )).
:- assert_must_succeed((kernel_equality:in_nat_range_test(int(11),int(10),int(12),R), R==pred_true )).
:- assert_must_fail((kernel_equality:in_nat_range_test(int(11),int(10),int(12),pred_false))).

in_nat_range_test(int(X),int(Y),int(Z),Res) :- in_nat_range_test1(X,Y,Z,Res).

in_nat_range_test1(X,Y,Z,Res) :- number(X),number(Y),number(Z),!,
% for improved performance, e.g., inside large set differences with intervals
   (  X < Y -> Res = pred_false
    ; X > Z -> Res = pred_false
    ; Res = pred_true).
% this clause does not seem to add any power:  (x:x..x <=> b=TRUE) already determines b=TRUE deterministically
%in_nat_range_test1(X,Y,Z,ResYZ) :- Y==Z, !, print(in_nat_eq_obj(X,Y,ResYZ)),nl,
%   equality_objects(int(X),int(Y),ResYZ).
in_nat_range_test1(X,Y,Z,ResYZ) :-
   prop_eq_01(ResYZ,R01),
   % print(posting((X#>=Y #/\ X#=<Z) #<=> R01)),nl,
   % TO DO: check CHR + possibly X==Y or X==Z
   clpfd_interface:post_constraint( (X#>=Y #/\ X#=<Z) #<=> R01), % reify constraint if CLP(FD) active
   % print(in_nat_range_test(X,Y,Z,R01,ResYZ)),nl,
   %was : convert_eq_to_mem(RYZ,Res),
   and_equality(RY,RZ,ResYZ),
   geq_test(X,Y,RY),
   geq_test(Z,X,RZ).

:- block geq_test(?,-,-),geq_test(-,?,-).
geq_test(X,Y,Res) :- % TO DO: use CHR ?
   (  Res==pred_true -> kernel_objects:less_than_equal_direct(Y,X)
    ; Res==pred_false -> kernel_objects:less_than_direct(X,Y)
    ; X >= Y -> Res=pred_true
    ; Res = pred_false
   ).



/* SUBSET CHECK REIFICATION */

:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([int(2)],[],pred_false,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([int(2),int(1)],[int(2)],pred_false,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:subset_test([int(2),int(1)],[int(2)],pred_true,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([int(2),int(1)],[int(2),int(1)],pred_true,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([int(2),int(1)],[int(2),int(1),int(3)],pred_true,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([int(2),int(1)],global_set('NATURAL1'),pred_true,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_test([],[int(2),int(1)],pred_true,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:subset_test([],[int(2),int(1)],pred_false,_WF))).
:- assert_must_succeed((kernel_equality:subset_test([],[int(3)],R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test([int(2),int(1)],[int(2),int(1),int(3)],R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test([int(2),int(1),int(4)],[int(2),int(1),int(3)],R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(1)],global_set('NATURAL'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(1)],global_set('NATURAL1'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(1),int(0)],global_set('NATURAL1'),R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(1),int(0)],X,R,_WF),X=global_set('NATURAL1'),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(-11)],global_set('INTEGER'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test([int(222),int(-11)],global_set('INTEGER'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_test(global_set('INTEGER'),global_set('NATURAL1'),R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_test(global_set('NATURAL1'),global_set('NATURAL1'),R,_WF),R==pred_true)).

:- use_module(kernel_objects,[check_subset_of_wf/3, not_subset_of_wf/3,
                              check_subset_of_global_sets/2, check_not_subset_of_global_sets/2,
                              both_global_sets/4]).

:- use_module(custom_explicit_sets,[expand_custom_set_to_list/4,is_definitely_maximal_set/1]).
:- block subset_test(-,-,-,?).
subset_test(_S1,S2,R,_WF) :-   % print_term_summary(subset_test(_S1,S2,R,WF)), %%
   is_definitely_maximal_set(S2),!,R=pred_true.
subset_test(S1,S2,R,WF) :-  subset_test0(S1,S2,R,WF).

:- block subset_test0(-,?,-,?).
subset_test0(S1,S2,R,WF) :-   % print_term_summary(subset_test0(S1,S2,R,WF)), %%
   nonvar(R),!,
   (R==pred_true -> check_subset_of_wf(S1,S2,WF)
    ; R==pred_false -> not_subset_of_wf(S1,S2,WF)
    ; add_error_fail(subset_test0,'Illegal result value: ',R)).
subset_test0([],_,R,_WF) :- !,R=pred_true.
subset_test0(S1,S2,R,WF) :- subset_test1(S1,S2,R,WF).


:- use_module(coverage_tools_annotations,['$NOT_COVERED'/1]).

:- block subset_test1(-,?,-,?),subset_test1(?,-,-,?).
subset_test1(S1,S2,R,WF) :- nonvar(R),!,
   (R==pred_true -> check_subset_of_wf(S1,S2,WF) ; not_subset_of_wf(S1,S2,WF)).
subset_test1([],_,R,_WF) :- !,R=pred_true, '$NOT_COVERED'('cannot be covered').  /* cannot be covered; emtpy set treated above */
subset_test1(Set1,Set2,R,_WF) :-  kernel_objects:both_global_sets(Set1,Set2,G1,G2),!, % also deals with two intervals
    % print(test_subset_of_global_sets(G1,G2,R)),nl,
    test_subset_of_global_sets(G1,G2,R).
subset_test1(_Set1,Set2,R,_WF) :-
   is_definitely_maximal_set(Set2),!, % TO DO: avoid checking it again if have already checked it above in subset_test
   R=pred_true.
subset_test1(Set1,Set2,R,_WF) :- Set2==[],!, eq_empty_set(Set1,R).
subset_test1(Set1,Set2,R,WF) :- test_subset_of_explicit_set(Set1,Set2,R,WF,Code),!,
   call(Code).
subset_test1(Set1,Set2,R,WF) :-
  expand_custom_set_to_list(Set1,ESet1,_,subset_test1),
  subset_test2(ESet1,Set2,R,WF).

:- block subset_test2(-,?,?,?).
subset_test2([],_,R,_WF) :- !, R=pred_true.
subset_test2([H|T],Set2,R,WF) :-
   membership_test_wf(Set2,H,MemRes,WF),
   subset_test3(MemRes,T,Set2,R,WF).

:- block subset_test3(-,?,?,?,?).
subset_test3(pred_false,_T,_Set2,R,_WF) :- R=pred_false.
subset_test3(pred_true,T,Set2,R,WF)   :- subset_test2(T,Set2,R,WF).

%:- block test_subset_of_global_sets(-,?,?), not_subset_of_global_sets(?,-,?).
%test_subset_of_global_sets(interval(Low1,Up1),interval(Low2,Up2)) :- !, test_interval_subset(Low1,Up1,Low2,Up2).
test_subset_of_global_sets(G1,G2,R) :- G1==G2,!,R=pred_true.
test_subset_of_global_sets(G1,G2,R) :- %print(check_subset(G1,G2)),nl,
   when((nonvar(R);(ground(G1),ground(G2))),test_subset_of_global_sets_aux(G1,G2,R)).
% TO DO: provide better reified version, especially for intervals with variables !
% e.g.,  1..x <: NATURAL1  ==> R=pred_true, ...
test_subset_of_global_sets_aux(G1,G2,R) :- R==pred_true,!, check_subset_of_global_sets(G1,G2).
test_subset_of_global_sets_aux(G1,G2,R) :- R==pred_false,!, check_not_subset_of_global_sets(G1,G2).
test_subset_of_global_sets_aux(G1,G2,R) :-
    (check_subset_of_global_sets(G1,G2) -> R=pred_true ; R=pred_false).

% ---------------------------- strict subset reification


:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([int(2)],[],pred_false,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([int(2),int(1)],[int(2)],pred_false,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:subset_strict_test([int(2),int(1)],[int(2)],pred_true,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([int(2),int(1)],[int(2),int(1)],pred_false,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([int(2),int(1)],[int(2),int(1),int(3)],pred_true,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([int(2),int(1)],global_set('NATURAL1'),pred_true,_WF))).
:- assert_must_succeed(exhaustive_kernel_check(kernel_equality:subset_strict_test([],[int(2),int(1)],pred_true,_WF))).
:- assert_must_succeed(exhaustive_kernel_fail_check(kernel_equality:subset_strict_test([],[int(2),int(1)],pred_false,_WF))).
:- assert_must_succeed((kernel_equality:subset_strict_test([],[int(3)],R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(2),int(1)],[int(2),int(1),int(3)],R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(2),int(1),int(4)],[int(2),int(1),int(3)],R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(1)],global_set('NATURAL'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(1)],global_set('NATURAL1'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(1),int(0)],global_set('NATURAL1'),R,_WF),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(1),int(0)],X,R,_WF),X=global_set('NATURAL1'),R==pred_false)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(-11)],global_set('INTEGER'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test([int(222),int(-11)],global_set('INTEGER'),R,_WF),R==pred_true)).
:- assert_must_succeed((kernel_equality:subset_strict_test(global_set('INTEGER'),global_set('NATURAL1'),R,_WF),R==pred_false)).

:- assert_must_succeed((kernel_equality:subset_strict_test(global_set('NATURAL1'),global_set('NATURAL1'),R,_WF),R==pred_false)).

:- block subset_strict_test(-,-,-,?).
subset_strict_test(_S1,S2,R,_WF) :-   % print_term_summary(subset_test(_S1,S2,R,WF)), %%
   S2==[],!,R=pred_false.
subset_strict_test(S1,S2,R,_WF) :-   % print_term_summary(subset_test(_S1,S2,R,WF)), %%
   S1==[],!,eq_empty_set(S2,Empty), bool_pred:negate(Empty,R).
subset_strict_test(S1,S2,R,WF) :-  subset_strict_test0(S1,S2,R,WF).

:- use_module(kernel_objects,[strict_subset_of_wf/3, not_strict_subset_of_wf/3]).
:- block subset_strict_test0(-,?,-,?).
subset_strict_test0(S1,S2,R,WF) :-   % print_term_summary(subset_test0(S1,S2,R,WF)), %%
   nonvar(R),!,
   (R==pred_true -> strict_subset_of_wf(S1,S2,WF)
    ; R==pred_false -> not_strict_subset_of_wf(S1,S2,WF)
    ; add_error_fail(subset_strict_test0,'Illegal result value: ',R)).
subset_strict_test0(S1,S2,R,WF) :-
    subset_test(S1,S2,IsSubset,WF),
    (IsSubset==pred_false -> R=pred_false
     ; conjoin_test(IsSubset,NotEqual,R,WF),
       bool_pred:negate(NotEqual,Equal),
       equality_objects_wf(S1,S2,Equal,WF)
     ).

conjoin_test(A,B,AB,WF) :-
    conjoin(A,B,AB,b(truth,pred,[]),b(truth,pred,[]),WF).
