% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)


:- module(kernel_waitflags,
   [init_wait_flags/1, waitflag0_is_set/1,
    get_wait_flag_infos/2, get_wait_flag_info/2, is_wait_flag_info/2,
    add_wait_flag_info/3,
    get_wait_flag/3, get_wait_flag/4, get_bounded_wait_flag/4,
    get_wait_flag0/2,
    get_wait_flag1/2,  get_wait_flag1/3, % deterministic wait flag
    get_middle_wait_flag/3, get_last_wait_flag/3,
    get_idle_wait_flag/3,
    get_binary_choice_wait_flag/3,
    get_pow2_binary_choice_priority/2,
    get_binary_choice_wait_flag_exp_backoff/3, get_binary_choice_wait_flag_exp_backoff/4,
    %refresh_wait_flag/4,
    add_fd_variable_for_labeling/3, add_fd_variable_for_labeling/2, add_fd_variables_for_labeling/2,
    get_new_subsidiary_wait_flag/4,
    update_waitflag/4,
    ground_constraintprop_wait_flags/1,
    check_is_wait_flag/1,
    create_inner_wait_flags/3, copy_wf01e_wait_flags/2,
    copy_wf_start/2, copy_wf_finish/2,
    copy_waitflag_store/2,

    clone_wait_flags_from1/2, clone_wait_flags_from1_finish/2,
    ground_wait_flag0/1,
    %create_wdguarded_wait_flags/4,
    %get_abort_error_flag/2,
    get_large_finite_wait_flag/3,
    get_enumeration_starting_wait_flag/3, ground_wait_flag_to_enumeration/1,
    get_enumeration_finished_wait_flag/2,
    %get_integer_enumeration_wait_flag/3,
    ground_det_wait_flag/1, ground_wait_flag_to/2,
    ground_wait_flags/1, ground_inner_wait_flags/1,

    add_abort_error_span/5,
    add_wd_error_span/4,
    add_wd_error/3, add_wd_error_set_result/6,
    pending_abort_error/4,
    %add_wd_error_span_now/3,
    assert_must_abort_wf/2,

    get_minimum_waitflag_prio/3,
    portray_waitflags/1, quick_portray_waitflags/1, quick_portray_waitflags_store/1,
    find_waitflag_info/4,

    integer_priority/1, large_finite_priority/1, integer_pow2_priority/1,

    my_get_kodkod_predicates/3, my_add_predicate_to_kodkod/3,
    my_get_satsolver_predicates/3, my_add_predicate_to_satsolver/3,

    set_silent/1]).
/* File: kernel_waitflags.pl */
/* Created: 17/3/06 by Michael Leuschel */

/*
Phases of the ProB Kernel:

 -> all Waitflags unbound

* Phase 0:  propagation of completely known ground values in a deterministic fashion

 -> WF0 is bound

* Phase 1:  deterministic propagation (not necessarily of completely known full values)

 -> WF1 is bound

* Phase 2:  boolean propagation (e.g., P or Q will now try P or Q)

 -> WF2 is bound

* Phase 3:  non-deterministic propagation (e.g, x: {1,2,3})

 -> Enumeration Starting WF is bound

* Enumeration Phase:  (later to be split into enumeration of finite and infinite types (e.g., integer)

  -> Enumeration Finished Waitflag bound

* Error Generation Phase: errors for, e.g., division by zero, f(x) with x outside domain of f,... are generated
* This phase should not create choice points and should not detect failure
  (e.g., in the
  case of negated existential quantifiers the Enumeration Finished Waitflag will be set outside of delay_not)


We now distinguish
 - Enumeration Wait Flags for Decision Variables with estimated enumeration domain size as priority
 - Binary Choice Point Wait Flags (stemming e.g. from disjunctions)

Note: all floats like 1.0, 1.5, 2.0 ... are dealt with before all integers (even 1) !

*/

:- use_module(error_manager).
:- use_module(self_check).
:- use_module(typechecker).
:- use_module(library(clpfd),[labeling/2, fd_size/2, fd_degree/2]).
:- use_module(tools).
:- use_module(coverage_tools_annotations,['$NOT_COVERED'/1]).
:- use_module(preferences,[preference/2]).

:- use_module(module_information,[module_info/2]).
:- module_info(group,kernel).
:- module_info(description,'This module manages the various choice points and enumerators of ProB, by maintaining the kernel Waitflags store.').

:- type wait_flag +--> (var ; wfx(any,var,any,list(any))).

:- use_module(library(avl)).
:- use_module(library(atts)).

:- attribute wf_store/1,wf_fdvars/1, wf_kodkod/2, wf_satsolver/2.

% the Call must succeed and generate one abort_error
:- meta_predicate assert_must_abort_wf(0,*).
%:- meta_predicate assert_must_abort2_wf(0,-).
assert_must_abort_wf(M:Call,WF) :- assert_must_succeed_any(M:(kernel_waitflags:assert_must_abort2_wf(M:Call,WF))).

assert_must_abort2_wf(_,_) :- get_error(well_definedness_error,M),
   add_internal_error('Abort error prior to call in assert_must_abort: ',M),fail.
assert_must_abort2_wf(Call,WF) :- (real_error_occurred -> RE=true ; RE=false),
  init_wait_flags(WF),
  call(Call), %print(called(Call)),nl,
  (ground_wait_flags(WF)
  -> add_internal_error('Call did not fail after grounding WF: ',assert_must_abort2_wf(Call,WF)),fail
  ;  (get_error(well_definedness_error,_), % we have 1 abort_error
       print(got_well_definedness_error),nl,
       (RE=false -> reset_real_error_occurred ; true)
     )
  ).

:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF), kernel_waitflags:check_is_wait_flag(WF) )).

check_is_wait_flag(WFX) :- var(WFX),!,
  add_internal_error('Variable as waitflag: ',check_is_wait_flag(WFX)).
check_is_wait_flag(WFX) :-
  WFX = wfx(_,Store,_,_)
    -> (get_atts(Store,+wf_store(_))
         -> true
          ; print('WF does not yet contain heap: '), print(WFX),nl)
            %add_internal_error('WF does not contain heap: ',check_is_wait_flag(WFX)))
    ;  add_internal_error('Illegal WF Format: ',check_is_wait_flag(WFX)).

verify_attributes(Var, Value, [] ) :-
  nonvar(Value) -> print('Trying to unify waitflag_heap_store'),nl
                 ;  print(verify_attributes(Var,Value)),nl.

init_wait_flags(wfx(_,_,_,[])).
%init_wait_flags(wfx(_,Store,_,[])) :- init_wait_flags_store(Store).
% we no longer do this because of performance, (see, e.g., LotsOfInvariants.mch) :
init_wait_flags_store(_).
%init_wait_flags_store(Store) :-
%	empty_avl(Heap),
%	put_atts(Store,+wf_store(Heap)),
%	put_atts(Store,+wf_fdvars([])).


:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
     kernel_waitflags:ground_wait_flag0(WF), kernel_waitflags:waitflag0_is_set(WF) )).
:- assert_must_fail(( kernel_waitflags:init_wait_flags(WF), kernel_waitflags:waitflag0_is_set(WF) )).
waitflag0_is_set(wfx(WF0,_Store,_EF,_INFOS)) :- ground(WF0).
waitflag0_is_set(no_wf_available).

get_new_subsidiary_wait_flag(OldWF,Info,WFX,NewWF) :-
   (ground(OldWF) /* then we have created a choice point with the old WF: double new WF prio */
      -> NewPrio is OldWF*2, get_wait_flag(NewPrio,Info,WFX,NewWF)
      ;  NewWF = OldWF
   ).

% not used at the moment:
%refresh_wait_flag(OldWF,Info,WFX,NewWF) :-
%   (ground(OldWF) /* then we have created a choice point with the old WF: double new WF prio */
%      -> get_wait_flag(OldWF,Info,WFX,NewWF)
%      ;  NewWF = OldWF
%   ).

:- volatile silent/0.
:- dynamic silent/0.
set_silent(true) :- !,(silent -> true ; assert(silent)).
set_silent(false) :- retractall(silent).
%waitflag_not_init :-  silent -> true ; print('waitflag-store not initialised'),nl, '$NOT_COVERED'('This should not happen').

:- use_module(eventhandling,[register_event_listener/3]).
:- register_event_listener(start_unit_tests,set_silent(true),'Allow wf not to be set up without printing a warning').
:- register_event_listener(stop_unit_tests, set_silent(false), 'Printing wf warnings again').

:- use_module(library(lists),[maplist/2]).
add_fd_variables_for_labeling(Vars,WF) :- WF==no_wf_available,!,
   add_internal_error('Cannot add FDVars to waitflag store: ',Vars).
add_fd_variables_for_labeling(Vars,WF) :-
   WF=wfx(_,Store,_EnumFinished,_INFOS),
   !,
   my_get_fdvars_atts(Store,FDVARS),
   l_add_fd_var_to_FDVARS(Vars,FDVARS,NewFDVARS),
   put_atts(Store,+wf_fdvars(NewFDVARS)).
add_fd_variables_for_labeling(Vars,WF) :-
   add_internal_error('Illegal waitflag store: ',add_fd_variables_for_labeling(Vars,WF)).

l_add_fd_var_to_FDVARS([],Acc,Acc).
l_add_fd_var_to_FDVARS([Var|VT],FDVARS,NewFDVARS) :-
    (nonvar(Var) -> F2=FDVARS;  add_fd_var_to_FDVARS(Var,FDVARS,F2)),
    l_add_fd_var_to_FDVARS(VT,F2,NewFDVARS).

% adds a CLP(FD) variable with given domain Size, will be enumerated by labeling (even if clpfd_solver preference is false !)
% Size could be computed by fd_size(FDVariable,Size)
add_fd_variable_for_labeling(FDVariable,WF) :-
    add_fd_variable_for_labeling_aux(FDVariable,WF).
add_fd_variable_for_labeling(FDVariable,Size,WF) :- Size = not_used_anymore,
    add_fd_variable_for_labeling_aux(FDVariable,WF).


add_fd_variable_for_labeling_aux(FDVariable,_WF) :- nonvar(FDVariable),!.
add_fd_variable_for_labeling_aux(FDVariable,WF) :-
  % otherwise: store the FDVariable in the separate list of FD Variables (all mixed together)
  WF=wfx(_,Store,_EnumFinished,_),
  my_get_fdvars_atts(Store,FDVARS),
  add_fd_var_to_FDVARS(FDVariable,FDVARS,NewFDVARS),
   %%  fd_size(FDVariable,Size),print(added(FDVariable,Size,NewFDVARS,old(FDVARS))),nl,
    %%  print(added_fd_variable(FDVariable,_Size,NewFDVARS)),nl,portray_waitflags_store(Store),nl,nl,%%
  put_atts(Store,+wf_fdvars(NewFDVARS)).

add_fd_var_to_FDVARS(FDVariable,FDVARS,NewFDVARS) :-
  %insert_fd_var(FDVARS,FDVariable,Size,LWF,NewFDVARS),
  ((FDVARS=[HH|_],HH==FDVariable) -> NewFDVARS = FDVARS % variable already in list [cheap check at front]
    ; NewFDVARS = [FDVariable|FDVARS]). %% this adds new variables at front; will be given priority over older ones
  %% add_fd_var(FDVARS,FDVariable,NewFDVARS),  %% this adds new variables at the end and does more thorough duplicate check


% a bit like append: but cleans up list: removes numbers + checks if var already in list
% runtimes for tests: 349 stays at 220ms, test 1088 4660 -> 5040ms
%add_fd_var([],Var,[Var]).
%add_fd_var([V|T],Var,Res) :-
%  (V==Var -> Res = [V|T] %,print(dup(V)),nl,
%   ; nonvar(V) -> add_fd_var(T,Var,Res) %,print(nonvar(V)),nl
%   ; Res = [V|RT], add_fd_var(T,Var,RT)).

my_get_fdvars_atts(Store,FDVARS) :-
   (get_atts(Store,+wf_fdvars(FDVARS)) -> true
	  ;  var(Store),FDVARS=[]
   ).
	   /* store not initialised; initialise it below but print msg */
	  %   ,waitflag_not_init).

:- use_module(library(random),[random_select/3]).
% reverse the list at each step; alternating taking from end and front; makes no sense with randomise_enumeration_order
%my_get_next_fdvar_to_enumerate_rev(FDVARS,NextFDVar,RemainingFDVARS) :- lists:reverse(FDVARS,RevFDVARS),
%   my_get_next_fdvar_to_enumerate(RevFDVARS,NextFDVar,RemainingFDVARS).

my_get_next_fdvar_to_enumerate([],_NextFDVar,_RemainingFDVARS) :- !,fail.
my_get_next_fdvar_to_enumerate([V1|FDVARS],NextFDVar,RemainingFDVARS) :-
   nonvar(V1),
   !, % skip over non-variables (numbers); clpfd_get_next_variable_to_label does not delete variable from list and may crash with leading non-variable terms !
   my_get_next_fdvar_to_enumerate(FDVARS,NextFDVar,RemainingFDVARS).
my_get_next_fdvar_to_enumerate(FDVARS,NextFDVar,RemainingFDVARS) :-
   preferences:preference(randomise_enumeration_order,true),
   !,
   get_min_fd_size_elements(FDVARS,MinFDVars),
   random_select(NextFDVar,MinFDVars,_),
   %print(random_enum(NextFDVar,FDVARS,MinFDVars)),nl,
   RemainingFDVARS=FDVARS. % still contains variable; but no problem as it will be discarded later
my_get_next_fdvar_to_enumerate(FDVARS,NextFDVar,RemainingFDVARS) :-
   %clpfd_interface:clpfd_get_next_variable_to_label_ffc(FDVARS,NextFDVar,RemainingFDVARS).
   clpfd_interface:clpfd_get_next_variable_to_label(FDVARS,NextFDVar),
   % print(enumerating(NextFDVar,FDVARS)),nl, maplist(tools_printing:print_arg,FDVARS),nl,
   RemainingFDVARS=FDVARS.
%   ((FDVARS=[H|T],H==NextFDVar) -> RemainingFDVARS=T ;  RemainingFDVARS=FDVARS).% does not seem to buy any speed

get_min_fd_size_elements([V1|FDVARS],MinFDVars) :- my_get_fd_size(V1,Size),
   get_min_aux(FDVARS,Size,[V1],MinFDVars).

get_min_aux([],_,MinFDVars,MinFDVars).
get_min_aux([V1|FDVARS],Size,MAcc,MinFDVars) :- number(V1),!,
   get_min_aux(FDVARS,Size,MAcc,MinFDVars).
get_min_aux([V1|FDVARS],Size,MAcc,MinFDVars) :-
   my_get_fd_size(V1,Size1),
   (lt_size(Size1,Size)
    -> get_min_aux(FDVARS,Size1,[V1],MinFDVars)
     ; Size1=Size -> get_min_aux(FDVARS,Size,[V1|MAcc],MinFDVars)
     ; get_min_aux(FDVARS,Size,MAcc,MinFDVars)
    ).

lt_size(X,Y) :- number(X),(Y=sup -> true ; X<Y).

% -------------

% using this is relevant for e.g. tests 415, 416, 1096 : to give variables without frozen goals a lower priority
% TO DO: we could filter out those before we call clpfd_get_next_variable_to_label !?
fd_priority_leq_limit(FDVAR,Prio,Limit) :-
  my_get_fd_size(FDVAR,Size),
  number(Size), % not =sup or =inf
  geq_limit(Limit,Size), % the calculation below will only increase the Size
  (Size < 2 -> Prio = Size
   ; geq_limit(Limit,Size*4) -> Prio = Size % avoid calling frozen and fd_degree, Priority only used for warnings and bisect decisions
   ; %print(sz(Size,Limit)),nl,
     (enumeration_only_fdvar(FDVAR) -> Prio is Size*4, Prio =< Limit
           ; Prio = Size)
  ).

my_get_fd_size(FDVAR,ResSize) :- fd_size(FDVAR,Size),!, ResSize=Size.
  %(Size=sup -> ResSize = 10000000 /* same as for inf */  ; ResSize=Size).
my_get_fd_size(FDVAR,Size) :- add_internal_error('Could not get fd_size: ',my_get_fd_size(FDVAR,Size)), Size=sup.

gt_limit(inf,_) :- !.
gt_limit(sup,_) :- !.
gt_limit(X,Limit) :- X > Limit.

geq_limit(inf,_) :- !.
geq_limit(sup,_) :- !.
geq_limit(X,Limit) :- X >= Limit.



% -------------


enumeration_only_fdvar(FDVAR) :- var(FDVAR),fd_degree(FDVAR,D),!,
  D=0,
  frozen(FDVAR,Goal), %print(frozen(FDVAR,D,Goal)),nl,
  enum_only_frozen_goal(Goal,FDVAR).
enumeration_only_fdvar(_).

% TO DO: maybe use enumeration_only_goal in b_enumerate
enum_only_frozen_goal((A,B),Var) :- !, enum_only_frozen_goal(A,Var), enum_only_frozen_goal(B,Var).
enum_only_frozen_goal(clpfd:_,_) :- !.
enum_only_frozen_goal(kernel_objects:G,V) :- !, enum_only_frozen_goal_k_obj(G,V).
enum_only_frozen_goal(true,_).

enum_only_frozen_goal_k_obj(safe_less_than_equal(_,X,Y),_) :- (number(X);number(Y)), !.
enum_only_frozen_goal_k_obj(call_enumerate_int(X,_,_,_),Var) :- Var==X,!.
enum_only_frozen_goal_k_obj(enumerate_int_wf(X,_,_),Var) :- Var==X,!.
enum_only_frozen_goal_k_obj(true,_).


% get a wait-flag that will be triggered first next time;
% can be used to ensure that all pending co-routines complete
get_idle_wait_flag(Info,wfx(_WF0,Store,EnumFinished,_),LWF) :- !,
    (nonvar(EnumFinished) -> LWF=1  % this waitflag store has already been grounded to completion; nobody may drive it anymore
     ; get_waitflag_from_store(0.9,Info,Store,LWF)).


get_wait_flag(Prio,Store,WF) :- get_wait_flag(Prio,'??',Store,WF).
get_wait_flag(0,_Info,wfx(WF0,_Store,_EnumFinished,_),WF) :- !, WF=WF0.
get_wait_flag(1.0,Info,wfx(WF0,Store,_EnumFinished,_),WF) :- !,
   (var(WF0) -> get_waitflag_from_store(1.0,Info,Store,WF) ; WF=1).
get_wait_flag(Prio,Info,wfx(_,Store,EnumFinished,_),WF) :- !,
   (ground(EnumFinished) -> WF=Prio /* enumeration has finished: return a ground WF */
    ; get_waitflag_from_store(Prio,Info,Store,WF)).
get_wait_flag(Prio,Info,Store,WF) :-
   (Store=no_wf_available -> WF=Prio ;
    Store=none -> print(get_wait_flag_deprecated(Prio,Info,Store,WF)),nl, WF=Prio ;
    add_internal_error('Illegal call: ',get_wait_flag(Prio,Info,Store,WF))).

get_waitflag_from_store(Prio,Info,Store,WF) :-
   % print(get_waitflag_from_store(Prio,Info,Store,WF)),nl,
    (Prio=inf -> integer_priority(X), RealPrio = float(X)
       % , print(infinite_priority(Prio,Info)),nl %%
     % ; (Prio>10, Prio<10000) -> RealPrio = float(100) % perform left-to-right enumeration in that range
      ; RealPrio=float(Prio)),
	my_get_wf_store_atts(Store,Heap),
	(avl_fetch(RealPrio,Heap,wf(WFs,_OldInfo))
	 -> (is_finite_priority(RealPrio) % TO DO: investigate performance if reduce the threshold
	      -> WF1 = WFs % here we re-use the waitflag instead of storing a new one
	      ;  (var(WFs) -> WFs2 = [WF1,WFs] ; WFs2 = [WF1|WFs]), % else: store waitflags as list; important for test 1194
	         %print(not_reusing(WFs2)),nl,
	         avl_store(RealPrio,Heap,wf(WFs2,Info),NewHeap),
	         put_atts(Store,+wf_store(NewHeap))
	    )
	  ; avl_store(RealPrio,Heap,wf(WF1,Info),NewHeap),
	    put_atts(Store,+wf_store(NewHeap))
	),!,
	%avl_size(NewHeap,Sz),
	%(Sz=10 -> print(heap_size(10)),nl ; true),
	%(Sz=1000 -> print(heap_size(1000)),nl, portray_avl(NewHeap) ; true),
	WF=WF1.
get_waitflag_from_store(Prio,Info,Store,WF) :-
    add_internal_error('Error getting Waitflag: ', get_waitflag_from_store(Prio,Info,Store,WF)),fail.

% -------------------

my_get_wf_store_atts(Store,Heap) :-
   (get_atts(Store,+wf_store(Heap)) -> true
	  ;  var(Store),empty_avl(Heap)
	).
	%,  /* store not initialised; initialise it below but print msg */
	%     waitflag_not_init).

% -------------------

% storing Predicates for Kodkod external function:
my_get_kodkod_predicates(wfx(_,Store,_,_),ID,PredicateList) :- my_get_wf_kodkod(Store,ID,PredicateList).
my_get_wf_kodkod(Store,ID,PredicateList) :-
   (get_atts(Store,+wf_kodkod(ID,PredicateList)) -> true
	  ;  var(Store),PredicateList = []
	).
%my_store_wf_kodkod(Store,ID,PredicateList) :-  put_atts(Store,+wf_kodkod(ID,PredicateList)).
my_add_predicate_to_kodkod(wfx(_,Store,_,_),ID,Predicate) :-
   my_get_wf_kodkod(Store,ID,PredicateList),
   put_atts(Store,+wf_kodkod(ID,[Predicate|PredicateList])).

% -------------------

% storing Predicates for external function calling satsolver:
my_get_satsolver_predicates(wfx(_,Store,_,_),ID,PredicateList) :- my_get_wf_satsolver(Store,ID,PredicateList).
my_get_wf_satsolver(Store,ID,PredicateList) :-
    (get_atts(Store,+wf_satsolver(ID,PredicateList)) -> true
	  ;  var(Store),PredicateList = []
    ).
my_add_predicate_to_satsolver(wfx(_,Store,_,_),ID,Predicate) :-
    my_get_wf_satsolver(Store,ID,PredicateList),
    put_atts(Store,+wf_satsolver(ID,[Predicate|PredicateList])).

% -------------------


ground_wait_flag0(wfx(WF0,_,_,_)) :- grd_wf0(WF0).

ground_wait_flags(wfx(WF0,Store,EnumFinish,_)) :-  !,
    grd_wf0(WF0),
    ground_waitflags_store_clpfd(Store),
    grd_ef(EnumFinish).
ground_wait_flags(no_wf_available) :-  !.
ground_wait_flags(W) :-
    add_internal_error('Waitflags in wrong format: ',ground_wait_flags(W)).

grd_wf0(E) :- var(E) -> E=0 %%,print(grd_wf0_done),nl
                     ; (E==0 -> true ; add_internal_error('Illegal WF0 Waitflag value: ',grd_wf0(E))).
grd_ef(E) :- var(E) -> E=0
             ; ((E==0;E==wd_guarded) -> true ; add_internal_error('Illegal EF Waitflag value: ',grd_ef(E))).

/* currently unused:
create_wdguarded_wait_flags(wfx(WF0,Store,EnumFinish,Info),Res,Expected,
                            wfx(WF0,Store,LocalEnumFinish,Info)) :-
  %% if Res=Expected then the computations of the inner waitflags should be well-defined;
  %%   otherwise they may not
    copy_wfe(EnumFinish,Res,Expected,LocalEnumFinish).
:- block copy_wfe(-,?,?,?), copy_wfe(?,-,?,?).
copy_wfe(EnumFinish,Res,Expected,LocalEnumFinish) :-
  (Res==Expected -> LocalEnumFinish=EnumFinish
   ; LocalEnumFinish=wd_guarded %% branch is pruned by well-definedness condition
   ).
*/

ground_constraintprop_wait_flags(wfx(WF0,Store,_WFE,_Infos)) :-
  %% print(ground_constraintprop_wait_flags(WF0,Store,_WFE)),nl, %%
  !,grd_wf0(WF0),% print(finished_grounding_wf0),nl,portray_waitflags_store(Store),
  ground_waitflags_store_clpfd(Store).
ground_constraintprop_wait_flags(no_wf_available) :-  !.
ground_constraintprop_wait_flags(W) :-
   add_internal_error('Waitflags in wrong format',ground_constraintprop_wait_flags(W)).

:- use_module(clpfd_interface,[clpfd_in_domain/1, clpfd_labeling/2]).
:- use_module(library(lists),[reverse/2]).

%ground_wf(Prio,WF) :- nl,print(ground(WF,Prio)),nl,WF=Prio, print(done_ground(WF,Prio)),nl.
% ground_wf(Prio,WF) :- Prio=WF.
ground_wf(Prio,WF) :- (var(WF) -> WF=Prio ; if(WF=Prio,true,add_internal_error('Illegal waitflag: ',ground_wf(Prio,WF)))).

ground_prio(Prio,WF,Info) :- %print(ground_prio(Prio,WF,_Info)),nl,
   nonvar(WF),!,
   (WF=[H|T] ->  % we have a list of waitflags; ground one-by one; TO DO: only ground one and store remaining list in wfx Heap
          reverse([H|T],Ls),  %print(list(Ls)),nl,
          maplist(ground_wf(Prio),Ls)
        ; ground(WF) -> true % already grounded by somebody else
        ; add_internal_error('Waitflag already partially grounded: ',ground_prio(Prio,WF,Info))).
ground_prio(Prio,WF,_Info) :-
   %external_functions:indent_format(user_output,'~w (~w)~n',[Prio,Info],no_trace),portray_attached_goals(WF),nl,
   %% (debug:debug_mode(on) -> external_functions:indent_format(user_output,'~w (~w)~n',[Prio,Info],no_trace) ; true),
		        % print(ground(Prio,WF,Info)),nl, translate:print_frozen_info(WF),nl,
				WF = Prio.                              % ground waitflag
	            % print(grounded(Prio,Info)),nl,nl, %%

%:- use_module(kernel_objects,[enum_warning_large/3]).
label_clpfd_variable(Variable) :- label_clpfd_variable(2097153,Variable).
label_clpfd_variable(Prio,Variable) :-
    ((gt_limit(Prio,2097152),  % note Prio > test also works with float(X) terms !
      check_if_labeling_domain_large(Variable,2097152,RANGE))  % {x|x:1..2**23 & x mod 2 = x mod 1001} takes about 2 minutes; {x|x:1..2**21 & x mod 2 = x mod 1001} takes about 30 seconds; 2**21 = 2097152, 2**23 = 8388608
		-> kernel_objects:enum_warning_large(Variable,'INTEGER VARIABLE',RANGE),  % could in principle also be enumerated set; but extremely unlikely given size
		   clpfd_labeling([Variable],[bisect]) % ffc no longer required: just a single variable
		    % we use bisecting rather than default enumeration; e.g., x:1..8388610 & y:1..8388610 & x*x*x = y*y + y + 10 is relatively quickly determined to be unsatisfiable (380 ms)
		    % Note: test 1099 tests this
		    % TO DO: investigate whether it makes sense to use bisect for smaller values already
		; preferences:preference(randomise_enumeration_order,true) -> clpfd_interface:clpfd_randomised_in_domain(Variable)
		; clpfd_in_domain(Variable)). % TO DO: use random ?

check_if_labeling_domain_large(H,Limit,Res) :-
    my_get_fd_size(H,Size),
    gt_limit(Size,Limit),
    clpfd_interface:clpfd_domain(H,From,To),
    Res=From:To.


% mix ProB labeling with CLP(FD) labeling
ground_waitflags_store_clpfd(Store) :-
			my_get_wf_store_atts(Store,Heap),!,
			ground_waitflags_store_clpfd_aux(Store,Heap).
ground_waitflags_store_clpfd(Store) :-
   add_internal_error('Illegal Waitflag: ',ground_waitflags_store_clpfd(Store)),fail.

ground_waitflags_store_clpfd_aux(Store,Heap) :-
			%my_get_fdvars_atts(Store,FDVars),
			get_atts(Store,+wf_fdvars(FDVars)), % there is a list of FD vars available
			!,
			( avl_del_min(Heap,Prio,wf(WF,Info),NewHeap) ->          % we have non-ground waitflags with a matching priority
			 ((my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,RemainingFDVARS),
			   fd_priority_leq_limit(NextFDVar,FDPrio,Prio)
			   )
			   -> enum_fd_variable_store(FDPrio,Store,NextFDVar,RemainingFDVARS)
			   ;  put_atts(Store,+wf_store(NewHeap)),   % write the modified heap back
			        % debug:hit_counter(XXX),print(grounding(XXX,Prio,_Info)),nl, (XXX=16556->trace ; true),
				  %WF = Prio,                            % ground waitflag
				  % print(grounding(Prio,Info)),nl,
				  ground_prio(Prio,WF,Info),
			      ground_waitflags_store_clpfd(Store)         % continue recursively
	          )
			; /* otherwise -> */
	           (my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,RemainingFDVARS)
			      -> enum_fd_variable_store(Store,NextFDVar,RemainingFDVARS)
			      ; true  % heap and FDVars list is empty, nothing more to do
			    )
			).
ground_waitflags_store_clpfd_aux(Store,Heap) :- % no FDVars are available to enumerate
              avl_del_min(Heap,Prio,wf(WF,Info),NewHeap), !,
			  put_atts(Store,+wf_store(NewHeap)),   % write the modified heap back
				% debug:hit_counter(XXX),print(grounding(XXX,Prio,_Info)),nl, (XXX=16556->trace ; true),
			  ground_prio(Prio,WF,Info),                            % ground waitflag
			  ground_waitflags_store_clpfd(Store).         % continue recursively
ground_waitflags_store_clpfd_aux(_Store,_Heap).   % heap and FDVars list is empty, nothing more to do

%:- use_module(library(lists),[exclude/3]).
enum_fd_variable_store(Store,NextFDVar,RemainingFDVARS) :-
   %exclude(nonvar,RemainingFDVARS,RemainingFDVARS2),
   put_atts(Store,+wf_fdvars(RemainingFDVARS)),
    %%  print(labeling(NextFDVar,RemainingFDVARS)),nl, portray_fd_vars([NextFDVar|RemainingFDVARS]),nl,nl,
   label_clpfd_variable(NextFDVar),
   ground_waitflags_store_clpfd(Store).
% a version of the above where we know the priority
enum_fd_variable_store(FDPrio,Store,NextFDVar,RemainingFDVARS) :-
   %exclude(nonvar,RemainingFDVARS,RemainingFDVARS2), % does not seem to buy a lot
   put_atts(Store,+wf_fdvars(RemainingFDVARS)),
    %%  print(labeling(NextFDVar,RemainingFDVARS)),nl, portray_fd_vars([NextFDVar|RemainingFDVARS]),nl,nl,
   label_clpfd_variable(FDPrio,NextFDVar),
   ground_waitflags_store_clpfd(Store).

:- if(false).
% small utility to trace through waitflag enumeration
:- dynamic spying_on/1.
spying_on(1).
spy_waitflags(_,_) :- \+ spying_on(_) , !.
spy_waitflags(Msg,Store) :- retract(spying_on(Nr)),print(Msg),nl,
   (Nr=1
    -> assert(spying_on(Nr)),
       print(' (j,t,p,q) ==> '),read_term(T,[]),
       action(T,Store)
    ;  N1 is Nr-1, assert(spying_on(N1))
   ).
action(t,_Store) :- !,trace.
action(p,Store) :- !,portray_waitflags_store(Store), spy_waitflags('',Store).
action(q,_Store) :- !,retractall(spying_on).
action(Nr,_) :- number(Nr),!, retractall(spying_on(_)), assert(spying_on(Nr)).
action(_,_).
:- endif.

:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF), kernel_waitflags:portray_waitflags(WF),
      kernel_waitflags:get_wait_flag(2,'Test',WF,WF2), when(nonvar(WF2),print(wf2)),
      clpfd:in(FD,'..'(1,2)),kernel_waitflags:add_fd_variable_for_labeling(FD,WF),
      kernel_waitflags:portray_waitflags(WF), kernel_waitflags:ground_wait_flags(WF)  )).
% PORTRAYING Waitflags
portray_waitflags(wfx(WF0,Store,WFE,Infos)) :- flush_output,
            print('*WAITFLAG STORE: '),print(wfx(WF0,Store,WFE,Infos)),nl,
            (var(WF0) -> print('> WF0 :: '), portray_attached_goals(WF0),nl ; true),
            portray_waitflags_store(Store),!,
            (var(WFE) -> print('> WFE :: '), portray_attached_goals(WFE),nl ; true),
            print('*END WAITFLAG STORE'),nl.
			%my_portray_avl(Heap),nl.
portray_waitflags(none) :- !, print('*WAITFLAG STORE:  none'),nl.
portray_waitflags(no_wf_available) :- !, print('*WAITFLAG STORE:  none'),nl.
portray_waitflags(Store) :-
   add_internal_error('Illegal Waitflag: ',portray_waitflags(Store)),flush_output.

portray_fd_vars([]) :- !, '$NOT_COVERED'('This is debugging code!').
portray_fd_vars([V|T]) :- print('> '), print(V), nonvar(V),!, nl,
   portray_fd_vars(T), '$NOT_COVERED'('This is debugging code!').
portray_fd_vars([V|T]) :- my_get_fd_size(V,Size), !, print(' : '), print(Size),
   portray_attached_goals(V),
   portray_fd_vars(T),
   '$NOT_COVERED'('This is debugging code!').
portray_fd_vars(X) :- print('> *** UNKNOWN FDVARS: '), print(X), nl,
   '$NOT_COVERED'('This is debugging code!').

portray_waitflags_store(Store) :-
            my_get_fdvars_atts(Store,FDList),
            (FDList=[] -> true ; print('FDVARS:'),nl,portray_fd_vars(FDList),nl),
			my_get_wf_store_atts(Store,Heap),
			avl_to_list(Heap,List), portray_avl_els(List).

portray_avl_els([]).
portray_avl_els([Prio-wf(LWF,Info)|T]) :-
  print('> '),print(Prio), print(' : '), print(LWF), print(' : '), portray_info(Info),nl,
  portray_attached_goals(LWF),
  portray_avl_els(T).

portray_info(label_clpfd_vars(Vars)) :- print('label_clpfd_vars :: '),
        translate:l_print_frozen_info(Vars).
portray_info(I) :- print(I).

portray_attached_goals(V) :- V==[],!.
portray_attached_goals(LWF) :- nonvar(LWF),LWF=[H|T],!,
   portray_attached_goals(H),portray_attached_goals(T).
portray_attached_goals(LWF) :- frozen(LWF,Goal), print('   :: '), portray_goal(Goal),nl.

:- use_module(tools_printing,[print_term_summary/1]).
portray_goal((A,B)) :- !, portray_goal(A), print(','), print('     '), portray_goal(B).
portray_goal(A) :- print_term_summary(A).

:- public my_portray_avl/1.
my_portray_avl(V) :- var(V), !, add_internal_error('Variable: ',my_portray_avl(V)).
my_portray_avl(V) :- portray_avl(V).


quick_portray_waitflags(wfx(WF0,Store,WFE,_)) :-
      my_get_fdvars_atts(Store,FDList),
      my_get_wf_store_atts(Store,Heap),
			avl_domain(Heap,List),
			format('% Waitflags-Store WF0=~w,WFE=~w, Flags=~w, FDVars=~w~n',[WF0,WFE,List,FDList]).
quick_portray_waitflags_store(Store) :-
      my_get_fdvars_atts(Store,FDList),
      my_get_wf_store_atts(Store,Heap),
			avl_domain(Heap,List),
			format('% Waitflags-Store Flages=~w, FDVars=~w~n',[List,FDList]).

:- assert_must_succeed(( kernel_waitflags:init_wait_flags(WF),
 kernel_waitflags:get_wait_flag1(WF,WF1), when(ground(WF1),X=a),
 kernel_waitflags:ground_wait_flags(WF),  X==a )).

% create a copy of the waitflags with WF0 and WF0 as variable; should be finished with copy_wf_finish
% can be used if you temporarily want to disable non-deterministic enumeration, ...
copy_wf_start(wfx(_,WF,_,Infos),wfx(_,WF,_,Infos)).
copy_wf_start(no_wf_available,no_wf_available).

% just copy WF0 and WFE over to new waitflag
copy_wf_finish(wfx(WF0,_,WFE,_),wfx(WF0,_,WFE,_)).
copy_wf_finish(no_wf_available,no_wf_available).


/* create a copy where the enumeration_finished waitflag is shared;
  the inner enumeration_finished wait flags should normally not be grounded */
%create_inner_wait_flags(wfx(_,_,WFE),wfx(WF0Inner,S,WFE)) :- !,init_wait_flags_store(S). % old version
create_inner_wait_flags(wfx(_,_,WFE,Infos),Info,wfx(WF0Inner,S,WFEInner,Infos)) :-
   init_wait_flags_store(S),
   get_inner_enumeration_over_wait_flag(wfx(WF0Inner,S,WFEInner,Infos),Info,LastWF),
   copy_wfe_to_inner(LastWF,WFE,WFEInner). % delay copying WFE at least until inner enumeration has been run
create_inner_wait_flags(no_wf_available,_,no_wf_available).
:- block copy_wfe_to_inner(-,?,?).
copy_wfe_to_inner(_,WFE,WFE).

ground_inner_wait_flags(wfx(WF0,Store,_EWF,_)) :- !,
 % does not ground enumeration finished waitflag
   grd_wf0(WF0),
   ground_waitflags_store_clpfd(Store).
ground_inner_wait_flags(no_wf_available) :-  !.
ground_inner_wait_flags(W) :-
   add_internal_error('Inner waitflags in wrong format: ',ground_inner_wait_flags(W)).

% copy WF0 and WFE, set up a new store for the rest
copy_wf01e_wait_flags(wfx(WF0,WFX,WFE,Infos),wfx(WF0,S,WFE,Infos)) :-
   init_wait_flags_store(S),
   get_waitflag_from_store(1,copy_wf01e,WFX,WF1), % now ground wait flag 1 if it is grounded in the orginal WF store
   copy_wf01e_wait_flags_aux(WF1,S).
copy_wf01e_wait_flags(no_wf_available,no_wf_available).

:- block copy_wf01e_wait_flags_aux(-,?).
copy_wf01e_wait_flags_aux(_,S) :- ground_waitflags_store_up_to_no_clpfd(1,S). % no CLPFD - labeling required until prio 1


clone_wait_flags_from1(wfx(_WF0,S,WFE,Infos),wfx(_NewWF0,S,WFE,Infos)).
clone_wait_flags_from1(no_wf_available,no_wf_available).
% creates a copy of the wait_flags, except for WF0; this needs to be grounded separately with grd_wf0/1
clone_wait_flags_from1_finish(wfx(WF0,_,_,_),wfx(WF0,_,_,_)). % finish copying by copying over WF0
clone_wait_flags_from1_finish(no_wf_available,no_wf_available).

get_wait_flag0(wfx(WF0,_,_,_),WF0).
get_wait_flag0(no_wf_available,0).

%get_wait_flag0(WFX,WF0) :-   get_wait_flag(0,WFX,WF0).  /* when this is ground you can do det. propagatons */
get_wait_flag1(WFX,WF1) :-   get_wait_flag(1.0,WFX,WF1). /* when this is ground you can do det. propagatons */
get_wait_flag1(Info,WFX,WF1) :-   get_wait_flag(1.0,Info,WFX,WF1).

get_binary_choice_wait_flag(Info,WFX,WF2) :- !, get_binary_choice_wait_flag_exp_backoff(4,Info,WFX,WF2).
%get_binary_choice_wait_flag(Info,WFX,WF2) :- % use 3 rather than 2 to give priority over enumeration choicepoints
%  get_wait_flag(4,Info,WFX,WF2). /* when this is ground you can do binary propagatons */

get_middle_wait_flag(Info,WFX,WFn) :- get_wait_flag(10,Info,WFX,WFn).

get_last_wait_flag(Info,WFX,Res) :- last_finite_priority(P),get_wait_flag(P,Info,WFX,Res).

% ------------------------
get_wait_flag_infos(wfx(_,_,_,Infos),Infos).
get_wait_flag_infos(no_wf_available,[]).

:- use_module(library(ordsets),[ord_member/2,ord_add_element/3]).
get_wait_flag_info(wfx(_,_,_,Infos),Info) :- member(Info,Infos).

% use if the Info field is known, requires ordered/sorted Infos list
is_wait_flag_info(wfx(_,_,_,Infos),Info) :-
  (nonvar(Infos) -> ord_member(Info,Infos) ; print(var_waitflag),nl,fail).

%set_wait_flag_infos(wfx(WF0,Store,WFE,_),Infos,wfx(WF0,Store,WFE,Infos)).
%set_wait_flag_infos(no_wf_available,_,no_wf_available).

add_wait_flag_info(wfx(WF0,Store,WFE,Infos),Info,wfx(WF0,Store,WFE,NewInfos)) :-
   ord_add_element(Infos,Info,NewInfos). %, print(added_wf_infos(NewInfos)),nl.
add_wait_flag_info(no_wf_available,_,no_wf_available).

% ------------------------

%  BINARY CHOICE WAITFLAGS

:- use_module(tools,[max_tagged_pow2/1]).
% last power of 2 exponent that is still a finite priority
last_finite_pow2_prio_exponent(Lim) :- tools:max_tagged_pow2(Pow),
   Lim is Pow-1. % stay below last_finite_priority, i.e., max_tagged_integer-1024

% utility to get power of two priority:
get_pow2_binary_choice_priority(Exp,Prio) :- last_finite_pow2_prio_exponent(Lim),
  (Exp =< Lim -> Prio is integer(2**Exp)
   ; Prio is integer(2**Lim)).

% get a binary choice wait flag, start with 2 but for every further call double the priority (exponential backoff)
% idea: give priority to data enumerations when too many choice points arise
get_binary_choice_wait_flag_exp_backoff(Info,WFX,WF2) :-
    get_binary_choice_wait_flag_exp_backoff(2,Info,WFX,WF2).
get_binary_choice_wait_flag_exp_backoff(_,_Info,no_wf_available,WF2) :- !, WF2=2.

% StartPrio should be a power of 2:
get_binary_choice_wait_flag_exp_backoff(StartPrio,Info,wfx(_,Store,_,_),WF2) :-
	my_get_wf_store_atts(Store,Heap),
	last_finite_pow2_prio_exponent(Lim),
	get_bin_aux(StartPrio,Lim,Info,Heap,NewHeap,WF2),
	put_atts(Store,+wf_store(NewHeap)).

% to do : maybe store attribute of current exponential ??
get_bin_aux(Prio,Lim,Info,Heap,NewHeap,WF2) :- avl_fetch(float(Prio),Heap,wf(WFs,OldInfo)),
    !,
    (Prio >= Lim -> WF2 = WFs, NewHeap=Heap  /* for finite priorities single waitflag stored */
     ; OldInfo = '$binary'(_) -> P2 is Prio*2, get_bin_aux(P2,Lim,Info,Heap,NewHeap,WF2)
     ; otherwise -> WF2=WFs, avl_store(float(Prio),Heap,wf(WFs,'$binary'(Info)),NewHeap) % just update Info; so that next time we double
     ).
get_bin_aux(Prio,_,Info,Heap,NewHeap,WF2) :- % we have found an unused power of 2
   %print(get_bin(Prio,Info)),nl,
   avl_store(float(Prio),Heap,wf(WF2,'$binary'(Info)),NewHeap).

% ------------------------

get_enumeration_finished_wait_flag(wfx(_,_,WFE,_),WFE) :- !.
get_enumeration_finished_wait_flag(no_wf_available,WFE) :-  !,WFE=1.
get_enumeration_finished_wait_flag(W,E) :-
   add_internal_error('Waitflags in wrong format: ',get_enumeration_finished_wait_flag(W,E)).


ground_det_wait_flag(wfx(WF0,Store,_,_)) :- !, %% print(ground_det_wait_flag),nl,
     grd_wf0(WF0),
     ground_waitflags_store_up_to_no_clpfd(1,Store). % no CLPFD labeling required for prio until 1
ground_det_wait_flag(no_wf_available) :-  !.
ground_det_wait_flag(W) :-
    add_internal_error('Waitflags in wrong format: ',ground_det_wait_flag(W)).

ground_wait_flag_to(wfx(WF0,Store,_,_),Limit) :- !,%% print(ground_det_wait_flag),nl,
     grd_wf0(WF0),
     ground_waitflags_store_clpfd_up_to(Limit,Store).
ground_wait_flag_to(no_wf_available,_) :-  !.
ground_wait_flag_to(W,Limit) :-
    add_internal_error('Waitflags in wrong format: ',ground_wait_flag_to(W,Limit)).

ground_waitflags_store_up_to_no_clpfd(Prio,Store) :- % print(prio(Prio)),nl,
	my_get_wf_store_atts(Store,Heap),  % portray_waitflags_store(Store), %
	( avl_min(Heap,Min,wf(WF,Info)) -> % heap is non-empty, first element is WF with priority Min
	    ( Min =< Prio ->          % we have non-ground waitflags with a matching priority
		((avl_del_min(Heap,AMin,_,NewHeap),AMin==Min)     % remove the first element
		 -> true ; add_internal_error('Could not remove min. in ground_waitflags_store_up_to_no_clpfd: ',Min:Info:Heap), fail
		),
		put_atts(Store,+wf_store(NewHeap)),   % write the modified heap back
			 	%%  print(grounding2(Prio,_Info)),nl, %%
		ground_prio(Min,WF,Info),
		ground_waitflags_store_up_to_no_clpfd(Prio,Store)         % continue recursively
	    ; /* otherwise -> */            % no more matching waitflags present
		true)                 % stop here
	; /* otherwise -> */               % heap is empty, nothing more to do
	    true).                    % stop here

ground_waitflags_store_clpfd_up_to(Limit,Store) :-
			my_get_wf_store_atts(Store,Heap),
			my_get_fdvars_atts(Store,FDVars),!,
			( avl_del_min(Heap,Prio,wf(WF,Info),NewHeap) ->          % we have non-ground waitflags with a matching priority
			 ((my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,RemainingFDVARS),
			   fd_priority_leq_limit(NextFDVar,FDPrio,Limit))
			   -> (FDPrio > Limit -> true
			         ; enum_fd_variable_store(FDPrio,Store,NextFDVar,RemainingFDVARS))
			   ;  (Prio > Limit -> true
			       ; put_atts(Store,+wf_store(NewHeap)),   % write the modified heap back
			         %print(grounding(Prio,_Info)),nl,
				     ground_prio(Prio,WF,Info),                            % ground waitflag
			         ground_waitflags_store_clpfd_up_to(Limit,Store)         % continue recursively
			      )
	          )
			; ((my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,RemainingFDVARS),
	             fd_priority_leq_limit(NextFDVar,FDPrio,Limit))
			      -> enum_fd_variable_store(FDPrio,Store,NextFDVar,RemainingFDVARS)
			      ; true  % heap and FDVars list is empty, nothing more to do
			    )
			).
ground_waitflags_store_clpfd_up_to(Limit,Store) :-
   add_internal_error('Illegal Waitflag: ',ground_waitflags_store_clpfd_up_to(Limit,Store)),fail.


update_waitflag(Prio,CurrentWaitflag,NewWaitFlag,WF) :-
  /* if the CurrentWaitflag is already ground and now a new WF with lower priority has been added to the store:
      create a new non-ground waitflag to give priority to new WF with lower priority value */
   (var(CurrentWaitflag) -> NewWaitFlag=CurrentWaitflag
     ; WF=wfx(_WF0,Store,_,_),my_get_wf_store_atts(Store,Heap),
       ((avl_min(Heap,Min,wf(_NewerWF,_Info)),(Prio=inf;Min<Prio))
         -> get_waitflag_from_store(Prio,updated_wf,Store,NewWaitFlag)  % could be optimized
           % ,print(updating_wf(Prio,CurrentWaitflag,NewWaitFlag,Min)),nl
         ;  fdvar_with_higher_prio_exists(Store,Prio) -> get_waitflag_from_store(Prio,updated_wf,Store,NewWaitFlag)
         ;  NewWaitFlag=CurrentWaitflag
       )
   ).

fdvar_with_higher_prio_exists(Store,Limit) :-
   my_get_fdvars_atts(Store,FDVars),
   my_get_next_fdvar_to_enumerate(FDVars,NextFDVar,_),
   fd_priority_leq_limit(NextFDVar,_,Limit).

% get the minimum waitflag, if it exists
get_minimum_waitflag_prio(wfx(_,Store,_,_),MinPrio,Info) :-
    my_get_wf_store_atts(Store,Heap),
    avl_min(Heap,MinPrio,wf(_LWF,Info)).


% currently not used:
%/* same as get_enumeration_finished_wait_flag, but will fail if user does not want
%   to look for abort errors */
%get_abort_error_flag(wfx(_,_,WFE,_),WFE) :- !, preferences:preference(find_abort_values,true).
%get_abort_error_flag(W,_) :-
%   add_internal_error(get_abor_error_flag,'Waitflags in wrong format: ',W).


add_wd_error(Msg,Term,WF) :- add_wd_error_span(Msg,Term,unknown,WF).
add_wd_error_set_result(Msg,Term,Result,ResultValue,Span,WF) :-
   is_wd_guarded_result(Result),  % just dummy co-routine to detect variables which have WD assignments pending on them; should not be enumerated
   % add well-definedness error but also set Result Variable to ResultValue to trigger pending co-routines if wd_guarded
   add_abort_error7(well_definedness_error,Msg,Term,Result,ResultValue,Span,WF).
add_wd_error_span(Msg,Term,Span,WF) :- add_abort_error_span(well_definedness_error,Msg,Term,Span,WF).

:- block is_wd_guarded_result(-).
is_wd_guarded_result(_).

add_abort_error_span(ErrType,Msg,Term,Span,WF) :- add_abort_error7(ErrType,Msg,Term,0,0,Span,WF).

add_abort_error7(_ErrType,_Msg,_Term,_Result,_ResultValue,_Span,_WF) :-
    preference(disprover_mode,true),
    !, % When Disproving we can assume well-definedness of the PO; sometimes the relevant hypotheses will be filtered out
    fail.
add_abort_error7(ErrType,Msg,Term,Result,ResultValue,Span,WF) :-
    get_enumeration_finished_wait_flag(WF,AWF),
    add_abort_error2(AWF,ErrType,Msg,Term,Result,ResultValue,Span).

% no longer used:
%add_wd_error_span_now(Msg,Term,Span) :- add_abort_error2(true,well_definedness_error,Msg,Term,0,0,Span).

:- use_module(state_space,
  [store_abort_error_for_context_state_if_possible/4]).
:- block add_abort_error2(-,?,?,?,?,?,?).
add_abort_error2(wd_guarded,_Err,_Msg,_Term,Result,ResultValue,_Span) :- !,% ignoring error message; no-wd problem
  % set Result to default ResultValue; useful to get rid of pending co-routines; result will not be used
  (Result=ResultValue -> true ; print(add_abort_error_failure(Result,ResultValue)),nl).
add_abort_error2(_AWF,ErrType,Msg,Term,_,_,Span) :-  % print(adding_abort_error(Msg)),nl,trace,
  % add_error(abort_error,Msg,Term),
    (store_abort_error_for_context_state_if_possible(ErrType,Msg,Term,Span)
      -> true ; add_error(add_abort_error2,'Could not store error: ',(Msg,Term),Span)),
    % print(done),nl,
    fail.

/*
ground_wait_flags_catch_enumeration_warnings(WF) :-
    on_exception(enumeration_warning(A,B,C,D,E),
                 ground_wait_flags(WF),
                 (pending_abort_error(WF,Msg,Term,Span),
                  format('WD-Error may be causing Enumeration Warning : ~w~n',[Msg]),
                  fail
                  ;
                  throw(enumeration_warning(A,B,C,D,E)))).
            */

% succeed once for every pending abort error in Waitflag store
pending_abort_error(wfx(_WF0,_WFX,WFE,_),Msg,Term,Span) :-
  frozen(WFE,Goal), %print(goal(Goal)),nl,
  pending_abort_error_aux(Goal,Msg,Term,Span).

pending_abort_error_aux(kernel_waitflags:add_abort_error2(_,_ErrType,Msg,Term,_Result,_ResultValue,Span), Msg,Term,Span).
pending_abort_error_aux(add_abort_error2(_,_ErrType,Msg,Term,_Result,_ResultValue,Span), Msg,Term,Span).
pending_abort_error_aux((A,B),Msg,Term,Span) :-
   (pending_abort_error_aux(A,Msg,Term,Span) ;
    pending_abort_error_aux(B,Msg,Term,Span) ).

% ---------------------------------------------

ground_wait_flag_to_enumeration(WF) :-
   last_finite_priority(P),
   ground_wait_flag_to(WF,P).

get_large_finite_wait_flag(Info,WFX,WF2) :-
    large_finite_priority(P),
    get_wait_flag(P,Info,WFX,WF2).

get_enumeration_starting_wait_flag(Info,WFX,WF2) :-
    last_finite_priority(P),
    get_wait_flag(P,enumeration_starting(Info),WFX,WF2).

get_inner_enumeration_over_wait_flag(WFX,Info,WF2) :-
   tools:max_tagged_integer(P), % a bit of a hack; what if inf_type_prio generates something higher ? TO DO : bound get_max_cardinality_as_priority/inf_type_prio result
   get_wait_flag(P,inner_enumeration(Info),WFX,WF2).

% unused:
%get_integer_enumeration_wait_flag(Info,WFX,WF2) :-
%  integer_priority(Prio),
%  get_wait_flag(Prio,Info,WFX,WF2).

% A few hardcoded priorities
large_finite_priority(P) :- integer_priority(I), P is I-1000.
last_finite_priority(P) :- integer_priority(I), P is I-2. % Note: enumerate_integer_with_min_range_wf did set its priority to I-1
%integer_priority(10000000). % old value
integer_priority(Prio) :-  % priority for X : INTEGER enumerations
  tools:max_tagged_integer(P),
  Prio is P - 1024.
  % note inf_type_prio sets some priorities such as seq(INTEGER) ... higher

% priority for sets of infinite type (POW(INTEGER)...)
integer_pow2_priority(Prio) :-  % was 10000010
   integer_priority(P), Prio is P+10.


is_finite_priority(P) :-
   P < 100000, % TO DO: put max_tagged_integer for 32-bit system minus 1026 or so (see above)
   last_finite_priority(LFP),
   P =< LFP.

%integer_pow_priority(10000020) :- tools:max_tagged_integer(MAX).

% ensure that we do not exceed priority of type enumeration or use too big numbers
get_bounded_wait_flag(Prio,Info,WF,LWF) :-
    bound_prio(Prio,P),
    get_wait_flag(P,Info,WF,LWF).

bound_prio(Prio,P) :- number(Prio),
    Prio>268435452,  % 268435455 is max_tagged_integer on 32 bit platforms
    last_finite_priority(MAX),
    Prio>MAX,
    !, P=MAX.
bound_prio(P,P).

% -------------

% copy waitflags from one waitflag store to another
copy_waitflag_store(wfx(WF0,Store1,WFE1,_),wfx(WF0,Store2,WFE2,Info2)) :-
   %print(copy_wf(WF0,Store1,Store2,WFE1,WFE2)),nl,
   my_get_fdvars_atts(Store1,FDList1),
   (FDList1 = [] -> true
     ; my_get_fdvars_atts(Store2,FDList2),
       append(FDList1,FDList2,NewFDList),
       put_atts(Store2,+wf_fdvars(NewFDList))),
   my_get_wf_store_atts(Store1,Heap),
   avl_to_list(Heap,WF_List),
   %print(copying(FDList1,WF_List)),nl,
   maplist(add_wf(wfx(WF0,Store2,WFE2,Info2)),WF_List), % we could do this faster
   clear_store(Store1), Store1=Store2, % ensure that calls which add waitflags now get-redirected to waitflag store 2
   %portray_waitflags(wfx(WF0,Store2,WFE2,Info2)),
   WFE1=WFE2.

clear_store(Store) :- put_atts(Store,-wf_store(_)), put_atts(Store,-wf_fdvars(_)).

add_wf(WF,Prio-wf(WFlag,Info)) :- % purpose: copy waitflag WFlag to WF
   (Prio=float(P) -> true ; P=Prio),
   (var(WFlag) -> get_wait_flag(P,Info,WF,WFlag) % we just have a single waitflag
    ; maplist(get_wait_flag(P,Info,WF),WFlag)). % we have a list of waitflags


% try and find an exact match for a waitflag info; this is linear in size of waitflag store !
find_waitflag_info(wfx(_WF0,Store,_WFE1,_),Info,Prio,WF1) :-
    my_get_wf_store_atts(Store,Heap),
    avl_member(Prio,Heap,wf(WF1,Info)).

% --------------------------

 :- assert_must_succeed(( kernel_waitflags:test_waitflags(1) )).
 :- assert_must_succeed(( kernel_waitflags:test_waitflags(2) )).

test_waitflags(1) :-
	init_wait_flags(Waitflags),
	get_wait_flag(10,Waitflags,WF1),
	when(nonvar(WF1),print('WF1 - Prio 10\n')),
	get_wait_flag(5,Waitflags,WF2),
	when(nonvar(WF2),(print('WF2 - Prio 5\n'),
	       get_wait_flag(6,Waitflags,WF4),
         	when(nonvar(WF4),print('WF4 - Prio 6\n')),
	       get_wait_flag(2,Waitflags,WF5),
	      when(nonvar(WF5),print('WF5 - Prio2\n')))),
        get_wait_flag(5,Waitflags,WF3),
	when(nonvar(WF3),print('WF3 - Prio 5\n')),
	ground_wait_flags(Waitflags),nl.

test_waitflags(2) :-
  init_wait_flags(Waitflags),
  get_wait_flag(10,Waitflags,WF1),
  when(nonvar(WF1),print('WF1 Prio 10\n')),
  get_wait_flag(5,Waitflags,WF2),
   when(nonvar(WF2),(print('WF2 - Prio 5\n'),
   get_wait_flag(6,Waitflags,WF4),
 	when(nonvar(WF4),print('WF4 - Prio 6\n')),
   get_wait_flag(2,Waitflags,WF5),
  when(nonvar(WF5),print('WF5 - Prio2\n')))),
  get_wait_flag(5,Waitflags,WF3),
  when(nonvar(WF3),print('WF3 - Prio 5\n')),
  ground_wait_flags(Waitflags),nl.
