% (c) 2016-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(meta_interface,[command_description/3, command_call/4,
                          command_unavailable/2,
                          command_arguments/2, command_optional_arguments/4,
                          command_category/2,
                          command_preferences_category/2,
                          command_additional_info/2,
                          get_command_preferences/2,
                          call_command/5,
                          is_dot_command/1, call_dot_command/2, call_dot_command/3,
                          is_dot_command_for_expr/1, call_dot_command_for_expr/3, call_dot_command_for_expr/4]).

:- use_module(module_information).
:- module_info(group,animator).
:- module_info(description,'This module provides access and description to a variety of Prolog commands.').

:- use_module(error_manager).
:- use_module(debug).
:- use_module(tools,[ajoin/2]).
:- use_module(specfile,[animation_mode/1, animation_minor_mode/1,
                    b_mode/0, z_mode/0, b_or_z_mode/0, b_or_z_mode/1, eventb_mode/0,
                    csp_mode/0, csp_mode/1,
                    csp_with_bz_mode/0, z_or_tla_minor_mode/0,
                    process_algebra_mode/0,
                    spec_file_has_been_successfully_loaded/0]).
:- use_module(state_space, [current_state_corresponds_to_fully_setup_b_machine/0]).

% dot command with no arguments
is_dot_command(Command) :- command_category(Command,dot), command_arguments(Command,[dotfile]).
call_dot_command(Command,File) :- call_dot_command(Command,File,[]).
call_dot_command(Command,File,OptionalArgs) :- %print(cmd(Command,File,OptionalArgs)),nl,
   call_command(dot,Command,[dotfile],[File],OptionalArgs).

is_dot_command_for_expr(Command) :- command_category(Command,dot), command_arguments(Command,[expression_atom,dotfile]).
call_dot_command_for_expr(Command,Expr,File) :-
   call_dot_command_for_expr(Command,Expr,File,[]).
call_dot_command_for_expr(Command,Expr,File,OptionalArgs) :-  % print(cmd(Command,Expr,File,OptionalArgs)),nl,
   call_command(dot,Command,[expression_atom,dotfile],[Expr,File],OptionalArgs).

% call a command described in this module
call_command(Category,Command,FormalArgs,ActualArgs,OptionalArgs) :-
   command_category(Command,Category),
   command_arguments(Command,FormalArgs), % to do: more flexible matching of args ?
   !,
   (command_unavailable(Command,Msg)
    -> tools:ajoin(['Command ', Command, ' not available: '],Msg1),
       add_error(meta_interface,Msg1,Msg),
       fail
    ; setup_optional_args(Command,OptionalArgs,OptArgValueList),
      (command_call(Command,ActualArgs,OptArgValueList,Call)
         -> debug_println(4,command_call(Call)),
            call(Call)
        ;   add_internal_error('Cannot construct command call: ',command_call(Command,ActualArgs,OptArgValueList,_)),
            fail
      )
    ).
call_command(Category,Command,_FormalArgs,_ActualArgs,_OptArgs) :- command_category(Command,Category),
    !,
    add_error(meta_interface,'Argument mis-match when calling command: ',Command),fail.
call_command(Category,Command,_,_,_) :-
    add_error(meta_interface,'Unknown command: ',Category:Command),fail.


setup_optional_args(Command,OptionalArgs,OptArgValueList) :-
     command_optional_arguments(Command,Names,_Types,DefaultValues),!,
     setup_opt_args_aux(Names,DefaultValues,OptionalArgs,Command,OptArgValueList).
setup_optional_args(_,[],Res) :- !, Res=[].
setup_optional_args(Cmd,Opt,Res) :- add_internal_error('Illegal optional arguments: ',setup_optional_args(Cmd,Opt,Res)),
   Res=[].

:- use_module(library(lists),[select/3]).

setup_opt_args_aux([],[],OptArgs,Command,Res) :-
    (OptArgs = [] -> true ; add_internal_error('Too many optional arguments: ',Command:OptArgs)),
    Res = [].
setup_opt_args_aux([Name|NT],[Default|DT],OptArgs,Command,Res) :-
    (select('='(Name,Value),OptArgs,OA2)
      -> Res = [Value|RT], setup_opt_args_aux(NT,DT,OA2,Command,RT)
      ;  Res = [Default|RT], setup_opt_args_aux(NT,DT,OptArgs,Command,RT)
    ).


% TO DO: provide automatic conversion ?  expression_atom -> expression_codes -> typed_expression ?

:- use_module(preferences,[virtual_preference_category/2]).
% get all preferences for a command
get_command_preferences(Command,Preferences) :-
   findall(Pref, (command_preferences_category(Command,PrefCat),
                  virtual_preference_category(Pref,PrefCat)), Preferences).

% --------------------
% COMMAND DESCRIPTIONS
% --------------------

:- discontiguous command_description/3, command_call/4,
                 command_unavailable/2, command_arguments/2,
                 command_optional_arguments/4,
                 command_category/2, command_preferences_category/2,
                 command_additional_info/2.

:- use_module(b_machine_hierarchy,[write_dot_event_hierarchy_to_file/1,write_dot_hierarchy_to_file/1]).

command_description(machine_hierarchy,'Machine Hierarchy','Shows the machine hierarchy of a classical B model').
command_call(machine_hierarchy,[File],[],b_machine_hierarchy:write_dot_hierarchy_to_file(File)).
command_unavailable(machine_hierarchy,'only available for B,Z or Event-B models') :- \+ b_or_z_mode.
command_arguments(machine_hierarchy,[dotfile]).
command_category(machine_hierarchy,dot).


command_description(event_hierarchy,'Event Hierarchy','Shows the event hierarchy of an Event-B model').
command_call(event_hierarchy,[File],[],b_machine_hierarchy:write_dot_event_hierarchy_to_file(File)).
command_unavailable(event_hierarchy,'only available for Event-B models') :- \+ eventb_mode.
command_arguments(event_hierarchy,[dotfile]).
command_category(event_hierarchy,dot).
command_preferences_category(event_hierarchy,dot_event_hierarchy).


:- use_module(visualize_graph,[tcltk_print_states_for_dot/2]).
command_description(state_space,'State Space','Show state space').
command_call(state_space,[File],[ColourTransitions],visualize_graph:tcltk_print_states_for_dot(File,ColourTransitions)).
command_unavailable(state_space,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(state_space,[dotfile]).
command_optional_arguments(state_space,[colouring],[atom],[no_colour_for_transitions]).
command_category(state_space,dot).
command_preferences_category(state_space,dot_state_space).

command_description(state_space_sfdp,'State Space (Fast)','Show state space (fast)').
command_call(state_space_sfdp,[File],[ColourTransitions],visualize_graph:tcltk_print_states_for_dot_sfdp(File,ColourTransitions)).
command_unavailable(state_space_sfdp,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(state_space_sfdp,[dotfile]).
command_optional_arguments(state_space_sfdp,[colouring],[atom],[no_colour_for_transitions]).
command_category(state_space_sfdp,dot).
command_preferences_category(state_space_sfdp,dot_state_space). % TODO: maybe remove those that are fixed and have no effect
command_additional_info(state_space_sfdp,sfdp_preferred). % sfdp preferred

:- use_module(state_space_reduction,[write_transition_diagram_for_expr_to_dotfile/2,
                                     write_signature_merge_to_dotfile/2]).


command_description(signature_merge,'Signature Merge','Show signature-merged reduced state space').
command_call(signature_merge,[File],[IgnoredEvents],state_space_reduction:write_signature_merge_to_dotfile(IgnoredEvents,File)).
command_unavailable(signature_merge,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(signature_merge,[dotfile]).
command_optional_arguments(signature_merge,[ignored_events],[list(operations)],[[]]).
command_category(signature_merge,dot).


:- use_module(reduce_graph_state_space,[print_dot_for_dfa_from_nfa/1]).
command_description(dfa_merge,'DFA Merge','Show state space as deterministic automaton (DFA)').
command_call(dfa_merge,[File],[],reduce_graph_state_space:print_dot_for_dfa_from_nfa(File)).
command_unavailable(dfa_merge,'only available when model successfully loaded') :- \+ spec_file_has_been_successfully_loaded.
command_arguments(dfa_merge,[dotfile]).
command_category(dfa_merge,dot).

:- use_module(graph_canon,[print_cstate_graph/1]).
command_description(state_as_graph,'State Graph','Show values in current state as a graph').
command_call(state_as_graph,[File],[],graph_canon:print_cstate_graph(File)).
command_unavailable(state_as_graph,'only available for B,Z or Event-B models') :- \+ b_or_z_mode.
command_arguments(state_as_graph,[dotfile]).
command_category(state_as_graph,dot).
% we could use tcltk_print_current_state_as_graph_for_dot

% from tcltk_interface.pl:
command_description(expr_as_graph,'Expression Graph','Show expression value as a graph').
command_call(expr_as_graph,[Expression,File],[],user:tcltk_show_expression_as_dot(Expression,File)).
command_unavailable(expr_as_graph,'only available for initialised B,Z or Event-B models') :- \+ current_state_corresponds_to_fully_setup_b_machine.
command_arguments(expr_as_graph,[expression_atom,dotfile]).
command_category(expr_as_graph,dot).

command_description(formula_tree,'Formula Tree','Show predicate and sub-expressions as a tree').
command_call(formula_tree,[Expression,File],[],user:generate_dot_from_pred(Expression,File)).
command_unavailable(formula_tree,'only available for initialised B,Z or Event-B models') :- \+ current_state_corresponds_to_fully_setup_b_machine.
command_arguments(formula_tree,[expression_atom,dotfile]). % is actually predicate_atom
command_category(formula_tree,dot).

command_description(invariant,'Invariant Formula Tree','Show invariant as a formula tree').
command_call(invariant,[File],[],user:generate_dot_from_invariant(File)).
command_unavailable(invariant,'only available for initialised B,Z or Event-B models') :- \+ current_state_corresponds_to_fully_setup_b_machine.
command_arguments(invariant,[dotfile]).
command_category(invariant,dot).

command_description(properties,'Properties Formula Tree','Show properties as a formula tree').
command_call(properties,[File],[],user:generate_dot_from_properties(File)).
command_unavailable(properties,'only available for initialised B,Z or Event-B models') :- \+ b_or_z_mode. % for non-initialised machines an existential quantifier is added
command_arguments(properties,[dotfile]).
command_category(properties,dot).

command_description(assertions,'Assertions Formula Tree','Show assertions as a formula tree').
command_call(assertions,[File],[],user:generate_dot_from_assertions(File)).
command_unavailable(assertions,'only available for initialised B,Z or Event-B models') :- \+ b_or_z_mode. % for non-initialised machines an existential quantifier is added
command_arguments(assertions,[dotfile]).
command_category(assertions,dot).

command_description(deadlock,'Deadlock Formula Tree','Show deadlocking status as a formula tree').
command_call(deadlock,[File],[],user:generate_dot_from_deadlock_po(File)).
command_unavailable(deadlock,'only available for initialised B,Z or Event-B models') :- \+ current_state_corresponds_to_fully_setup_b_machine.
command_arguments(deadlock,[dotfile]).
command_category(deadlock,dot).

command_description(goal,'Goal Formula Tree','Show GOAL as a formula tree').
command_call(goal,[File],[],user:generate_dot_from_goal(File)).
command_unavailable(goal,'only available for initialised B,Z or Event-B models') :- \+ b_or_z_mode.
command_arguments(goal,[dotfile]).
command_category(goal,dot).

command_description(last_error,'Last Error Formula Tree','Show last error source as a formula tree').
command_call(last_error,[File],[],user:generate_dot_from_last_span_predicate(File)).
command_unavailable(last_error,'only available when error occured') :- \+ user:can_generate_dot_from_last_state_error.
command_arguments(last_error,[dotfile]).
command_category(last_error,dot).

:- use_module(state_space_reduction,[write_transition_diagram_for_expr_to_dotfile/2]).
command_description(transition_diagram,'Transition Diagram Projection','Project state space onto expression values').
command_call(transition_diagram,[Expression,File],[],
             state_space_reduction:write_transition_diagram_for_expr_to_dotfile(Expression,File)).
command_unavailable(transition_diagram,'only available for B,Z or Event-B models') :- \+ b_or_z_mode.
command_arguments(transition_diagram,[expression_atom,dotfile]).
command_category(transition_diagram,dot).
command_preferences_category(transition_diagram,dot_projection).

:- use_module(probporsrc(dot_graphs_static_analysis), [tcltk_create_dependence_graph/1, tcltk_create_enable_graph/1]).
command_description(enable_graph,'Enable Graph','Show enabling graph of events').
command_call(enable_graph,[File],[],dot_graphs_static_analysis:tcltk_create_enable_graph(File)).
command_unavailable(enable_graph,'only available for B,Z or Event-B models') :- \+ b_or_z_mode.
command_arguments(enable_graph,[dotfile]).
command_category(enable_graph,dot).

command_description(dependence_graph,'Dependence Graph','Show dependence graph of events').
command_call(dependence_graph,[File],[],dot_graphs_static_analysis:tcltk_create_dependence_graph(File)).
command_unavailable(dependence_graph,'only available for B,Z or Event-B models') :- \+ b_or_z_mode.
command_arguments(dependence_graph,[dotfile]).
command_category(dependence_graph,dot).

:- use_module(visualize_graph,[tcltk_print_definitions_as_graph_for_dot/1]).
command_description(definitions,'Definitions Graph','Show dependence graph of DEFINITIONS').
command_call(definitions,[File],[],visualize_graph:tcltk_print_definitions_as_graph_for_dot(File)).
command_unavailable(definitions,'only available for B,Z or Event-B models') :- \+ b_or_z_mode.
command_arguments(definitions,[dotfile]).
command_category(definitions,dot).
command_preferences_category(definitions,dot_definitions).

:- use_module(visualize_graph,[tcltk_print_predicate_dependency_as_graph_for_dot/2]).
command_description(predicate_dependency,'Predicate Dependency Graph','Show dependence graph of conjuncts of a predicate').
command_call(predicate_dependency,[Pred,File],[],visualize_graph:tcltk_print_predicate_dependency_as_graph_for_dot(Pred,File)).
command_unavailable(predicate_dependency,'only available for B,Z or Event-B models') :- \+ b_or_z_mode.
command_arguments(predicate_dependency,[expression_atom,dotfile]).
command_category(predicate_dependency,dot).
