% (c) 2009-2015 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(pathes,[
                  set_search_pathes/0,       % install run-time search pathes
                  runtime_application_path/1
                 ]).

:- use_module(module_information).

:- module_info(group,infrastructure).
:- module_info(description,'This module contains code to compute compile-time and runtime-pathes.').

:- use_module(library(system),[environ/2]).
:- use_module(library(file_systems),[file_exists/1,
                                     current_directory/1]).
:- use_module(library(lists),[reverse/2]).

:- multifile user:file_search_path/2.
:- dynamic user:file_search_path/2.

% some pathes may be set by environment variables during compilation,
:- dynamic compile_time_env_path/2.
lookup_env_path(Pathname, Varname) :-
    ( environ(Varname,Value) ->
        print('Hard-wired path for alias '),print(Pathname),print(': '),print(Value),nl,
        assert(compile_time_env_path(Pathname,Value))
    ; otherwise ->
        true).
% compile-time pathes
:- lookup_env_path(prob_comp_home,'PROB_COMP_HOME'). % hard-wired version of the runtime_application_path
                                                     % This is used on systems where it absolutely clear where the
                                                     % ProB's application directory will be, namely Debian/Ubuntu packages
% run-time pathes
:- lookup_env_path(prob_home,'PROB_HOME').
:- lookup_env_path(examples,'PROB_EXAMPLES').        % hard-wired pathes, see above (about Debian/Ubuntu packages)
:- lookup_env_path(prob_lib,'PROB_LIB').             % hard-wired pathes, see above (about Debian/Ubuntu packages)
:- lookup_env_path(prob_tcl,'PROB_TCL').             % hard-wired pathes, see above (about Debian/Ubuntu packages)

% removes an (optional) trailing /src directory from the path,
% the result will not end with a slash
remove_src_dir(Orig,Dir) :-
    name(Orig,COrig),
    ( append(CDir,"/src/",COrig) -> true
    ; append(CDir,"/src",COrig)  -> true
    ; append(CDir,"/src/proz",COrig) -> true
    ; append(CDir,"/src/proz/",COrig) -> true
    ; append(CDir,"/",COrig)     -> true
    ; otherwise -> COrig=CDir),
    name(Dir,CDir).

% returns the path to the application directory (at run-time)
runtime_application_path(Dir) :-
   ( environ('PROB_HOME',Dir) -> % the user has set the environment variable PROB_HOME and thus overrides any other setting
       true                      % the Rodin plugin used this mechanism
   ; application_path2(Dir),is_correct_prob_home_path(Dir) ->
       true
   ; application_path2(D),absolute_file_name(D,D1),parent_dir(D1,Dir,_),
       is_correct_prob_home_path(Dir) ->
       true
   ; compile_time_env_path(prob_comp_home,Dir) ->  % e.g. Debian/Ubuntu systems: the path was hard-wired at compile-time
       true
   ; otherwise ->  % usually a run from source code, where we the current directory is prob/src
       current_directory(Current), remove_src_dir(Current,Dir)).
application_path2(Dir) :-
   \+(prolog_flag(system_type,development)),
   environ('SP_APP_DIR',Dir).  /* /usr/local/bin/sicstus on development systems */

% check if the path points to a directory with a lib subdirectory and the probcliparser.jar file
is_correct_prob_home_path(D) :- atom_concat(D,'/lib/probcliparser.jar',ConParser),
   file_exists(ConParser).

parent_dir(FullFile,Dir,Basename) :-
    atom_chars(FullFile,CFullFile),                   % "path/to/file"
    reverse(CFullFile,RCFullFile),                    % "elif/ot/htap"
    append(RCFile,['/'|RCDir],RCFullFile),!,          % "elif" "/ot/htap"
    reverse(RCFile,CFile),reverse(['/'|RCDir],CDir),  % "file" "path/to/"
    atom_chars(Basename,CFile),atom_chars(Dir,CDir).

% we do not use our standard test case schema here because we do not
% want to introduce a dependency to the test framework at this level
parent_dir_test1 :-
    parent_dir('/aaaa/bbb/cc/d.app',D,B),D=='/aaaa/bbb/cc/',B=='d.app'.
parent_dir_test2 :-
    parent_dir('/aaaa/bbb/cc/',D,B),D=='/aaaa/bbb/cc/',B==''.
parent_dir_test :-
    parent_dir_test1, parent_dir_test2.
:- ( parent_dir_test -> true
   ; otherwise -> write(user_error,'testcase for pathes:parent_dir/2 failed, aborted.\n'),
       halt(1)).

% returns the path to the application directory (compile-time)
compiletime_application_path(Dir) :-
    compile_time_env_path(prob_comp_home,Dir),!.
compiletime_application_path(Dir) :-
    runtime_application_path(Dir).

check_if_hard_wired(Alias, _Prefix, _Dir, Full) :-
    compile_time_env_path(Alias,Full),!.
check_if_hard_wired(_Alias, Prefix, Dir, Full) :-
    atom_concat(Prefix,Dir,Full).

set_search_path(Alias, Prefix, Dir) :-
    check_if_hard_wired(Alias, Prefix, Dir, Full),
    %%  print(setting_path(Alias,Full)),nl, %%
    ( catch( user:file_search_path(Alias,Full), _, fail) -> true
    ; otherwise ->
        catch( retractall(user:file_search_path(Alias,_)), _, true),  % remove any old defintion
        assertz(user:file_search_path(Alias,Full)) % , print(setting_search_path(Alias,Full)),nl
    ).

% pathes that can be used by use_module(...), but
% must not be used in run-time code
set_compile_time_search_pathes :-
    compiletime_application_path(App),
    set_search_path(extension, App, '/extensions'),
    set_search_path(probsrc, App, '/src'),
    set_search_path(chrsrc, App, '/src/chr'),
    set_search_path(probcspsrc, App, '/src/cia'),
    set_search_path(probltlsrc, App, '/src/ltl'),
    set_search_path(probporsrc, App, '/src/por'),
    set_search_path(probpgesrc, App, '/src/pge'),
    set_search_path(prozsrc, App, '/src/proz'),
    set_search_path(kodkodsrc, App, '/src/kodkod'),
    set_search_path(symbolic_model_checker, App, '/src/symbolic_model_checker'),
    set_search_path(partial_evaluator, App, '/src/partial_evaluator'),
    set_search_path(plugins, App, '/plugins'),
    set_search_path(abstract_domains, App, '/plugins/absint/domains'),
    set_search_path(tclsrc, App, '/tcl'),
    set_search_path(probpromela,App,'/src/promela'),
    set_search_path(prob_rewrite_rules, App, '/src/rewrite_rules'),
    set_search_path(smtlib_solver, App, '/src/smtlib_solver'),
    set_search_path(smt_solvers_interface, App, '/src/smt_solvers_interface'),
    set_search_path(synthesis, App, '/src/synthesis').
:- set_compile_time_search_pathes.

% pathes that can be used in run-time code;  should not be called at compile time
set_search_pathes :-
    runtime_application_path(App),
    %print(runtime_application_path(App)),nl,
    set_search_path(prob_home, App, ''),
    set_search_path(examples, App, '/../probprivate/public_examples'),
    set_search_path(prob_lib, App, '/lib'),
    set_search_path(prob_tcl, App, '/tcl'),

    check_if_hard_wired(prob_lib, App, '/lib', LibDir),
    ( catch(user:library_directory(LibDir), _, fail) -> true % ,print(already_set(LibDir)),nl
    ; otherwise ->
        catch( retractall(user:library_directory(_)), _, true), % remove any old defintion
        assertz(user:library_directory(LibDir))).
