% (c) 2009-2017 Lehrstuhl fuer Softwaretechnik und Programmiersprachen,
% Heinrich Heine Universitaet Duesseldorf
% This software is licenced under EPL 1.0 (http://www.eclipse.org/org/documents/epl-v10.html)

:- module(proz, [open_proz_file/2]).

:- use_module(library(lists)).
:- use_module(library(ordsets),[ord_union/3,ord_subtract/3]).

:- use_module(probsrc(debugging_calls)).
:- use_module(probsrc(tools)).
:- use_module(probsrc(self_check)).
:- use_module(probsrc(b_state_model_check)).
:- use_module(probsrc(store)).
:- use_module(probsrc(error_manager)).

:- use_module(probsrc(bmachine_structure)).
:- use_module(probsrc(bsyntaxtree)).
:- use_module(probsrc(b_ast_cleanup),[clean_up/3,clean_up_l_wo_optimizations/4, clean_up_l_with_optimizations/4]).
%:- use_module(probsrc(translate),[print_machine/1]).

:- use_module(z_tools).
:- use_module(fuzzfile).
:- use_module(schemaexpansion).
:- use_module(dependence,[schema_dependence/2]).
:- use_module(zenvironment).
:- use_module(z_typechecker).
:- use_module(ztransformations).
:- use_module(consistencycheck).
:- use_module(subexpressions).

:- use_module(probsrc(module_information)).
:- module_info(group,proz).
:- module_info(description,'This is the central ProZ module for translating Z specification into B.').

/* --------------- FUZZ ----------------- */

reset_proz :-
    reset_fuzzfile,
    retractall(unchanged_vars_for_partial_zoperation(_,_)).

open_proz_file(File,BMachine) :-
    reset_proz,
    load_fuzzfile(File),
    get_z_filename(File,Filename),
    init_fuzz_mode(Filename,BMachine).
%    ( translate:print_machine(BMachine) -> true
%    ; otherwise -> write('pretty-printing the translated B machine failed.\n')).

get_z_filename(File,Filename) :-
    get_tail_filename(File,Filename1),
    ( split_filename(Filename1,Filename,_) -> true
    ; otherwise -> Filename1 = Filename).


/* Initialize all basic types */
init_fuzz_mode(Filename,BMachine) :-
    typed_specification(Dependencies,TypedSpecification),

    /* initialise the z schemas */
    ~~mnf( fuzzschemas(TypedSpecification,Schemas) ),
    ~~mnf( classify_schemas(Schemas, Dependencies, Classification) ),

    /* construct any statements for initialisation and operations */
    ~~mnf( construct_init(Schemas,Classification,InitAny) ),
    ~~mnf( construct_operations(Schemas,Classification,Ops) ),

    ~~mnf( optimize_anys(anys(InitAny,Ops),anys(InitAny2,Ops2)) ),

    ~~mnf( get_proz_settings(Schemas,Classification,Filename,TypedSpecification,Settings) ),

    create_z_machine(Filename,TypedSpecification,InitAny2,Ops2,Schemas,Classification,Settings,BMachine).

:- public fuzzschemas/2, classify_schemas/3, construct_init/3, construct_operations/3,
          optimize_anys/2, get_proz_settings/5. % because Spider cannot deal with ~~mnf

typed_specification(Dependencies,TypedSpecification) :-
    ~~mnf( get_z_definitions(InitialZSpec) ),

    ~~mnf( schema_dependence(InitialZSpec,Dependencies) ),

    /* simplify without having type information */
    ~~mnf( simplify_untyped(InitialZSpec,ZSpecification) ),

    /* add type informations */
    ~~mnf( global_environment(ZSpecification,_,TypedSpecification1) ),

    /* simplify the specification */
    ~~mnf( simplify_typed(TypedSpecification1,TypedSpecification) ).

create_z_machine(Filename, TypedSpecification, InitAny, Ops, Schemas, Classification, Settings, BMachine) :-
    create_machine(Filename,Empty),

    /* register new data types */
    ~~mnf( set_given_sets(TypedSpecification,Empty,MSets) ),
    ~~mnf( set_freetypes(TypedSpecification,MSets,MFreetypes) ),

    /* register initialisation and operations */
    ~~mnf( set_init(InitAny,MFreetypes,MInit) ),
    ~~mnf( set_operations(Ops,MInit,MOps) ),

    ~~mnf( set_variables(Schemas,Classification,MOps,MVars) ),
    /* mnf( register_scope(Schemas,Classification) ), */
    ~~mnf( set_constants(TypedSpecification,MVars,MConstants) ),

    ~~mnf( set_settings(Settings, MConstants, MSettings) ),

    ~~mnf( clean_up_machine(MSettings,BMachine) ).

:- public set_init/3, set_operations/3, set_given_sets/3, set_freetypes/3,
          set_variables/4, set_constants/3, set_settings/3,
          clean_up_machine/2. % because Spider does not deal with ~~mnf

simplify_untyped -->
    zcheck('initial Z specification',[]),

    /* replace simple freetypes with named sets */
    freetype2namedsets, zcheck('after replacing simple freetypes by named sets',[]),

    /* remove the declaration of the ignore pregen */
    remove_ignore_decl, zcheck('after removing proz-ignore declarations',[]),

    /* exchanged simple references to schemas by SEXPR(sref(...)) */
    clearsrefs, zcheck('after clearsrefs',[]),

    /* substitute references to abbreviations with their definition */
    removeabbreviations, zcheck('after removal of abbreviations',[]),

    /* flatten the specification */
    sexpansion, zcheck('after schema expansion',[nlabel(scalc)]).

simplify_typed -->
    /* remove schema expressions */
    removesexpr, ztcheck('after removing schema expressions',[nlabel(scalc)]),

    /* some general simplifications */
    ggsimplify, ztcheck('after some general simplifications',[nlabel(scalc)]).

zcheck(Msg,A,Z,Z) :- check_z_structure(Z,Msg,A).
ztcheck(Msg,A,Z,Z) :- check_z_structuret(Z,Msg,A).

%*******************************************************************************
% optimization of ANY-statements
optimize_anys -->
        exists_to_anys,
    optimize_global_predicates.

optimize_global_predicates(anys(InitAny,OldOps),anys(InitAny,NewOps)) :-
    global_predicates(InitAny,OldOps,GlobalPredicates),
    remove_global_predicates(OldOps,GlobalPredicates,NewOps).

% if the predicate is just of one exists predicate, we
% merge any and exists.
exists_to_anys(anys(OldInitAny,OldOps),anys(NewInitAny,NewOps)) :-
    exists_to_anys2(OldInitAny,NewInitAny),
    exists_to_anys_ops(OldOps,NewOps).
exists_to_anys_ops([],[]).
exists_to_anys_ops([op(Op,Params,Results,OldAny,ParaTypes,ResultTypes)|ORest],
                   [op(Op,Params,Results,NewAny,ParaTypes,ResultTypes)|NRest]) :-
    exists_to_anys2(OldAny,NewAny),
    exists_to_anys_ops(ORest,NRest).
exists_to_anys2(any(Vars,Types,[ti(exists(Body,Pred),bool)],LHS,RHS),NewAny) :-
    !,Body=ti(body(_,Wheres),set(schema(ZFields))),
    append(Wheres,[Pred],NPreds1),
    clean_wheret(NPreds1,NPreds),
    % todo: not safe with name clashes
    zfields2b(ZFields,EVars,ETypes),
    append(Vars,EVars,NVars),
    append(Types,ETypes,NTypes),
    exists_to_anys2(any(NVars,NTypes,NPreds,LHS,RHS),NewAny).
exists_to_anys2(Any,Any).

%*******************************************************************************
/* extract constants, their types and properties from the specification */
set_constants(TypedDefs,In,Out) :-
    write_section(abstract_constants,TypedConstants,In,M),
    write_section(properties,Property,M,Out),
    extract_constants2(TypedDefs, Constants, Types, Properties),
    conjunct_predicates(Properties,Property),
    z_translate_typed_identifiers(Constants,Types,[],TypedConstants).

extract_constants2([],[],[],[]).
extract_constants2([Def|Rest], Constants, Types, Properties) :-
    extract_constants3(Def, LocalConstants, LocalTypes, ZPropertiesTi),
    !,
    z_translate(ZPropertiesTi,LocalProperties),
    append(LocalConstants, RestConstants, Constants),
    append(LocalTypes, RestTypes, Types),
    append(LocalProperties, RestProperties, Properties),
    extract_constants2(Rest, RestConstants, RestTypes, RestProperties).
extract_constants2([_|Rest], Constants, Types, Properties) :-
    extract_constants2(Rest, Constants, Types, Properties).
extract_constants3(Axdef, Constants, Types, Predicates) :-
    ti_expr(Axdef,axdef(BodyTi)),!,
    extract_btypes(Axdef,Constants,Types),
    ti_expr(BodyTi,body(_,Predicates)).
extract_constants3(PredTi, [], [], [Predicate]) :-
    ti_expr(PredTi, pred(Predicate)).


/* appends a decoration to the each variable name of the given list of
   variable names */
/* unused at the moment:
decorate([],_,[]) :- !.
decorate([Var|Vars], Deco, [Decorated|Rest]) :-
    !,
    decorate(Var, Deco, Decorated),
    decorate(Vars, Deco, Rest).
decorate(Var, Deco, Decorated) :-
    z_translate_identifier(name(Var,Deco),Decorated).
*/


remove_vars(Orig,OrigTypes,[],Orig,OrigTypes) :- !.
remove_vars(Orig,OrigTypes,[Var|Rest],ResVars,ResTypes) :-
    remove_var(Orig,OrigTypes,Var,Vars,Types),
    remove_vars(Vars,Types,Rest,ResVars,ResTypes).
remove_var([],[],_,[],[]) :- !.
remove_var([Var|Orig],[_|OrigTypes],Var,ResVars,ResTypes) :-
    remove_var(Orig,OrigTypes,Var,ResVars,ResTypes).
remove_var([Var|Orig],[Type|OrigTypes],OtherVar,[Var|ResVars],[Type|ResTypes]) :-
    Var \= OtherVar,
    remove_var(Orig,OrigTypes,OtherVar,ResVars,ResTypes).

/* Extract all schemas from the fuzz schema definitions */
fuzzschemas([],[]).
fuzzschemas([Def|Rest],FSchemas) :-
    ( schema_id_body(Def,Id,Body) ->
        FSchemas = [fschema(Id, Body, Vars, Types)|FRest],
        extract_btypes(Def,Vars,Types)
    ; otherwise ->
        FSchemas = FRest),
    fuzzschemas(Rest,FRest).

schema_id_body(Ti,Id,Body) :- ti_expr(Ti,Expr),schema_id_body2(Expr,Id,Body).
schema_id_body2(sdef(shead(Id,[]),Body),Id,Body).
schema_id_body2(defeq(shead(Id,[]),Body),Id,Body).

%*******************************************************************************
% Extract settings from specifications and put them into the meta data of
% a machine
get_proz_settings(Schemas,Classification,Filename,TypedSpecification,Settings) :-
    findall(Setting,
            get_proz_setting(Schemas,Classification,Filename,TypedSpecification,Setting),
            Settings).
get_proz_setting(Schemas,Classification,_Filename,TypedSpecification,Setting) :-
    get_schema_by_class(settings,Classification,Schemas,ZDef,ZVars,ZTypes),
    get_proz_expr_setting(ZDef,ZVars,ZTypes,TypedSpecification,Setting).
get_proz_setting(Schemas,Classification,_Filename,_TypedSpecification,goal(BGoal)) :-
    get_schema_by_class(goal,Classification,Schemas,ZDef,_,_),
    ti_expr(ZDef,body(_,ZGoals)),
    z_translate(ZGoals,BGoals1),
    clean_up_l_wo_optimizations(BGoals1,[],BGoals,schemas),
    conjunct_predicates(BGoals,BGoal).
get_proz_setting(_Schemas,_Classification,_Filename,_TypedSpecification,[]).
get_proz_expr_setting(ZDef,_ZVars,_ZTypes,TypedSpecification,animation_function(BAnim,BDefaultAnim,Images)) :-
    ( extract_def_by_equality(ZDef,'animation_function',AFType,BAnim) ->
        ( extract_def_by_equality(ZDef,'animation_function_default',DFType,BDefaultAnim) -> true
        ; otherwise -> BDefaultAnim = none)
    ; extract_def_by_equality(ZDef,'animation_function_default',DFType,BDefaultAnim) ->
        BAnim = none),
    ( AFType = DFType -> true
    ; otherwise ->
        add_error(proz,'animation_function and animation_function_default should be of the same type')),
    ( AFType = set(tuple([tuple([_,_]),ImageType])) ->
        ( extract_image_filenames(ImageType,TypedSpecification,Images) -> true
        ; otherwise -> add_error(proz,'Unable to map animation function to images'))
    ; otherwise ->
        add_error(proz,'Unexpected type of animation_function'),fail).

extract_def_by_equality(ZDef,Name,Type,BExpr) :-
    get_equality(ZDef,ti(ref(name(Name,''),[]),Type),ZExpr),
    z_translate(ZExpr,BExpr1),
    clean_up(BExpr1,[],BExpr).
get_equality(ti(body(_,ZWhere),_),Name,Equal) :-
    !,member(A,ZWhere),get_equality(A,Name,Equal),!.
get_equality(ti(equal(X,Y),_),TName,Equal) :-
    ( X=TName -> Equal=Y
    ; Y=TName -> Equal=X).
extract_image_filenames(given(IName),TypedSpecification,Images) :-
    memberchk(ti(namedset(IName,Elements),_),TypedSpecification),
    z_translate_identifiers(Elements,PImages),
    maplist(append_gif,PImages,Images).
append_gif(Name,Gif) :- atom_concat(Name,'.gif',Gif).

set_settings(Settings,In,Out) :-
    write_section(meta,[proz_settings(Settings)],In,Out).

%*******************************************************************************
% Schema classification

/* Classifies each schema, i.e. assign a type (state, init, op, other)
   to each given schema */
classify_schemas(Schemas,Deps,[cls(init,Init),cls(state,State)|Operations]) :-
        ( classify_init_state(Deps, Init, State) -> true
        ; Deps= [] ->
          add_error(proz,'This Z specification has no schema and cannot be initialised.'),fail
        ; otherwise ->
            %debug:print_quoted(Deps),nl,
            ajoin(['Was not able to identify Z state schema.\n',
                   'Please check if there is a schema called Init that\n',
                   'references exactly one other schema (the state) in its\n',
                   'declaration part.'], Msg),
            add_error(proz,Msg),fail),
        classify_schemas(Schemas, Schemas, Deps, Init, State, Operations).

:- use_module(library(lists),[sublist/5]).
classify_init_state(Deps,'Init',State) :-
    findall(S,member(dep('Init',S),Deps),[State]).
classify_init_state(Deps,SchemaName,State) :-
   % see if we can find another Schema with Init in its name, which references exactly one other atomic Schema
    member(dep(SchemaName,_),Deps),
    atom_codes(SchemaName,SCodes),
    atom_codes('Init',InitCodes1), atom_codes('init',InitCodes2),
    (sublist(SCodes,InitCodes1,_,_,_) -> true ; sublist(SCodes,InitCodes2,_,_,_)),
    findall(S,member(dep(SchemaName,S),Deps),[State]), % we should check that State is not Delta or Xi state schema
    add_warning(proz,'Did not find Init Schema; assuming following schema to define the initialisation: ',SchemaName).

classify_schemas([], _, _, _, _, []).
classify_schemas([Schema|SRest], AllSchemas, Deps, Init, State, [cls(Class,Id)|CRest]) :-
    classify_schema(Schema, AllSchemas, Deps, Init, State, Class),
    !,Schema = fschema(Id,_,_,_),
    classify_schemas(SRest, AllSchemas, Deps, Init, State, CRest).
classify_schemas([_|SRest], AllSchemas, Deps, Init, State, CRest) :-
    classify_schemas(SRest, AllSchemas, Deps, Init, State, CRest).

classify_schema(fschema('Scope',_,ScopeVars,_), AllSchemas, _, _, State, scope) :-
    member(fschema(State,_,StateVars,_),AllSchemas),
    all_members(ScopeVars,StateVars),!.
classify_schema(fschema('ProZ_Goal',_,ScopeVars,_), AllSchemas, _, _, State, goal) :-
    member(fschema(State,_,StateVars,_),AllSchemas),
    all_members(ScopeVars,StateVars),!.
classify_schema(fschema('ProZ_Settings',_,_,_), _AllSchemas, _, _, _State, settings) :- !.
classify_schema(fschema('Invariant',_,InvVars,_), AllSchemas, _, _, State, inv) :-
    member(fschema(State,_,StateVars,_),AllSchemas),
    all_members(InvVars,StateVars),!.
classify_schema(Schema, AllSchemas, Deps, Init, State, op) :-
    Schema = fschema(Id,_,_,_),
    Id \= State, Id \= Init,
    is_operation(AllSchemas, Deps, Schema, State),
    !.

:- dynamic unchanged_vars_for_partial_zoperation/2.
:- use_module(library(ordsets),[list_to_ord_set/2,ord_member/2]).
/* Is the given schema an operation?
   1) It does include all variables of the state schema
   2) It does include all variables of the after state
   3) No other schema includes this schema */
is_operation(AllSchemas, Deps, fschema(Id,_,OpVars,_), State) :-
    \+ member(dep(_,Id),Deps), % not included by other schema
    %format('Try operation ~w with OpVars ~w and State schema is ~w~n',[Id,OpVars,State]),
    member(fschema(State,_,StateVars,_),AllSchemas), % get the variables of the state schema
    !,
    decorate_vars(StateVars,'\'',AfterStateVars),
    append(StateVars,AfterStateVars,NeededVars),
    (all_members(NeededVars,OpVars) -> true
      ; %format('~nSchema not recognised as operation: ~w over state variables ~w~n',[Id,StateVars]),
        OpVars \= [],
        list_to_ord_set(StateVars,SVS),
        list_to_ord_set(OpVars,OVS),
        primed_unprimed_ok(OVS,SVS), % check that all variables that are referenced occur with and without prime
        ord_subtract(SVS,OVS,Unchanged),
        format('Promoting Schema ~w to Z Operation by adding unchanged variables: ~w~n',[Id,Unchanged]),
        % TO DO: in test 1094 for System.fuzz the InitSystem operation is also recognised
        assert(unchanged_vars_for_partial_zoperation(Id,Unchanged))
    ).

primed_unprimed_ok([],_).
primed_unprimed_ok([name(ID,''),name(ID,'\'')|T],SVS) :- !,primed_unprimed_ok(T,SVS).
primed_unprimed_ok([NAME|T],SVS) :- \+ ord_member(NAME,SVS),primed_unprimed_ok(T,SVS).

/* unused:
operation_variables([],[]).
operation_variables([name(_,Deco)|Rest],NeededVars) :-
    member(Deco,['?','!']),!,operation_variables(Rest,NeededVars).
operation_variables([Var|Rest],NeededVars) :-
    select(Var,NeededVars,RestVars),
    !,operation_variables(Rest,RestVars).
 */

get_schema_by_class(Cls,Classification,Schemas,Def,Vars,Types) :-
    memberchk(cls(Cls,Name),Classification),
    memberchk(fschema(Name,Def,Vars,Types),Schemas).

%*******************************************************************************
% Initialisation

construct_init(Schemas,Classification,Any) :-
    member(cls(init,Init),Classification),
    member(fschema(Init,Def,_,_),Schemas),

    member(cls(state,State),Classification),
    member(fschema(State,_,StateVars,StateTypes),Schemas),!,

    (translate_initialisation(Def,StateVars,StateTypes,Any) -> true
    ; otherwise ->
        add_error(proz,'Internal error while translating initialisation'),fail).

translate_initialisation(TDef,StateVars,StateTypes,Any) :-
    check_z_substructuret(TDef,body,'invalid Z structure in initialisation',[nlabel(scalc)]),
    findall(rename(name(Name,'\''),name(Name,'')),
            member(name(Name,''), StateVars),
            Renamings),
    rename_variablest(TDef,Renamings,TypedDef),
    ti_expr(TypedDef,body(_,TypedWhere)),

    decorate_vars(StateVars,'\'',ZAnyVars),
    z_translate_typed_identifiers(StateVars,StateTypes,[],Lhs),
    z_translate_typed_identifiers(ZAnyVars,StateTypes,[],Rhs),
    Any = any(ZAnyVars,StateTypes,TypedWhere,Lhs,Rhs).

%create_ti(Name,Type,TI) :- ti_expr(TI,Name),ti_type(TI,Type).


set_init(InitAny,In,Out) :-
    write_section(initialisation,Statement,In,Out),
    construct_any(InitAny,[],Statement).

%*******************************************************************************
% Variables and invariant

set_variables(Schemas,Classification,In,Out) :-
    write_section(abstract_variables,Variables,In,M),
    write_section(invariant,Invariant,M,Out),

    member(cls(state,State),Classification),
    member(fschema(State,_,ZVars,Types),Schemas),!,

    z_translate_typed_identifiers(ZVars,Types,[],Variables),

    ( memberchk(cls(inv,Inv),Classification) ->
        memberchk(fschema(Inv,TypedDef,_,_),Schemas),
        print('Z invariant found'),nl,
        ti_expr(TypedDef,body(_,ZInvs))
    ; otherwise ->
        ZInvs = []),
    ( z_translate(ZInvs,Invariants) -> true
    ; otherwise -> add_error(proz, 'Internal error while translating invariant')),
    conjunct_predicates(Invariants,Invariant).


%*******************************************************************************
% Scope

/* currently unused:
register_scope(Schemas,Classification) :-
    ( memberchk(cls(scope,Scope),Classification) ->
        ( memberchk(fschema(Scope,TypedDef,_,_),Schemas),!,
            print('Z scope schema found'),nl,
            ti_expr(TypedDef,body(_,ZScope)),
            ( (z_translate(ZScope,ScopePredicates),conjunct_predicates(ScopePredicates,ScopePredicate)) ->
                true
            ; otherwise ->
                add_error(proz, 'Internal error while translating scope'),ScopePredicate=none
            )
        )
    ; otherwise ->
        ScopePredicate = none),
    set_scope_expression(ScopePredicate).
*/

%*******************************************************************************
% Operations

construct_operations(Schemas,Classification,Operations) :-
    member(cls(state,State),Classification),
    member(fschema(State,_,StateVars,StateTypes),Schemas),!,
    findall(Op,member(cls(op,Op),Classification),Ops),
    construct_operations2(Ops,Schemas,StateVars,StateTypes,Operations).
construct_operations2([],_,_,_,[]).
construct_operations2([Op|ORest],Schemas,StateVars,StateTypes,[Operation|OpRest]) :-
    member(fschema(Op,Def,Vars,Types),Schemas),!,
    ( translate_operation(Op,Def,Vars,Types,StateVars,StateTypes,Params,ParaTypes,Results,ResultTypes,Any) ->
        Operation = op(Op,Params,Results,Any,ParaTypes,ResultTypes),
        OpArgs =.. [Op|Params],print('Z operation: '),print(OpArgs),nl
    ; otherwise ->
        add_error(proz,'Internal error while translating operation', Op)
    ),
    construct_operations2(ORest,Schemas,StateVars,StateTypes,OpRest).

set_operations(Ops,In,Out) :-
    write_section(promoted,Promoted,In,M),
    write_section(operation_bodies,Bodies,M,Out),
        get_operations(Ops,Bodies,Promoted).
get_operations([],[],[]).
get_operations([Operation|ORest],[BOperation|Brest],[Promoted|PromotedRest]) :-
    get_operation(Operation,BOperation,Promoted),
        get_operations(ORest,Brest,PromotedRest).
get_operation(op(Op,Params,Results,Any,ParaTypes,ResultTypes),TOp,OpId) :-
    OpType = op(ResultTypes,ParaTypes),
    create_texpr(identifier(op(Op)),OpType,[],OpId),
    create_b_identifiers(Params,ParaTypes,BParams),
    create_b_identifiers(Results,ResultTypes,BResults),
    create_texpr(operation(OpId,BResults,BParams,Statement),OpType,OpInfo,TOp), % [] == Info ?? extract from Statement ??
    add_locals([BParams,BResults],[],Locals),
    construct_any(Any,Locals,Statement),
    get_texpr_info(Statement,StmtInfo),
    include(is_operation_info,StmtInfo,OpInfo).
create_b_identifiers([],[],[]).
create_b_identifiers([Name|Nrest],[Type|Trest],[Id|Irest]) :-
    create_texpr(identifier(Name),Type,[],Id),
    create_b_identifiers(Nrest,Trest,Irest).

is_operation_info(modifies(_)).
is_operation_info(reads(_)).

translate_operation(OpName,Def,Vars,Types,StateVars,StateTypes,Params,ParamTypes,Result,ResultTypes,Any) :-
    check_z_substructuret(Def,body,'invalid Z structure in operation',[nlabel(scalc)]),
    filter_vars_by_deco(Vars,Types,'?',ZParams,ParamTypes),
    filter_vars_by_deco(Vars,Types,'!',ZResult,ResultTypes),
    z_translate_identifiers(ZParams,Params),
    z_translate_identifiers(ZResult,Result),

    copy_result_vars(ZResult,ZAnyResults),

    find_unchanged(Def,UnchangedVars1,Def2),
    (unchanged_vars_for_partial_zoperation(OpName,AdditionalUnchanged)
     % check if we have an operation that does not talk about all variables, but we still classified it as an operation
      -> append(UnchangedVars1,AdditionalUnchanged,UnchangedVars)
       ; UnchangedVars = UnchangedVars1
    ),
    findall(rename(name(Name,''),name(Name,'\'')),
            member(name(Name,''),UnchangedVars),
            Renamings1),
    findall(rename(name(Name,Deco),name(Name,'!')),
            member(name(Name,Deco),ZAnyResults),
            Renamings2),
    append(Renamings1,Renamings2,Renamings),

    rename_variablest(Def2,Renamings,TypedAnyPreconditions),

    ti_expr(TypedAnyPreconditions,body(_,TypedWheres)),

    append([ZParams,ZResult,UnchangedVars],NotAffectedVars),
    remove_vars(StateVars,StateTypes,NotAffectedVars,ZChangedVars,ChangedVarTypes),
    decorate_vars(ZChangedVars,'\'',ZChangedVarsPrimed),

    % extract other vars (neither input,output nor state variables)
    decorate_vars(StateVars,'\'',AfterStateVars),
    append([StateVars,AfterStateVars,ZParams,ZResult],InOutStateVars),
    remove_vars(Vars,Types,InOutStateVars,ZOther,OtherTypes),

    % Z Vars and their types in the ANY declaration
    append([ZChangedVarsPrimed,ZAnyResults,ZOther],ZAnyVars),
    append([ChangedVarTypes,ResultTypes,OtherTypes],AnyTypes),

    % References in the substitution
    append(ZChangedVarsPrimed,ZAnyResults,ZRightHandSubst),
    append(ZChangedVars,ZResult,ZLeftHandSubst),
    append(ChangedVarTypes,ResultTypes,LeftRightHandTypes),
    z_translate_typed_identifiers(ZLeftHandSubst,LeftRightHandTypes,[],Lhs),
    z_translate_typed_identifiers(ZRightHandSubst,LeftRightHandTypes,[],Rhs),

    Any = any(ZAnyVars,AnyTypes,TypedWheres,Lhs,Rhs).

copy_result_vars([],[]).
copy_result_vars([name(Name,Deco)|Rest],[name(Name,NewDeco)|NRest]) :-
    z_counter(Deco,NewDeco),
    copy_result_vars(Rest,NRest).

filter_vars_by_deco([],[],_,[],[]).
filter_vars_by_deco([V|VRest],[T|TRest],Deco,Param,ParamTypes) :-
    ( V = name(_,Deco) ->
        (Param,ParamTypes) = ([V|PRest],[T|PTRest])
    ; otherwise ->
        (Param,ParamTypes) = (PRest, PTRest)),
    filter_vars_by_deco(VRest,TRest,Deco,PRest,PTRest).

%*******************************************************************************
% construction of B any statements from preformatted Z predicates and expressions
construct_any(any(Vars,Types,ZConds,LHS,RHS),Locals1,TExpr) :-
    % Translate the guard to B:
    z_translate(ZConds,BConds),conjunct_predicates(BConds,BCond),
    % Translate local variables to B:
    z_translate_typed_identifiers(Vars,Types,[],Identifiers),
    add_locals([Identifiers],Locals1,Locals),
    % Determine which variables have been used by the guard:
    find_identifier_uses(BCond,Locals,GuardReads),
    % Construct inner substitution:
    construct_substitution(LHS,RHS,Locals,InnerSubst,Writes,SubstReads),
    % Construct ANY or SELECT:
    ord_union(GuardReads,SubstReads,AllReads),
    Info = [modifies(Writes),reads(AllReads)],
    construct_any_or_select(Identifiers,BCond,Info,InnerSubst,AnyOrSelect),
    create_texpr(AnyOrSelect,subst,Info,TExpr).
construct_any_or_select([],BCond,Info,InnerSubst,select([When])) :- !,
    create_texpr(select_when(BCond,InnerSubst),subst,Info,When).
construct_any_or_select(Identifiers,BCond,_Info,InnerSubst,any(Identifiers,BCond,InnerSubst)).

construct_substitution(LHS,RHS,Locals,TExpr,Writes,Reads) :-
    create_texpr(Subst,subst,[modifies(Writes),reads(Reads)],TExpr),
    construct_substitution2(LHS,RHS,Locals,Subst,Writes,Reads).
construct_substitution2([],[],_Locals,skip,[],[]) :- !.
construct_substitution2(LHS,RHS,Locals,assign(LHS,RHS),Writes,Reads) :-
    get_texpr_ids(LHS,UWrites),sort(UWrites,Writes1),
    ord_subtract(Writes1,Locals,Writes),
    find_identifier_uses_l(RHS,Locals,Reads).

add_locals(IdentifiersL,LocalsIn,LocalsOut) :-
    append(IdentifiersL,Identifiers),
    get_texpr_ids(Identifiers,UIds),sort(UIds,Ids),
    ord_union(LocalsIn,Ids,LocalsOut).

%*******************************************************************************
% identify predicates that are in the initialisation and in every operation
% on the post-state
global_predicates(InitAny,Ops,GlobalPreds) :-
        pred_init(InitAny,InitPreds),
        findall(OpAny,member(op(_,_,_,OpAny,_,_),Ops),OpAnies),
        pred_ops(OpAnies,InitPreds,GlobalPreds).
pred_init(any(_,_,Preds,Lhs,Rhs),GPreds) :-
        create_renamings(Lhs,Rhs,Renamings),
        rename_variablest(Preds,Renamings,GPreds).
pred_ops([],GPreds,GPreds).
pred_ops([any(_,_,OpPreds,Lhs,Rhs)|ORest],OldPreds,GPreds) :-
        create_renamings(Rhs,Lhs,Renamings),
        filter_preds(OldPreds,OpPreds,Renamings,NewPreds),
        pred_ops(ORest,NewPreds,GPreds).
filter_preds([],_,_,[]).
filter_preds([Old|ORest],OpPreds,Renamings,NewPreds) :-
        rename_variablest(Old,Renamings,Comparable),
        (exists_equalt(OpPreds,Comparable) ->
            NewPreds = [Old|PRest];
            NewPreds = PRest),
        filter_preds(ORest,OpPreds,Renamings,PRest).
create_renamings([],[],[]).
create_renamings([L|LRest],[R|RRest],[rename(name(LN,LD),name(RN,RD))|Rest]) :-
    get_texpr_info(L,LInfos),member(zname(LN,LD),LInfos),
    get_texpr_info(R,RInfos),member(zname(RN,RD),RInfos),!,
        create_renamings(LRest,RRest,Rest).

%*******************************************************************************
% remove global predicates from operations
remove_global_predicates(Ops,[],Ops) :- !.
remove_global_predicates(Ops,GlobalPredicates,NewOps) :-
    remove_globalpreds(Ops,GlobalPredicates,NewOps).
remove_globalpreds([],_,[]).
remove_globalpreds([op(Op,Param,Results,Any,ParaTypes,ResultTypes)|ORest],GlobalPredicates,
                   [op(Op,Param,Results,NewAny,ParaTypes,ResultTypes)|NRest]) :-
    remove_globalpreds2(Any,GlobalPredicates,NewAny),
    remove_globalpreds(ORest,GlobalPredicates,NRest).
remove_globalpreds2(any(Vars,Types,Old,LHS,RHS),GlobalPredicates,
                    any(Vars,Types,New,LHS,RHS)) :-
    remove_globalpreds3(Old,GlobalPredicates,New).
remove_globalpreds3([],_,[]).
remove_globalpreds3([Pred|ORest],GlobalPredicates,New) :-
    ( exists_equalt(GlobalPredicates,Pred) ->
        New = NRest
    ; otherwise ->
        New = [Pred|NRest]),
    remove_globalpreds3(ORest,GlobalPredicates,NRest).

%**********************************************************************
% Dies und das

all_members([],_).
all_members([FirstMember|Members], List) :-
    member(FirstMember, List),!,
    all_members(Members, List).

%*******************************************************************************
% Z to ProB translation

z_translate([],[]) :- !.
z_translate([ZExpr|ZRest],[BExpr|BRest]) :-
    !,
    z_translate(ZExpr,BExpr),z_translate(ZRest,BRest).
z_translate(ti(ZExpr,ZType),BTExpr) :-
    internal2b(ZType,BType),
    z_translate1(ZExpr,ZType,BExpr),
    create_typed_bexpr(BExpr,BType,BTExpr).

create_typed_bexpr(symbolic_comprehension_set(Ids, Lhs),BType,Res) :- !,
     Res = b(comprehension_set(Ids, Lhs),BType,[prob_annotation('SYMBOLIC')]).
create_typed_bexpr(BExpr,BType,b(BExpr,BType,[])).

z_translate1(mu(Body,Computation), ZType, mu(BSet)) :-
    !, ZSet = ti(comp(Body,Computation),set(ZType)),
    z_translate(ZSet,BSet).
z_translate1(ZExpr,_ZType,BExpr) :-
    z_translate2(ZExpr,BExpr).

z_translate2(falsity, falsity) :- !.
z_translate2(truth, truth) :- !.
z_translate2(forall(BodyTi, ZConsequent), forall(Ids, Lhs, Rhs)) :-
    !, schema_to_typed_identifiers(BodyTi,Ids,Lhs),
    z_translate(ZConsequent,Rhs).
z_translate2(exists(BodyTi, ZCondition), exists(Ids, BCondition)) :-
    !, schema_to_typed_identifiers(BodyTi,Ids,Condition1),
    z_translate(ZCondition,Rhs),
    conjunct_predicates([Condition1,Rhs],BCondition).
z_translate2(lambda(BodyTi, ZExpr), lambda(Ids, Lhs, Rhs)) :-
    !, schema_to_typed_identifiers(BodyTi,Ids,Lhs),
    z_translate(ZExpr,Rhs).
z_translate2(ref(name('\\emptyset',''),[]),empty_set) :- !.
%z_translate2(ref(name('\\emptyset',''),[_TypeArg]),empty_set) :- !.  % support constructs like \emptyset[\nat] ; currently argument removed in typing phase
z_translate2(ref(name('\\nat',''),[]),integer_set('NATURAL')) :- !.
z_translate2(ref(name('\\nat','_1'),[]),integer_set('NATURAL1')) :- !.
z_translate2(ref(name('\\num',''),[]),integer_set('INTEGER')) :- !.
/* In some special cases Schemas are refered just by ref(name(..)..), not by SEXPR(..) */
%z_translate2(ref(name(Name,Deco),Params), Env, BExpr) :-
%   named_schema(Env,Name),!,
%   z_translate(sexpr(sref(Name,Deco,Params,[])),BExpr).
%z_translate2(ref(FreetypeName,[]), Env, 'FreetypeSet'(Id)) :-
%   zlookup(Env,data(FreetypeName,_)),
%   !,z_translate_identifier(FreetypeName,Id).
z_translate2(ref(Name,[]), identifier(Id)) :-
    !,z_translate_identifier(Name, Id).
z_translate2(inop(name('\\extract',''), Za, Zb), compaction(DomRes)) :-
    !, create_texpr(domain_restriction(Ba,Bb),BType,[],DomRes),
    z_translate(Za,Ba),z_translate(Zb,Bb),
    get_texpr_type(Bb,BType).
z_translate2(inop(name('\\filter',''), Za, Zb), compaction(RanRes)) :-
    !, create_texpr(range_restriction(Ba,Bb),BType,[],RanRes),
    z_translate(Za,Ba),z_translate(Zb,Bb),
    get_texpr_type(Ba,BType).
z_translate2(apply(Iteration,ZRel), Result) :-
    % iteration of relations. The problem of this operator is that
    % it can have negative arguments in Z but not in B
    ti_expr(Iteration, apply(Ref,ZN)),
    ti_expr(Ref, ref(name(iter,''),[])),
    !, z_translate(ZN,BN), z_translate(ZRel,BRel),
    get_texpr_type(BRel,BRelType),
    create_texpr(reverse(BRel),BRelType,[],Inverse),
    ( is_a_number(ZN,Value), Value >= 0 ->
        % R^x, where x is a non-negative number, we can use it directly
        Result = iteration(BRel,BN)
    ; is_a_negative_number(ZN,Value), Value >= 0 ->
        % R^-x, where x is a non-negative number, we can rewrite it into (R~)^x
        create_texpr(integer(Value),integer,[],Negation),
        Result = iteration(Inverse,Negation)
    ; otherwise ->
        % R^x, where x is an arbitrary expression, we surround it by if-then-else
        % to negate the x if necessary
        create_texpr(integer(0),integer,[],Zero),
        create_texpr(less_equal(Zero,BN),pred,[],NonNegative),
        create_texpr(iteration(BRel,BN),BRelType,[],Then),
        create_texpr(unary_minus(BN),integer,[],Negation),
        create_texpr(iteration(Inverse,Negation),BRelType,[],Else),
        Result = if_then_else(NonNegative,Then,Else)).
z_translate2(inop(ZopName, Za, Zb), BBinOp) :-
    z_translate_identifier(ZopName,Zop),
    is_binary_op(Zop),!,
    z_translate_binary_op(Zop,Za,Zb,BBinOp).
z_translate2(inrel(ZopName, Za, Zb), BBinOp) :-
    z_translate_identifier(ZopName,Zop),
    is_binary_op(Zop),!,
    z_translate_binary_op(Zop,Za,Zb,BBinOp).
z_translate2(ingen(ZopName, Za, Zb), BBinOp) :-
    z_translate_identifier(ZopName,Zop),
    is_binary_op(Zop),!,
    z_translate_binary_op(Zop,Za,Zb,BBinOp).
z_translate2(number(Value), integer(Value)) :- !.
z_translate2(ext(ZEntities), set_extension(BEntities)) :-
    !, z_translate(ZEntities, BEntities).
z_translate2(seq(ZEntities), sequence_extension(BEntities)) :-
    !, z_translate(ZEntities, BEntities).
z_translate2(apply(Items,ZExpr), bag_items(BExpr)) :-
    ti_expr(Items,ref(name('items',''),[])),!,
    z_translate(ZExpr,BExpr).
z_translate2(apply(ZFun, Zarg), BExpr) :-
    ti_expr(ZFun,ref(Name,[])),
    z_translate_identifier(Name,Op),
    is_unary_op(Op),!,
    z_translate_unary_op(Op,Zarg,BExpr).
z_translate2(apply(ZFun, ZArg), function(BFun,BArg)) :-
    !, z_translate(ZFun,BFun), z_translate(ZArg,BArg).
z_translate2(postop(Name, Zarg), BExpr) :-
    z_translate_identifier(Name,Op),
    is_unary_op(Op),!,
    z_translate_unary_op(Op,Zarg,BExpr).
z_translate2(prerel(Name, Zarg), BExpr) :-
    z_translate_identifier(Name,Op),
    is_unary_op(Op),!,
    z_translate_unary_op(Op,Zarg,BExpr).
z_translate2(pregen(Name, Zarg), BExpr) :-
    z_translate_identifier(Name,Op),
    is_unary_op(Op),!,
    z_translate_unary_op(Op,Zarg,BExpr).
z_translate2(inrel(name('\\prefix',''), Za, Zb), BExpr) :-
    !,z_translate(Za,Ba),z_translate(Zb,Bb),
    get_texpr_type(Ba,SeqType),
    create_texpr(size(Ba),integer,[],SizeA),
    create_texpr(size(Bb),integer,[],SizeB),
    create_texpr(greater_equal(SizeB,SizeA),pred,[],GreaterEqual),
    create_texpr(restrict_front(Bb,SizeA),SeqType,[],Prefix),
    create_texpr(equal(Ba,Prefix),pred,[],Equal),
    conjunct_predicates([GreaterEqual,Equal],TypedBExpr),
    get_texpr_expr(TypedBExpr,BExpr).
z_translate2(inrel(name('\\suffix',''), Za, Zb), BExpr) :-
    !,z_translate(Za,Ba),z_translate(Zb,Bb),
    get_texpr_type(Ba,SeqType),
    create_texpr(size(Ba),integer,[],SizeA),
    create_texpr(size(Bb),integer,[],SizeB),
    create_texpr(greater_equal(SizeB,SizeA),pred,[],GreaterEqual),
    create_texpr(minus(SizeB,SizeA),integer,[],Diff),
    create_texpr(restrict_tail(Bb,Diff),SeqType,[],Suffix),
    create_texpr(equal(Ba,Suffix),pred,[],Equal),
    conjunct_predicates([GreaterEqual,Equal],TypedBExpr),
    get_texpr_expr(TypedBExpr,BExpr).
z_translate2(inrel(name('\\inseq',''), Za, Zb), BExpr) :-
    !, create_in_sequence(Za,Zb,BExpr).
z_translate2(prerel(name('\\disjoint',''), Z), BExpr) :-
    !, z_translate(Z,B), create_disjoint(B,BExpr).
z_translate2(inrel(name('\\partition',''), Za, Zb), partition(BSet,BSeq)) :-
    ti_expr(Za,seq(ZSeq)),!,
    z_translate(Zb,BSet),
    z_translate(ZSeq,BSeq).
z_translate2(inrel(name('\\partition',''), Za, Zb), BExpr) :-
    !, create_partition(Za,Zb,BExpr).
z_translate2(select(ZRecord, ZField), record_field(BRecord, FieldIdentifier)) :-
    !, z_translate(ZRecord, BRecord),
    z_translate_identifier(ZField, FieldIdentifier).
/* Comprehension set without expression */
z_translate2(comp(Body, Nothing),Res) :-
    !,
    ( ti_expr(Nothing,nothing) -> true
    ; otherwise ->
        add_error(proz,'Encountered comprehension set with expression in translation. Should have been replaced.'),
        fail),
    ti_expr(Body,body(Decls,ZWheres)),
    ~~mnf( varlist_of_expanded_decls(Decls, Vars, Types) ),
    z_translate_typed_identifiers(Vars,Types,[],Ids),
    z_translate(ZWheres,Wheres),
    conjunct_predicates(Wheres,Lhs),
    %print(comp(Ids,Lhs)),nl,
    (is_symbolic_set_comprehension(Ids,Lhs)
     -> Res = symbolic_comprehension_set(Ids, Lhs)  %,print(symbolic(Ids)),nl
      ; Res = comprehension_set(Ids, Lhs)).
z_translate2(cross([Za,Zb]), cartesian_product(Ba,Bb)) :-
    !, z_translate(Za,Ba), z_translate(Zb,Bb).
z_translate2(cross([ZExpr|ZRest]), cartesian_product(BExpr,TBRest)) :-
    !, z_translate(ZExpr,BExpr), z_translate2(cross(ZRest),BRest),
    BRest=cartesian_product(BL,BR),get_texpr_type(BL,set(LType)),get_texpr_type(BR,set(RType)),
    create_texpr(BRest,set(couple(LType,RType)),[],TBRest).
z_translate2(tuple([Za,Zb]), couple(Ba,Bb)) :-
    !, z_translate(Za,Ba), z_translate(Zb,Bb).
z_translate2(tuple([ZExpr|ZRest]), couple(BExpr,TBRest)) :-
    !, z_translate(ZExpr,BExpr), z_translate2(tuple(ZRest),BRest),
    BRest=couple(BL,BR),get_texpr_type(BL,LType),get_texpr_type(BR,RType),
    create_texpr(BRest,couple(LType,RType),[],TBRest).
z_translate2(letexpr(ZLets,ZExpr), let_expression(Bids,BAssignments,BExpr)) :-
    !, translate_lets(ZLets,Bids,BAssignments),
    z_translate(ZExpr,BExpr).
z_translate2(letpred(ZLets,ZExpr), let_predicate(Bids,BAssignments,BExpr)) :-
    !, translate_lets(ZLets,Bids,BAssignments),
    z_translate(ZExpr,BExpr).
z_translate2(if(ZPred,ZIf,ZElse), if_then_else(BPred,BIf,BElse)) :-
    !, z_translate(ZPred,BPred),
    z_translate(ZIf,BIf), z_translate(ZElse,BElse).
z_translate2(inrel(Name,Za,Zb), member(Couple,Relation)) :-
    !,z_translate(Za,Ba),z_translate(Zb,Bb),
    z_translate_identifier(Name,Id),
    get_texpr_type(Ba,TypeA), get_texpr_type(Bb,TypeB),
    create_texpr(couple(Ba,Bb),couple(TypeA,TypeB),[],Couple),
    create_texpr(identifier(Id),set(couple(TypeA,TypeB)),[],Relation).

% Internal constructs
z_translate2(binding(Bindings), rec(Fields)) :-
    !, create_fields_for_binding(Bindings,Fields).
z_translate2(ftconstant(Freetype,Case),value(freeval(FreetypeId,ConstId,term(ConstId)))) :-
    !,z_translate_identifier(Freetype,FreetypeId),
    z_translate_identifier(Case,ConstId).
z_translate2(ftconstructor(Freetype,Case,ZExpr),
         freetype_constructor(FreetypeId,CaseId,BExpr)) :-
    !,z_translate(ZExpr,BExpr),
    z_translate_identifier(Freetype,FreetypeId),
    z_translate_identifier(Case,CaseId).
z_translate2(ftdestructor(Freetype,Case,ZExpr),
             freetype_destructor(FreetypeId,CaseId,BExpr)) :-
    !,z_translate(ZExpr,BExpr),
    z_translate_identifier(Freetype,FreetypeId),
    z_translate_identifier(Case,CaseId).
z_translate2(ftcase(Freetype,Case,ZExpr),
             freetype_case(FreetypeId,CaseId,BExpr)) :-
    !,z_translate(ZExpr,BExpr),
    z_translate_identifier(Freetype,FreetypeId),
    z_translate_identifier(Case,CaseId).

z_translate2(basetype(set(Type)), Expr) :-
    !,translate_basetype2(Type,Expr).

z_translate2(reclet(ZId,ZSet), recursive_let(BId,BSet) ) :-
    !,z_translate(ZId,BId),
    % TODO: recursive_let accepts only comprehension sets as argument,
    % maybe this should be checked here (lambda should be ok, too, because
    % it's converted to a comprehension set
    z_translate(ZSet,BSet1),
    add_texpr_infos(BSet1,[prob_annotation('SYMBOLIC')],BSet).

z_translate2(ZBinOp, BBinOp) :-
    ZBinOp =.. [ZopName,Za,Zb],
    is_binary_op(ZopName),!,
    z_translate_binary_op(ZopName,Za,Zb,BBinOp).
z_translate2(ZUnOp, BUnOp) :-
    ZUnOp =.. [ZopName, Zarg],
    is_unary_op(ZopName),!,
    z_translate_unary_op(ZopName,Zarg,BUnOp).

z_translate2(ZExpr,_) :-
    add_error(proz,'Unmatched Z expression',ZExpr),
    fail.

% added by Leuschel to mark certain set comprehensions as symbolic
% TO DO: look with Daniel Plagge whether rec__ should always be symbolic, and whether other sets should be put symbolic
is_symbolic_set_comprehension([TID],_) :-
    get_texpr_id(TID,ID),
    atom_codes(ID,C),
    append("rec__",_,C). % rec__ are special variables introduced in ztransformations:removesexpr2 for Schema fields

is_a_negative_number(ZExpr,N) :-
    ti_expr(ZExpr,apply(Ref,Num)),
    ti_expr(Ref,ref(name('(-)',''),[])),
    is_a_number(Num,N).
is_a_number(ZExpr,N) :-
    ti_expr(ZExpr,number(N)).

z_translate_binary_op(ZopName,Za,Zb,Bop) :-
    z_binary_op(ZopName,BopName),!,
    z_translate(Za,Ba),z_translate(Zb,Bb),
    Bop =.. [BopName,Ba,Bb].

z_translate_unary_op(ZopName,Zarg,Bop) :-
    z_unary_operator(ZopName, BopName),!,
    z_translate(Zarg, Barg),
    Bop =.. [BopName, Barg].

translate_basetype(Type,TExpr) :-
    create_texpr(BExpr,set(BType),[],TExpr),
    internal2b(Type,BType),
    translate_basetype2(Type,BExpr).

translate_basetype2(num,integer_set('INTEGER')).
translate_basetype2(given(Name), value(global_set(Id))) :- z_translate_identifier(Name,Id).
translate_basetype2(freetype(Name), freetype_set(Id)) :- z_translate_identifier(Name,Id).
translate_basetype2(set(T), pow_subset(B)) :- translate_basetype(T,B).
translate_basetype2(tuple([A,B]), cartesian_product(TA,TB)) :-
    !,translate_basetype(A,TA),
    translate_basetype(B,TB).
translate_basetype2(tuple([A,B,C|Rest]), cartesian_product(TA,TBRest)) :-
    translate_basetype(A,TA),
    translate_basetype(tuple([B,C|Rest]),BRest),
    BRest = cartesian_product(BL,BR),
    get_texpr_type(BL,set(LType)),get_texpr_type(BR,set(RType)),
    create_texpr(BRest,set(couple(LType,RType)),[],TBRest).
translate_basetype2(schema(ZFields),struct(Rec)) :-
    translate_bt_fields(ZFields,BFields,FieldTypes),
    create_texpr(rec(BFields),record(FieldTypes),[],Rec).
translate_bt_fields([],[],[]).
translate_bt_fields([ZField|ZRest],[BField|BRest],[FieldType|TRest]) :-
    translate_bt_field(ZField,BField,FieldType),
    translate_bt_fields(ZRest,BRest,TRest).
translate_bt_field(field(Name,T),field(FieldId,B),field(FieldId,set(BType))) :-
    internal2b(T,BType),
    z_translate_identifier(Name,FieldId),
    translate_basetype(T,B).

schema_to_typed_identifiers(BodyTi,Ids,Condition) :-
    ti_type(BodyTi,set(schema(ZFields))),
    zfields2b(ZFields,Vars,Types),
    z_translate_typed_identifiers(Vars,Types,[],Ids),

    ti_expr(BodyTi,body(_,WhereTi)),
    z_translate(WhereTi,Wheres),
    conjunct_predicates(Wheres,Condition).

create_in_sequence(Za,Zb,BExpr) :-
    z_translate(Za,Ba),z_translate(Zb,Bb),
    get_texpr_type(Ba,BSeqType), BSeqType = set(couple(integer,BElemType)),
    ti_type(Za,ZSeqType), ZSeqType = set(tuple([num,ZElemType])),
    translate_basetype(ZElemType,BaseSet),
    create_texpr(identifier('__card'),integer,[],C),
    create_texpr(identifier('__i'),integer,[],I),
    create_texpr(identifier('__la'),integer,[],La),
    create_texpr(identifier('__lb'),integer,[],Lb),
    create_texpr(identifier('__offset'),integer,[],Offset),
    create_texpr(let_predicate([La,Lb],[Ba,Bb],Conjunction),pred,[],TypedBExpr),
    create_texpr(seq(BaseSet),set(BSeqType),[],Sequences),
    create_texpr(member(La,Sequences),pred,[],Asequence),
    create_texpr(member(Lb,Sequences),pred,[],Bsequence),
    create_texpr(card(La),integer,[],CardA),
    create_texpr(card(Lb),integer,[],CardB),
    create_texpr(minus(CardB,CardA),integer,[],Diff),
    create_texpr(let_predicate([C],[Diff],Exists),pred,[],Let),
    create_texpr(integer(0),integer,[],Zero),
    create_texpr(interval(Zero,Diff),integer,[],Interval),
    create_texpr(member(Offset,Interval),pred,[],InInterval),
    create_texpr(conjunct(InInterval,Forall),pred,[],Pred),
    create_texpr(forall([I],InDomain,Equal),pred,[],Forall),
    create_texpr(domain(Ba),integer,[],Domain),
    create_texpr(member(I,Domain),pred,[],InDomain),
    create_texpr(equal(ElemA,ElemB),pred,[],Equal),
    create_texpr(function(Ba,I),BElemType,[],ElemA),
    create_texpr(add(I,Offset),integer,[],I2),
    create_texpr(function(Bb,I2),BElemType,[],ElemB),
    create_exists([Offset],Pred,Exists),
    conjunct_predicates([Asequence,Bsequence,Let],Conjunction),
    get_texpr_expr(TypedBExpr,BExpr).

create_disjoint(B, let_predicate([D],[Domain],Forall)) :-
    get_texpr_type(B,set(couple(integer,set(ElemType)))),
    create_texpr(identifier('__d'),set(integer),[],D),
    create_texpr(identifier('__i'),integer,[],I),
    create_texpr(identifier('__j'),integer,[],J),
    create_texpr(domain(B),set(integer),[],Domain),
    create_texpr(forall([I,J],Pre,IsDisjunct),pred,[],Forall),
    create_texpr(member(I,D),pred,[],M1),
    create_texpr(member(J,D),pred,[],M2),
    create_texpr(less(I,J),pred,[],Less),
    conjunct_predicates([M1,M2,Less],Pre),
    create_texpr(function(B,I),set(ElemType),[],SetA),
    create_texpr(function(B,J),set(ElemType),[],SetB),
    create_texpr(intersection(SetA,SetB),set(ElemType),[],Intersection),
    create_texpr(empty_set,set(ElemType),[],Empty),
    create_texpr(equal(Intersection,Empty),pred,[],IsDisjunct).

% maybe we could in some cases use the new partition operator from Event-B (if Ba is known?) ; or maybe provide better constraint propagation support
create_partition(Za,Zb,let_predicate([Seq],[Ba],Conjunct)) :-
    z_translate(Za,Ba),
    z_translate(Zb,Bb),
    % print('Z Partition : '), translate:print_bexpr(Ba),nl, translate:print_bexpr(Bb),nl,
    get_texpr_type(Ba,SeqType),
    get_texpr_type(Bb,set(Type)),
    create_texpr(identifier('__sets'),SeqType,[],Seq),
    create_disjoint(Seq,Disjoint1), create_texpr(Disjoint1,pred,[],Disjoint),
    create_texpr(general_union(Range),set(Type),[],Union),
    create_texpr(range(Seq),set(Type),[],Range),
    create_texpr(equal(Union,Bb),pred,[],Unification),
    create_texpr(conjunct(Disjoint, Unification),pred,[],Conjunct).
    % ,print('PARTITION: '), translate:print_bexpr(b(let_predicate([Seq],[Ba],Conjunct),pred,[])),nl.

is_binary_op(OpName) :- z_binary_op(OpName,_).
is_unary_op(OpName) :- z_unary_operator(OpName,_).

/* Boolean binary operators */
z_binary_op(and, conjunct).
z_binary_op(or, disjunct).
z_binary_op(implies, implication).
z_binary_op(equiv, equivalence).
/* Generic equals */
z_binary_op(equal, equal).
z_binary_op('\\neq', not_equal).
/* Set binary operators */
z_binary_op(member, member).
z_binary_op('\\notin', not_member).
z_binary_op('\\cup', union).
z_binary_op('\\cap', intersection).
z_binary_op('\\subseteq', subset).
z_binary_op('\\subset', subset_strict).
z_binary_op('\\setminus', set_subtraction).
/* Arithmetic binary operators */
z_binary_op('+', add).
z_binary_op('-', minus).
z_binary_op('*', multiplication).
z_binary_op('\\div', floored_div). % not div; B and Z semantics differ
z_binary_op('\\mod', modulo).
z_binary_op('\\upto', interval).
z_binary_op('<', less).
z_binary_op('\\leq', less_equal).
z_binary_op('\\geq', greater_equal).
z_binary_op('>', greater).
/* Relations and functions */
z_binary_op('\\rel', relations).
z_binary_op('\\pfun', partial_function).
z_binary_op('\\fun', total_function).
z_binary_op('\\pinj', partial_injection).
z_binary_op('\\inj', total_injection).
z_binary_op('\\psurj', partial_surjection).
z_binary_op('\\surj', total_surjection).
z_binary_op('\\bij', total_bijection).
z_binary_op('\\ffun', partial_function).
z_binary_op('\\finj', partial_injection).
z_binary_op('\\mapsto', couple).
z_binary_op('\\comp', composition).
z_binary_op('\\dres', domain_restriction).
z_binary_op('\\ndres', domain_subtraction).
z_binary_op('\\rres', range_restriction).
z_binary_op('\\nrres', range_subtraction).
z_binary_op('_(|_|)', image).
z_binary_op('\\oplus', overwrite).
/* Sequences */
z_binary_op('\\cat', concat).

z_unary_operator('(-)', unary_minus).
z_unary_operator('\\dom', domain).
z_unary_operator('\\ran', range).
z_unary_operator(power, pow_subset).
z_unary_operator('\\power_1', pow1_subset).
/* In ProB all sets are treated as finite, so we use here the same as above */
z_unary_operator('\\finset', fin_subset).
z_unary_operator('\\finset_1', fin1_subset).
z_unary_operator('\\bigcup', general_union).
z_unary_operator('\\bigcap', general_intersection).
z_unary_operator('\\dcat', general_concat).

z_unary_operator(not, negation).
z_unary_operator('\\id', identity).
z_unary_operator('\\inv', reverse).

z_unary_operator('\\plus', closure).

z_unary_operator('\\#', card).
z_unary_operator('min', min).
z_unary_operator('max', max).

z_unary_operator('\\seq', seq).
z_unary_operator('\\seq_1', seq1).
z_unary_operator('\\iseq', iseq).
z_unary_operator('rev', rev).
z_unary_operator('head', first).
z_unary_operator('last', last).
z_unary_operator('tail', tail).
z_unary_operator('front', front).
z_unary_operator('squash', compaction).

z_unary_operator('first', first_of_pair).
z_unary_operator('second', second_of_pair).


translate_lets([],[],[]).
translate_lets([LetdefTi|ZRest],[Bid|IdRest],[BExpr|BRest]) :-
    ti_expr(LetdefTi,letdef(ZName,ZExpr)),
    ti_type(ZExpr,ZType),internal2b(ZType,BType),
    z_translate_typed_identifier(ZName,BType,[],Bid),
    z_translate(ZExpr,BExpr),
    translate_lets(ZRest,IdRest,BRest).

create_fields_for_binding([],[]).
create_fields_for_binding([BindingTi|ZRest],
                          [field(BField,BExpr)|BRest]) :-
    ti_expr(BindingTi,bindfield(Name,ZExpr)),
    z_translate_identifier(Name,BField),
    z_translate(ZExpr,BExpr),
    create_fields_for_binding(ZRest,BRest).

:- public varlist_of_expanded_decls/3. % because Spider cannot deal with ~~mnf
varlist_of_expanded_decls([],[],[]).
varlist_of_expanded_decls([TDecl|Drest],Vars,Types) :-
        ti_expr(TDecl,decl(Names,TypeDecl)),
        ti_type(TypeDecl,set(ZType)),
        internal2b(ZType,BType),
        varlist_of_expanded_decl(Names,BType,LTypes),
        append(Names,Vrest,Vars),
        append(LTypes,Trest,Types),
        varlist_of_expanded_decls(Drest,Vrest,Trest).
varlist_of_expanded_decl([],_,[]).
varlist_of_expanded_decl([_|Nrest],Type,[Type|Trest]) :-
        varlist_of_expanded_decl(Nrest,Type,Trest).







%*******************************************************************************
% register_given_sets/1: registers all given sets in the specification
% The specification must not include type information

set_given_sets(Spec,In,Out) :-
    findall(namedset(Name,Members),member(ti(namedset(Name,Members),_),Spec),NamedSets),
    findall(Sets,member(ti(given(Sets),_),Spec),ListOfSetsTmp),
    append(ListOfSetsTmp,ListOfSets),
    get_given_sets(ListOfSets,ListOfDeferredSets),
    get_named_sets(NamedSets,BNamedSets,BEnumeratedElems),
    write_section(deferred_sets,ListOfDeferredSets,In,MDsets),
    write_section(enumerated_sets,BNamedSets,MDsets,MEnums),
    write_section(enumerated_elements,BEnumeratedElems,MEnums,Out).

get_given_sets([],[]).
get_given_sets([Setname|Rest],[BSet|Brest]) :-
    z_translate_identifier(Setname, Setid),
    print_message(given_set(Setid)),
    create_texpr(identifier(Setid),set(global(Setid)),[given_set],BSet),
    get_given_sets(Rest,Brest).
get_named_sets([],[],[]).
get_named_sets([namedset(Setname,Members)|Rest],[BSet|BSetRest],Enumerated) :-
    z_translate_identifier(Setname,Setid),
    create_texpr(identifier(Setid),set(global(Setid)),[given_set],BSet),
    z_translate_identifiers(Members,Memberids),
    create_enum_elements(Memberids,global(Setid),LocalEnumerated),
    append(LocalEnumerated,RestEnumerated,Enumerated),
    get_named_sets(Rest,BSetRest,RestEnumerated).
create_enum_elements([],_,[]).
create_enum_elements([Id|IRest],Type,[TId|TRest]) :-
    create_texpr(identifier(Id),Type,[enumerated_set_element],TId  ),
    create_enum_elements(IRest,Type,TRest).

%*******************************************************************************
% register_freetypes/2: Registers the freetypes of the given specification.
% The specification must be given with type informations.

set_freetypes(Spec,In,Out) :-
    write_section(freetypes,Freetypes,In,Out),
    get_freetypes(Spec,Freetypes).
get_freetypes([],[]).
get_freetypes([ExprTi|Rest],Freetypes) :-
    ( ti_expr(ExprTi,data(Freetype,Arms)) ->
        get_freetype(Freetype,Arms,BFreetype),
        Freetypes = [BFreetype|FRest]
    ; otherwise ->
        Freetypes = FRest),
    get_freetypes(Rest,FRest).
get_freetype(Freetype,ArmsTi,freetype(FreetypeId,Cases)) :-
    extract_cases_from_arms(ArmsTi,Cases),
    z_translate_identifier(Freetype,FreetypeId),
    print_message(freetype(FreetypeId)).
extract_cases_from_arms([],[]).
extract_cases_from_arms([ArmTi|RestArms],[case(CaseId,Type)|RestCases]) :-
    ti_expr(ArmTi,arm(Case,SwitchTi)),
    z_translate_identifier(Case,CaseId),
    ti_expr(SwitchTi,Switch),
    ( Switch = just(ExprTi) ->
        ti_type(ExprTi,set(ZType)),internal2b(ZType,Type)
    ; otherwise ->
        Type = constant([CaseId])
    ),
    extract_cases_from_arms(RestArms,RestCases).

%*******************************************************************************
% clean up the machine
clean_up_machine(Machine,Clean) :-
    clean_up_sections([concrete_constants,properties,
                       abstract_variables,invariant,
                       promoted,initialisation,operation_bodies],
                      Machine,Clean).
clean_up_sections([],M,M).
clean_up_sections([Section|Rest],In,Out) :-
    clean_up_section(Section,In,M),
    clean_up_sections(Rest,M,Out).
clean_up_section(Section,In,Out) :-
    select_section_texprs(Section,Old,New,In,Out),
    clean_up_l_with_optimizations(Old,[],New,Section).

%*******************************************************************************
% test the conversion of specifications
:- public startzconversiontests/0.
startzconversiontests :-
    startzconversiontest('Z/Daniel/testcases.fuzz').

:- use_module(library(system),[environ/2]).
startzconversiontest(Testfile) :-
    system:environ('PROB_EX_DIR',Dir),
    append_atoms(Dir,'/',Dir2),
    append_atoms(Dir2,Testfile,Fullpath),
    testconversion(Fullpath).

testconversion(File) :-
    ( proz:load_and_simplify(File,Tests,Specification) -> true
    ; otherwise ->
        add_error(proz,'Failed to load and simplify Z conversion test file: ',File)),
    runtests(Tests,Specification,Errors),
    ( Errors = [] ->
        true
    ; otherwise ->
        add_errors(Errors),
        fail).

add_errors([]).
add_errors([T|Rest]) :-
    add_error(proz,'Z conversion testcase failed: ', T),
    add_errors(Rest).

load_and_simplify(File,Tests,Specification) :-
    load_fuzzfile(File),
    get_z_definitions(Z),
    findtests(Z,Tests),
    simplify_untyped(Z,Z2),!,
    global_environment(Z2,_,Z3),!,
    simplify_typed(Z3,Specification).

findtests(Z,Tests) :-
    findall(Name,getschema(Z,Name,_),Schemas),
    findall(test(N,T),(member(T,Schemas),testmark(T,N),member(N,Schemas)),Tests).

getschema(Z,Name,Def) :- member(sdef(shead(Name,[]),Def),Z).
getschema(Z,Name,Def) :- member(defeq(shead(Name,[]),Def),Z).
getschemat(Z,Name,Def) :- member(ti(sdef(shead(Name,[]),Def),_),Z).
getschemat(Z,Name,Def) :- member(ti(defeq(shead(Name,[]),Def),_),Z).

testmark(Name,UName) :- append("Test",Rest,Complete),name(Name,Complete),name(UName,Rest).

runtests([],_,[]).
runtests([test(N,T)|Rest],Z,Errors) :-
    getschemat(Z,N,Def1),
    getschemat(Z,T,Def2),
    !,
    write('Z transformation test '),write(N),write(': '),
    ( equal_expressiont(Def1,Def2) ->
        write(ok),
        Errors = RestErrors
    ; otherwise ->
        write(failed),
        Errors = [T|RestErrors]),
    !,
    nl,
    runtests(Rest,Z,RestErrors).
