:- module(symmetry_reduction,[get_symmetry_reduction_constraint/4]).

:- use_module(library(lists)).
:- use_module(library(sets),[subtract/3]).
:- use_module(probsrc(bsyntaxtree)).

% Exclude symmetric changes that only cause syntactical changes when searching for a further solution
% and do not change the semantics of the synthesized program.
% Components are tuples of raw component name and internal component name like (add,add1).
get_symmetry_reduction_constraint(Components,LocationVarsIn,LocationVarsOut,SymmetryReduction) :-
    findall((Name,InternalName),(member((Name,InternalName),Components) , Name \= constant-_ ,
                                 is_symmetric(Name)),SymmetricComponents) ,
    maplist(get_single_operand_symmetry_reduction(LocationVarsIn),SymmetricComponents,OperandExclusionList) ,
    get_operator_symmetry_reduction(Components,LocationVarsOut,OperatorReductionPredicate) ,
    conjunct_predicates([OperatorReductionPredicate|OperandExclusionList],SymmetryReduction).

% Exclude symmetric changes within a component by asserting L(i1) < L(i2).
get_single_operand_symmetry_reduction(LocationVarsIn,Expression,SymmetryReduction) :-
    Expression = (_,InternalName) ,
    get_location_var_from_list(InternalName,left_input,LocationVarsIn,LeftLocationVarIn) ,
    get_location_var_from_list(InternalName,right_input,LocationVarsIn,RightLocationVarIn) ,
    SymmetryReduction = b(less(LeftLocationVarIn,RightLocationVarIn),pred,[]).

% Exclude symmetric changes of operands (components). E.g., given three additions add1,add2,add3 we assert
% L(add1) < L(add2) < L(add3) to hold, where L(add_) is the corresponding output location of the component.
get_operator_symmetry_reduction(Components,LocationVarsOut,SymmetryReduction) :-
    split_equivalent_components(Components,[],EquivalentComponents) ,
    maplist(get_operator_symmetry_reduction_aux(LocationVarsOut,[]),EquivalentComponents,SymmetryExclusionList) ,
    conjunct_predicates(SymmetryExclusionList,SymmetryReduction).

get_operator_symmetry_reduction_aux(_,Acc,[],SymmetryReduction) :-
    conjunct_predicates(Acc,SymmetryReduction).
get_operator_symmetry_reduction_aux(LocationVarsOut,Acc,[Components|T],SymmetryReduction) :-
    % skip single components
    length(Components,Length) ,
    Length > 1 , !,
    get_operator_symmetry_reduction_aux2(Components,LocationVarsOut,[],OperatorReduction) ,
    get_operator_symmetry_reduction_aux(LocationVarsOut,[OperatorReduction|Acc],T,SymmetryReduction).
get_operator_symmetry_reduction_aux(LocationVarsOut,Acc,[_|T],SymmetryReduction) :-
    get_operator_symmetry_reduction_aux(LocationVarsOut,Acc,T,SymmetryReduction).

% Components are all equivalent like [add1,add2,add3] and the list has a minimum length of 2.
% For just one component we don't need symmetry reduction on the level of operators.
get_operator_symmetry_reduction_aux2([_],_,Acc,OperatorReduction) :-
    conjunct_predicates(Acc,OperatorReduction).
get_operator_symmetry_reduction_aux2([(_,InternalName1),(Name,InternalName2)|T],LocationVarsOut,Acc,OperatorReduction) :-
    get_location_var_from_list(InternalName1,output,LocationVarsOut,LocationVarOutComp1) ,
    get_location_var_from_list(InternalName2,output,LocationVarsOut,LocationVarOutComp2) ,
    NewAcc = [b(less(LocationVarOutComp1,LocationVarOutComp2),pred,[])|Acc] ,
    get_operator_symmetry_reduction_aux2([(Name,InternalName2)|T],LocationVarsOut,NewAcc,OperatorReduction).

% Find all equivalent components like [add1,add2,add3].
split_equivalent_components([],Acc,Acc).
split_equivalent_components([(Name,InternalName)|T],Acc,EquivalentComponentsList) :-
    Name \= constant-_ ,
    findall(EquiInternalName,(member((_,EquiInternalName),T) , EquiInternalName = InternalName),EquivalentComponents) ,
    subtract(T,EquivalentComponents,NT) ,
    split_equivalent_components(NT,[(Name,InternalName)|Acc],EquivalentComponentsList).
split_equivalent_components([_|T],Acc,EquivalentComponentsList) :-
    split_equivalent_components(T,Acc,EquivalentComponentsList).

is_symmetric(not_equal).
is_symmetric(equal).
is_symmetric(conjunct).
is_symmetric(disjunct).
is_symmetric(equivalence).
is_symmetric(add).
is_symmetric(multiplication).
is_symmetric(union).
is_symmetric(intersection).

% Search a location variable for a given tuple of component name and internal name (like add1)
% and a component type (i.e., one of [left_input, right_input, output]).
get_location_var_from_list(InternalName,ComponentType,LocationVars,LocationVar) :-
    member(LocationVar,LocationVars) ,
    get_texpr_info(LocationVar,Info) ,
    member(synthesis(InternalName,ComponentType),Info).
