:- module(fuzzing,[fuzz/2,fuzz/3,reproduce_test/3,
                   generate/2,error_detected/1]).

:- use_module(library(lists)).
:- use_module(library(timeout),[time_out/3]).
:- use_module(library(random),[getrand/1,setrand/1]).
:- use_module(library(file_systems),[file_members_of_directory/3]).

:- multifile generate/2.
:- multifile shrink/3.

include_type_definition(_-FullPath) :- consult(FullPath).

:- file_members_of_directory('./types/','*.pl',FileList),
   maplist(include_type_definition,FileList).

:- meta_predicate fuzz(1,+).
:- meta_predicate fuzz(1,+,+).

fuzz(Module:Predicate/Arity,Arguments) :-
    fuzz(Module:Predicate/Arity,50000,Arguments).
% fuzzing with specific amount of tests
fuzz(Module:Predicate/Arity,Testcount,Arguments) :-
    (current_predicate(Module:Predicate/Arity) ->
        true
    ;   error_process(existence_error,Predicate,_,_,_) , fail) ,
    % split arguments by ':'
    get_types(Arguments,Types) ,
    length(Types,Arity) ,
    getrand(Seed) ,
    format('Start fuzzing predicate ~w/~w',[Predicate,Arity]) , nl ,
    format('First state is ~w',[Seed]) , nl ,
    % run randomized tests
    (run_tests(Predicate,Types,Module,Testcount,Result) ->
       (Result = true ->
            nl , format('All tests passed',[]) , nl
        ;   % error has already been printed by run_tests
            true)
    ;   % run tests failed, error in the code
        format('*** Test FAILED for seed ~w ***',[Seed]) , nl,
        error_process(generation_error,Predicate,_,Types,_)).
fuzz(_:Predicate/Arity,_,_) :- error_process(not_enough_arguments,Predicate/Arity,_,_,_).

:- meta_predicate reproduce_test(1,+,+).

% reproduce test case by using a given seed
reproduce_test(Module:Predicate/Arity,Arguments,Seed) :-
    get_types(Arguments,Types) ,
    length(Types,Arity) ,
    format('Start fuzzing predicate ~w/~w for given seed',[Predicate,Arity]) , nl ,
    setrand(Seed) ,
    % run single test
    (run_tests(Predicate,Types,Module,1,Result) ->
       (Result = true ->
            format('Test passed for seed ~w',[Seed]) , nl
        ;   % error has already been printed by run_tests
            true)
    ;   % run tests failed
        format('*** Test FAILED for seed ~w ***',[Seed]) , nl,
        error_process(generation_error,Predicate,_,Types,_)).
reproduce_test(_:Predicate/Arity,_,_) :- error_process(not_enough_arguments,Predicate/Arity,_,_,_).

:- dynamic error_detected/1.
error_detected(false).

set_error_detected :-
    error_detected(false) , ! ,
    retractall(error_detected(_)) ,
    assert(error_detected(true)).
set_error_detected.

run_tests(_,_,_,0,true).
% run tests with random arguments for a prolog predicate
run_tests(Predicate,Types,Module,Testcount,Result) :-
    getrand(Seed) ,
    random_arguments(Types,Values) ,
    Term =.. [Predicate|Values] ,
    call_term(Module,Term,Error) ,
    % write '.' every thousandth testcase
    RunningSign is mod(Testcount,1000) ,
    (RunningSign = 0 ->
        write('.')
    ;   true) ,
    (Error \= none ->
        Result = false ,
        % try shrinking arguments and print error
        nl , format('Error detected, shrink arguments',[]) , nl ,
        % don't print input from mutation(Input:Type) for user readability
        minimum_typelist(Types,NTypes) ,
        shrink_values(Predicate,Module,Types,Values,Shrunken) , nl ,
        set_error_detected ,
        error_process(Error,Predicate,Shrunken,NTypes,Seed) , nl
    ;   % go on with testing if no error detected
        Error = none ,
        NTestcount is Testcount - 1 ,
        run_tests(Predicate,Types,Module,NTestcount,Result)).

% calls a term within its given module with error and timeout exception
call_term(Module,Term,Error) :-
    (time_out(on_exception(_,Module:call(Term),fail),5000,Result) ->
        (Result = success ->
            Error = none
        ; % timeout
            Error = timeout)
    ; % predicate failed
        Error = false).

% shrink arguments
shrink_values(Predicate,Module,Types,Values,Result) :-
    % write sth at every shrinking step
    write('.') ,
    maplist(shrink_arguments,Types,Values,Shrunken) ,
    % termination condition
    Values \= Shrunken ,
    Term =.. [Predicate|Shrunken] ,
    % catch timeout and error exception
    \+ time_out(on_exception(_,Module:call(Term),fail),1000,success) ,
    shrink_values(Predicate,Module,Types,Shrunken,Result).
shrink_values(_,_,_,Result,Result).

shrink_arguments(Type,Value,Shrunken) :-
    shrink(Type,Value,Shrunken).
shrink_arguments(_,Value,Value).

% convert types divided by ':' to a list
get_types(Type:T,[Type|NT]) :-
    get_types(T,NT) , !.
get_types(Type,[Type]).

% generate random arguments from a list of types
random_arguments([],[]).
random_arguments([Type|T1],[Value|T2]) :-
    generate(Type,Value) ,
    random_arguments(T1,T2).

% make typelist readable, i.e. don't print input
% from mutation(Input:Type)
minimum_typelist([],[]).
minimum_typelist([mutation(_:Type)|T],[mutation(Type)|NT]) :- ! ,
    minimum_typelist(T,NT).
minimum_typelist([Type|T],[Type|NT]) :-
    minimum_typelist(T,NT).

% error handling
error_process(existence_error,Predicate,_,_,_) :-
    format('Predicate ~w does not exist.',[Predicate]) , nl.

error_process(not_enough_arguments,Predicate,_,_,_) :-
    format('Wrong amount of arguments for predicate ~w',[Predicate]) , nl.

error_process(generation_error,_,_,Types,_) :-  % error in generate/2
    format('Unknown type or an error in a prolog file for type of ~w',[Types]) , nl.

error_process(error(type_error(_,_),_),Predicate,_,Types,_) :-
    format('Wrong type of arguments in predicate ~w of type ~w',[Predicate,Types]) , nl.

error_process(timeout,Predicate,Values,Types,Seed) :-
    length(Types,Arity) ,
    format('Timeout in ~w/~w for input ~w of type ~w',[Predicate,Arity,Values,Types]) , nl ,
    format('Seed for reproducing test case: ~w',[Seed]) , nl.

error_process(false,Predicate,Values,Types,Seed) :-
    length(Types,Arity) ,
    format('Predicate ~w/~w false for input ~w of type ~w',[Predicate,Arity,Values,Types]) , nl ,
    format('Seed for reproducing test case: ~w',[Seed]) , nl.
